#!/usr/local/bin/ruby
# $Id: cutil.rb,v 1.11 2002/02/08 04:33:38 Black Exp $
#                                                      Project Gemini
#---------------------------------------------------------------------
#
#   Corpus Utility
#
#       Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp>
######################################################################
require File.dirname(File.expand_path(__FILE__)) + "/nlputil.rb"


#
#   Global Method
######################################################################

$VERSION = <<_VERSIONEND_
Corpus Utility   ver. 1.3                                 [gemini 1.2]
         Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp>

_VERSIONEND_

$USAGE = <<_USAGEEND_
usage:
  cutil.rb [mode] (option)

mode:
  extract [corpus files]
    Extract sentences from corpora
  (option)
    -c [count]  (default = 100)
      extract count (be padded if data is not enough)

  divide -p prefix [corpus files]
    Divide corpora by length of sentnece
  (option)
    -p [prefix]
      prefix name for divided corpus
    -w [width]  (default = 1)
      collect sentences according to width
      e.g.
        -w 2  =  1,2   ; 3,4   ; ...
        -w 3  =  1,2,3 ; 4,5,6 ; ...

  select [corpus file] [block names]
    Extract data blocks from a corpus

  count [corpus files]
    Count sentences in corpora

  sample [corpus file] [spec]
    Extract specified sentences from a corpus
  (spec format)
    1,2,3 = [1], [2] and [3]
    1-4,7 = [1], [2], [3], [4] and [7]

  unite [corpus files]
    Unite some corpora
_USAGEEND_

def print_read_count ( count )
  v_print { "\r>>> #{count} (sentences)" }
end


#
#   extract
######################################################################

def mode_extract ()
  v_print { " + MODE = extract\n" }
  
  $Count = 100
  $Count = Integer($OPT_c) if $OPT_c
  v_print { " + extract-size = #{$Count}\n" }
  
  # buffer ɕf[^~ςĂ
  buffer = [ ]
  
  $file = [ $stdin ] if $file.empty?
  $file.each do |filename|
    v_print { "   - #{io_to_name(filename)}\n" }
    if String === filename then
      unless File::exist?( filename ) then
        $_ = "No such file -- #{filename}\n"
        if $OPT_v then
          v_print { "     ! #{$_}" }
        else
          $stderr << $_
        end
        next
      end
      stream = open( filename )
    else
      stream = filename
    end
    
    CorpusStream.new( stream ) do |root|
      begin
        root.each do |corpus|
          readcount = 0
          t = Time.now.to_i
          corpus.each do |sentence|
            if t+1 <= Time.now.to_i then
              t = Time.now.to_i
              v_print { "     - #{readcount}\r" }
            end
            if buffer.size >= $Count and rand(100) < 50 then
              sentence.close() ; readcount += 1
              next
            end
            
            image = [ ]
            sentence.dump( image ) ; readcount += 1
            image = image.join("")
            
            if buffer.size < $Count then
              buffer << image
            else
              buffer[ rand( buffer.size ) ] = image
            end
          end
          v_print { "     - #{readcount}\n" }
        end
      rescue
        $stderr << "#{root.stream.path}:#{root.stream.lineno}: #{$!}\n"
        exit
      end
    end
  end
  
  # Ȃꍇ͖ߍ킹
  while buffer.size < $Count do
    $_ = rand( buffer.size )
    buffer << buffer[ $_ ]
  end
  Ext::Block.display( "pCorpus", $defout ) do
    buffer.each do |item|
      $defout << item
    end
  end
end


#
#   divide
######################################################################

# Ruby 1.6.5 Ȃǂł Range#eql? ɃoÔ Range ͎gȂ
def  length_to_range  ( len )
  # e.g.  ($Width == 4)
  #  1 --> [1,4], 2 --> [1,4]
  #  7 --> [5,8], 10 --> [9,12]
  if len == 0 then
    return [0, 0]
  end
  block = ((len-1) / $Width).to_i
  return [ block*$Width + 1, (block+1)*$Width ]
end

def  mode_divide  ()
  v_print { " + MODE = divide\n" }
  
  unless $OPT_p then
    $stderr << "ERROR: Prefix is empty. Use an option -p.\n"
    exit
  end
  
  $Prefix = $OPT_p
  $Width  = 1
  
  $Width = Integer($OPT_w) if $OPT_w
  v_print { " + width: #{$Width}\n" }
  
  buffer    = [ ]
  container = { }  #  range ɑΉ IO IuWFNg ( IO IuWFNg̔z)
  
  readcount = 0
  
  $file = [ $stdin ] if $file.empty?
  $file.each do |filename|
    v_print { "   - #{io_to_name(filename)}\n" }
    if String === filename then
      unless File::exist?( filename ) then
        $_ = "No such file -- #{filename}\n"
        if $OPT_v then
          v_print { "     ! #{$_}" }
        else
          $stderr << $_
        end
        next
      end
      stream = open( filename )
    else
      stream = filename
    end
    
    CorpusStream.new( stream ) do |root|
      begin
        root.each do |corpus|
          readcount = 0
          t = Time.now.to_i
          corpus.each do |sentence|
            if t+1 <= Time.now.to_i then
              t = Time.now.to_i
              v_print { "     - #{readcount}\r" }
            end
            readcount += 1
            
            # ܂̓ɓǂݍ (ǂݏIȂƕ킩炸o͐킩Ȃ)
            image    = [ ]
            keyrange = nil
            sentence.display( image ) do
              sentence.each( PoSBlock ) do |c, block|
                unless c then
                  block.dump
                else
                  block.out = $defout
                  block.display do
                    keyrange = length_to_range( c.new( block ).read().length )
                    block.close()
                  end
                end
              end
            end
            
            # t@Cւ̏o
            file = container[ keyrange ]
            unless file then
              if keyrange.first != keyrange.last then
                suffix = "#{keyrange.first}-#{keyrange.last}"
              else
                suffix = "#{keyrange.first}"
              end
              out_filename = "#{$Prefix}_#{suffix}.corpus"
              file = container[ keyrange ] = open( out_filename, "w" )
              file << "<pCorpus>\n"
            end
            for line in image do
              file << line
            end
          end
          v_print { "     - #{readcount}\n" }
        end
      rescue
        $stderr << "#{root.stream.path}:#{root.stream.lineno}: #{$!}\n"
        exit
      end
    end
  end
  v_print { " + written files are...\n" }
  container.each_pair do |range, file|
    v_print { "   - #{file.path}\n" }
    file << "</pCorpus>\n"
    file.close
  end
end


#
#   Select
######################################################################

def mode_select ()
  v_print { " + MODE = select\n" }
  
  if File::exist?( $file.first ) then
    targetfile = $file.shift
  else
    if /^[dDpP]/ === $file.first then
      targetfile = $stdin
    else
      $_ = "No such file -- #{$file.first}\n"
      if $OPT_v then
        v_print { "     ! #{$_}" }
      else
        $stderr << $_
      end
      exit
    end
  end
  targetblock = $file.to_a
  if targetblock.empty? then
    $stderr << "ERROR: No target block.\n"
    exit
  end
  stream = (String === targetfile)? open( targetfile ): targetfile
  
  v_print { " + target-block = #{targetblock.join(', ')}\n" \
            "   - #{io_to_name(targetfile)}\n" }
  readcount = 0
  
  CorpusStream.new( stream ) do |root|
    begin
      root.each do |corpus|
        t = Time.now.to_i
        corpus.display do
          corpus.each do |sentence|
            if t+1 <= Time.now.to_i then
              t = Time.now.to_i
              v_print { "     - #{readcount}\r" }
            end
            readcount += 1
            sentence.display do
              sentence.each do |c, block|
                if targetblock.include?( block.name ) then
                  block.dump( $defout )
                else
                  block.close()
                end
              end
            end
          end
        end
        v_print { "     - #{readcount}\n" }
      end
    rescue
      $stderr << "#{root.stream.path}:#{root.stream.lineno}: #{$!}\n"
      exit
    end
  end
end


#
#   Count
######################################################################

def mode_count ()
  v_print { " + MODE = count\n" }
  
  result = Hash.new( 0 )
  $file = [ $stdin ] if $file.empty?
  $file.each do |filename|
    v_print { " + #{io_to_name(filename)}\n" }
    
    stream = (String === filename)? open( filename ): filename
    root = CorpusStream.new( stream )
    begin
      root.each do |corpus|
        readcount = 0
        corpus.each do |sentence|
          readcount += 1
        end
        v_print { "   - #{readcount}\n" }
        result[ filename ] += readcount
      end
    rescue
      $stderr << "#{root.path}:#{root.lineno}: #{$!}\n"
      exit
    ensure
      root.close
    end
  end
  result.each_pair do |filename, n|
    $defout << "#{io_to_name(filename)}\t#{n}\n"
  end
end


#
#   Sample
######################################################################

class Fixnum
  def include? ( n )
    self == n
  end
end

def sample_spec_check ( n )
  $spec.each do |range|
    return true if range.include?( n )
  end
  return false
end

def sample_read_spec ( ary )
  ret = [ ]
  re_comma = /\s*,\s*/
  
  list = [ ]
  ary.each do |s|
    list.concat( s.split( re_comma ) )
  end
  list.each do |s|
    case s
    when /^(\d+)-(\d+)$/ then
      ret << ($1.to_i .. $2.to_i)
    when /^\d+$/ then
      ret << $&.to_i
    else
      unless s.empty? then
        $stderr << "ERROR: spec ``#{s}'' is invalid format.\n"
        exit
      end
    end
  end
  return ret
end

def sample_max_spec ()
  maxnum = 0
  $spec.each do |range|
    if range.is_a?( Range ) then
      maxnum = range.last if maxnum < range.last
    else
      maxnum = range if maxnum < range
    end
  end
  return maxnum
end

def mode_sample ()
  v_print { " + MODE = sample\n" }
  
  if File::exist?( $file.first ) then
    targetfile = $file.shift
  else
    if /^\s*\d+\s*([-,]\s*\d+\s*)?$/ === $file.first then
      $_ = "No such file -- #{$file.first}\n"
      if $OPT_v then
        v_print { "     ! #{$_}" }
      else
        $stderr << $_
      end
      exit
    end
    targetfile = $stdin
  end
  $spec  = sample_read_spec( $file )
  maxnum = sample_max_spec()
  stream = (String === targetfile)? open( targetfile ): targetfile
  
  v_print { " + spec = #{$spec.join(', ')}\n + maxnum = #{maxnum}\n" \
            "   - #{io_to_name(targetfile)}\n" }
  
  CorpusStream.new( stream ) do |root|
    begin
      readcount = 0
      corpus = root.getb()
      if corpus then
        t = Time.now.to_i
        corpus.display do
          corpus.each do |sentence|
            if t+1 <= Time.now.to_i then
              t = Time.now.to_i
              v_print { "     - #{readcount}\r" }
            end
            readcount += 1
            if sample_spec_check( readcount ) then
              sentence.dump
            end
            break if readcount >= maxnum
          end
        end
      end
      v_print { "     - #{readcount}\n" }
    rescue
      $stderr << "#{root.stream.path}:#{root.stream.lineno}: #{$!}\n"
      exit
    end
  end
end


#
#   Unite
######################################################################

def mode_unite ()
  v_print { " + MODE = unite\n" }
  
  totalcount = 0
  Ext::Block.display( "pCorpus" ) do
    $file.each do |filename|
      unless File::exist?( filename ) then
        v_print { "     ! No such file -- #{filename}\n" }
        next
      end
      
      v_print { "    - #{filename}\n" }
      stream = (String === filename)? open( filename ): filename
      CorpusStream.new( stream ) do |root|
        begin
          root.each do |corpus|
            t = 0
            readcount = 0
            corpus.each do |sentence|
              readcount += 1
              sentence.dump
              if t+1 <= Time.now.to_i then
                v_print { "      - #{readcount}\r" }
              end
            end
            v_print { "      - #{readcount}\n" }
            totalcount += readcount
          end
        rescue
          $stderr << "#{root.path}:#{root.lineno}: #{$!}\n"
          exit
        end
      end
    end
    v_print { " + total = #{totalcount}\n" }
  end
end


#
#   Main Program
######################################################################

getopts_ex( nil, 'c:', 'w:', 'p:' )
v_print $VERSION

$mode = ARGV.shift
$file = ARGV.dup
case $mode
when "extract" then
  mode_extract()
when "divide" then
  mode_divide()
when "select" then
  mode_select()
when "count" then
  mode_count()
when "sample" then
  mode_sample()
when "unite" then
  mode_unite()
else
  if $mode then
    $stderr << "ERROR: Mode `#{$mode}' is unknown."
  end
  printUsage()
end

