#!/usr/local/bin/ruby
# $Id: earleyparser.rb,v 1.15 2002/02/08 04:33:13 Black Exp $
#                                                      Project Gemini
#---------------------------------------------------------------------
#
#   Earley Parser
#
#       Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp>
######################################################################
require File.dirname(File.expand_path(__FILE__)) + "/nlputil.rb"


#
#   Global Method
######################################################################

# ftHgl
$StartName  =  '<start>'
$Epsilon    =  0.000001

$VERSION = <<"--------"
Earley Parser   ver. 1.6                                  [gemini 1.2]
         Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp>

--------

$USAGE = <<"--------"
usage:
  ruby earleyparser.rb (options) (corpus file)

options:
  -c [count]
    iteration count (> 0)
  -e [precision]
    threshold to stop EM iteration  (default = #{$Epsilon})
  -g [grammar file]
    using grammar
  -l [file (output)]
    enable EM learning (Stolcke\'s algorithm)
  -s [name]
    start symbol  (default = #{$StartName})
  --prune
    enable pruning unused Earley items
  --random
    force to set random probability
  --loop [count]
    set learning loop\'s count
--------


#
#   Main Program
######################################################################

getopts_ex( nil, 'a:', 'c:', 'e:', 'l:', 's:', 'prune', 'random', 'log:', 'time_r_u', 'bench:', 'loop:' )
v_print $VERSION

$files = ARGV.dup
$files = [ STDIN ] if $files.empty?

# -s IvV͊JnL (KI[LɂȂ)
$StartName = $OPT_s  if $OPT_s
$StartName = "<#{$StartName}>"  unless /\A<.*>\z/ =~ $StartName

# EM wK臒l
if $OPT_e then
  $OPT_e = $OPT_e.to_f
  $Epsilon = $OPT_e if $OPT_e > 0
end

# JԂ (Kl)
$OPT_c = $OPT_c.to_i if $OPT_c

# o͂Ɏgp郍Of[^p̃nbVe[u
$ResultLog = { }


#
#   Prepare grammar
######################################################################
# $StartName : ł̊JnL\
# $Start     : lID ł̊JnL\

printUsage() unless $grammar
v_print { " + grammar = #{$OPT_g}\n" }

$Start = $grammar.getcat( $StartName )
unless $Start then
  $stderr << "Error: `#{$StartName}' is an unknown symbol.\n"
  exit
end
$Start = $Start.id
v_print { " + start symbol = `#{$StartName}' (\##{$Start})\n" }

$extgrammar = toPCFG( Ext::Grammar.new( $grammar ), $OPT_random )


#
#   Create parser
######################################################################

if $OPT_bench then
  $parser = Ext::StolckeParser::Matrix::BenchMark.new( $extgrammar )
  $OPT_loop = $OPT_bench
else
  $parser = Ext::StolckeParser::Matrix.new( $extgrammar )
end
$parser.pruning = true if $OPT_prune


#
#   Helper
######################################################################

def output_log ( out, chartary )
  chartary.each do |chart|
    items = chart.list_items()
    items.each do |item|
      next if item.beta == 0 or item.gamma == 0
      out << sprintf( "(%d,%d) %s :: i=%g, o=%g\n", item.left, item.right, item.to_r().typestring( item.dot ), item.gamma, item.beta )
    end
    out << "\n"
  end
end

def print_likelihood ()
  v_print do
    tstr = Time.now.strftime('%Y/%m/%d %H:%M:%S')
    sprintf( "\#%4d (#{tstr}): likelihood = %.6e (%f sec)\n", $itrcount, $likelihood, $itrtime )
  end
end

def print_progress ()
  if $memtotal then
    v_print { sprintf( "    - \#%d (%.1f KB memory) (%.1f sec)\r", $total, $memtotal/1024, $totaltime.ptime ) }
  else
    v_print { sprintf( "    - \#%d (%.1f sec)\r", $total, $totaltime.ptime ) }
  end
end


#
#   Parsing
######################################################################

HistoryData = Struct.new( :num, :likelihood, :tms )

$history   = [ ]
$total     = 0
$success   = 0
$failure   = 0
$memtotal  = ($OPT_l or $OPT_log)? 0: nil
$totaltime = Struct::Tms.new( 0, 0, 0, 0 )
$itrcount  = 1

if $OPT_time_r_u then
  time_block { $parser.compute_unit_matrix() }
  $r_u = [ $time ]
  $stderr << " + R_U time: #{$time.ptime}\n"
end

$likelihood = 0.0

Ext::Block.display( "pCorpus" ) do
  $parser.update do
    $files.each do |filename|
      v_print { " + #{io_to_name(filename)}\n" }
      if String === filename then
        unless File::exist?( filename ) then
          if $OPT_v then
            v_print { "   ! No such file\n" }
          else
            $stderr << "No such file -- #{filename}\n"
          end
          next
        end
        stream = open( filename )
      else
        stream = filename
      end
      
      CorpusStream.new( stream ) do |root|
        root.each do |corpus|
          t = Time.now.to_i ; print_progress()
          corpus.each do |sentence|
            $total += 1
            begin
              sentence.out = $defout
              sentence.display do
                
                # ̓ǂݍ
                poslist = nil
                sentence.each( PoSBlock ) do |c, block|
                  if c then
                    if poslist then
                      raise NLPException, "Found too many <dPoS>."
                    else
                      block.display do
                        poslist = c.new(block).read
                      end
                    end
                  else
                    if ItemSetBlock === block then
                      block.skip
                    else
                      block.dump
                    end
                  end
                end
                break unless poslist
                
                # \
                poslist.compile( $grammar )
                $chart = time_block { $parser.parse( poslist ) }
                $totaltime += $time
                
                local_likelihood = $chart.likelihood()
                if local_likelihood then
                  $success += 1 # \͐
                  if $OPT_l or $OPT_log then # wKsꍇ̓obt@ɕۑ (p[^XV update ōs)
                    $memtotal   += $chart.memsize()
                    $likelihood += local_likelihood
                    $parser << $chart
                  end
                  if t+1 <= Time.now.to_i then # i󋵂̕\
                    t = Time.now.to_i ; print_progress()
                  end
                  Ext::Block.display( "dItemSet" ) do
                    $defout << "\#@ROOT = #{$StartName}\n"
                    $chart.display_itemset( $defout )
                  end
                else
                  $failure += 1 # \͎s
                  raise NLPException, "Parsing failed."
                end
              end # sentence.display
            rescue
              v_print "\n"
              $stderr << "\##{$total} (#{root.stream.path}:#{root.stream.lineno}): #{$!}\n"
              t = 0
            end
          end # corpus.each
          print_progress() ; v_print { "\n" }
        end # root.each
      end # CorpusStream.new()
    end # $files.each
  end # $parser.update
end # <pCorpus>.display

time_parse = $totaltime.ptime
time_learn = 0.0
$itrtime = time_parse

$ResultLog['parse-time'   ] = "#{time_parse} \# sec"
$ResultLog['corpus-size'  ] = $total
$ResultLog['success-count'] = $success
$ResultLog['failure-count'] = $failure if $failure > 0
Ext::Block.display( "dLog" ) do
  $defout << hash_to_pstr( $ResultLog )
end
if $OPT_log then
  $logfile = open( $OPT_log, "w" )
  $logfile << "---- iteration \#1\n"
  output_log( $logfile, $parser.chartlist )
end

# wKsȂꍇ͂ŏI
unless $OPT_l then
  v_print { " + finished.\n" }
  exit
end

$chart    = nil
$chartary = $parser.to_a
if $chartary.empty? then
  $stderr << "ERROR: No sentences.\n"
  exit
end

if $OPT_loop then
  $parser.loopcount = $OPT_loop.to_i
  v_print { " + loopcount = #{$parser.loopcount}\n" }
end

print_likelihood()
$history << HistoryData.new( $itrcount, $likelihood, $totaltime )


#
#   Learning
######################################################################

loopcount = $OPT_c || -1
begin
  until loopcount >= 0 and $itrcount >= loopcount do
    $likelihood = 0.0
    if $OPT_time_r_u then
      time_block { $parser.compute_unit_matrix() }
      $r_u << $time
    end
    $likelihood = time_block { $parser.update() }
    $itrtime    = $time.ptime
    $itrcount  += $parser.loopcount
    time_learn += $itrtime
    
    print_likelihood()
    $history << HistoryData.new( $itrcount, $likelihood, $time )
    
    if $OPT_log then
      $logfile << "---- iteration \##{$itrcount}\n"
      output_log( $logfile, $chartary )
    end
    
    break if $likelihood == 0
    
    delta = $history[-1].likelihood - $history[-2].likelihood
    if delta < 0 then
      $stderr << "Error: Likelihood is down. (fatal error)\n"
      exit
    end
    
    break if delta < $Epsilon
  end
rescue NLPException
  $stderr << "Error: #{$!}\n"
end


#
#   Output parameter
######################################################################

if $r_u then
  $_ = 0
  $r_u.each do |t|
    $_ += t.ptime
  end
  $r_u = $_ / $r_u.size
end

t_per_itr = ($itrcount <= 0)? 0: time_learn / ($itrcount-1)

$ResultLog['itr-count' ] = $itrcount
$ResultLog['R_u-time'  ] = "#{$r_u} \# sec/itr"      if $r_u
$ResultLog['learn-time'] = "#{t_per_itr} \# sec/itr" if $itrcount > 1
$ResultLog['loopcount' ] = $parser.loopcount         if $OPT_loop

open( $OPT_l, "w" ) do |io|
  $grammar.display( io )
  Ext::Block.display( "dLog", io ) do
    $defout << hash_to_pstr( $ResultLog )
  end
end
v_print { " + written #{$OPT_l}.\n" }

