/*                                                     project Gemini 
----------------------------------------------------------------------
    lCeBũubNXg[
      $Id: bstream.c,v 1.1 2002/02/06 12:45:22 Black Exp $

        Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp> 
----------------------------------------------------------------------
*/
#include "ruby.h"
#include <string.h>
#include <stdlib.h>
#include <memory.h>

#define PRINT_LOCK    0
#define PRINT_UNLOCK  0
#define PRINT_CLOSE   0


static VALUE rb_cPacketBlock;
static VALUE rb_cDataBlock;
static VALUE rb_cBlock;

static VALUE rbstr_nl;
static VALUE rbstr_opentag_head;
static VALUE rbstr_closetag_head;
static VALUE rbstr_tag_tail;
static VALUE global_object_cache;

static ID id_gets;
static ID id_close;


/*   ubNp̃f[^\
----------------------------------------------------------------------
*/

struct tagBlock
{
    VALUE  obj_cache; /* GC }[Np(0:parent/stream, 1:ubN, 2: _v) */
    
    VALUE  parent; /* eubN */
    VALUE  stream; /* f[^̃Xg[ */
    VALUE  out;    /* _ṽXg[ */
    
    int  locked; /* bNJE^ (-1: N[Yς) */
    int  lineno; /* Jn̍sʒu */
    
    VALUE  nameobj; /* ubN */
};
typedef  struct tagBlock  Block;


/*   `FbN
----------------------------------------------------------------------
 - ^O̒Ԃ
*/

static int is_tagopen_line ( const char * ptr )
{
    const char * s;
    
    if( *ptr != '<' )
        return -1;
    ++ptr;
    
    s = ptr;
    if( *ptr != 'p' && *ptr != 'P' && *ptr != 'd' && *ptr != 'D' )
        return -1;
    ++ptr;
    while( ('a' <= *ptr && *ptr <= 'z') || ('A' <= *ptr && *ptr <= 'Z') || *ptr == '-' || *ptr == '_' )
        ++ptr;
    if( ptr == s || ptr[0] != '>' || (ptr[1] != '\n' && ptr[1] != '\0') )
        return -1;
    
    return ptr - s;
}

static int is_tagclose_line ( const char * ptr, const char * tagname )
{
    if( tagname[0] == '\0' )
        return 0; /* [gpPbg̏I[ EOF ̂ */
    
    if( ptr[0] != '<' || ptr[1] != '/' )
        return 0;
    ptr += 2;
    
    while( *tagname != '\0' ) {
        if( *ptr != *tagname )
            return 0;
        ++ptr;
        ++tagname;
    }
    if( ptr[0] != '>' || (ptr[1] != '\n' && ptr[1] != '\0') )
        return 0;
    
    return 1;
}


/*   [eBeB
----------------------------------------------------------------------
*/

static VALUE return_nil ( VALUE self )
{
    return Qnil;
}

static int is_io_object ( VALUE obj )
{
    return (rb_obj_class(obj) == rb_cIO);
}

#define  is_root(BLOCK)  NIL_P((BLOCK)->parent)
#define  is_closed(BLOCK)  ((BLOCK)->locked < 0)


static char * get_block_tagname ( Block * block )
{
    return (NIL_P(block->nameobj))? "": RSTRING(block->nameobj)->ptr;
}

static void rb_raise_notfound_closetag ( const char * name )
{
    rb_raise( rb_eRuntimeError, "``</%s>'' is not found.", name );
}

static void block_lock ( Block * block )
{
    block->locked += 1;
}

static void block_unlock ( Block * block )
{
    if( block->locked <= 0 )
        rb_raise( rb_eRuntimeError, "[BUG/bstream.so] Too many unlocking." );
    block->locked -= 1;
}

static void block_is_locked ( Block * block )
{
    if( block->locked != 0 )
        rb_raise( rb_eRuntimeError, "Block ``%s'' is locked.", STR2CSTR(block->nameobj) );
}

static void block_unlock_parent ( Block * block )
{
    Block * parent;
    
    if( !is_root(block) ) {
        Data_Get_Struct( block->parent, Block, parent );
        block_unlock( parent );
    }
}

static void rb_write_opentag_internal ( VALUE out, VALUE name )
{
    rb_io_write( out, rbstr_opentag_head );
    rb_io_write( out, name );
    rb_io_write( out, rbstr_tag_tail );
}

static void rb_write_opentag ( VALUE out, Block * block )
{
    rb_write_opentag_internal( out, block->nameobj );
}

static void rb_write_closetag_internal ( VALUE out, VALUE name )
{
    rb_io_write( out, rbstr_closetag_head );
    rb_io_write( out, name );
    rb_io_write( out, rbstr_tag_tail );
}

static void rb_write_closetag ( VALUE out, Block * block )
{
    rb_write_closetag_internal( out, block->nameobj );
}

#define  Stream_Gets(STREAM,isIO)  ((isIO)? rb_io_gets(STREAM): rb_funcall(STREAM, id_gets, 1, rbstr_nl))


/*   ANZX\bh
----------------------------------------------------------------------
*/

static VALUE block_set_out ( VALUE self, VALUE out )
{
    const ID id_lineno = rb_intern("lineno");
    Block * block;
    
    Data_Get_Struct( self, Block, block );
    block->out = out;
    RARRAY(block->obj_cache)->ptr[2] = out;
    
    return out;
}

static VALUE block_get_out ( VALUE self )
{
    Block * block;
    
    Data_Get_Struct( self, Block, block );
    
    return block->out;
}

static VALUE block_get_stream ( VALUE self )
{
    Block * block;
    
    Data_Get_Struct( self, Block, block );
    
    return block->stream;
}

static VALUE block_is_closed ( VALUE self )
{
    Block * block;
    
    Data_Get_Struct( self, Block, block );
    
    return (is_closed(block))? Qtrue: Qfalse;
}


/*   ubN̐
----------------------------------------------------------------------
*/

static void block_free ( Block * block )
{
    if( block ) {
        rb_gc_unregister_address( &block->obj_cache );
    }
}

static Block * block_initialize_internal ( Block * block, VALUE parent, const char * name, int n )
{
    const ID id_lineno = rb_intern("lineno");
    
    if( name ) {
        /* parent ɂ͐eƂȂ Ext::PacketBlock w肳Ă */
        block->parent = parent;
        block->stream = block_get_stream( parent );
        block->out    = block_get_out( parent ); /* e out p */
    }
    else {
        /* parent ɂ͓͂ƂȂXg[w肳Ă */
        block->parent = Qnil; 
        block->stream = parent;
        block->out    = Qnil;
    }
    block->locked = 0;
    block->lineno = NUM2INT( rb_funcall( block->stream, id_lineno, 0 ) );
    
    if( n < 0 )
        rb_raise( rb_eRuntimeError, "[BUG/bstream.c] Block name's length is minus." );
    if( name )
        block->nameobj = rb_str_new( name, n );
    else
        block->nameobj = Qnil;
    
    block->obj_cache = rb_ary_new3( 3, parent, block->nameobj, Qnil );
    rb_gc_register_address( &block->obj_cache );
    
    return block;
}

static VALUE block_create ( VALUE parent, const char * name, int n )
{
    Block  *block;
    VALUE   obj;
    
    switch( (!name)? 'p': name[0] )
    {
    case 'p':  case 'P':
        obj = Data_Make_Struct( rb_cPacketBlock, Block, NULL, block_free, block );
        break;
    case 'd':  case 'D':
        obj = Data_Make_Struct( rb_cDataBlock, Block, NULL, block_free, block );
        break;
    default:
        rb_raise( rb_eRuntimeError, "[BUG/bstream.so] Illegal block name ``%s''.", name );
    }
    block_initialize_internal( block, parent, name, n );
    
    return obj;
}


/*   ubÑI[v
----------------------------------------------------------------------
*/

static VALUE block_open_child ( VALUE parent )
{
    Block  *block;
    VALUE   stream, line;
    char   *tagname, *parentname;
    int     n, is_io;
    
    Data_Get_Struct( parent, Block, block );
    stream = block->stream;
    is_io  = is_io_object( stream );
    
    tagname    = NULL;
    parentname = get_block_tagname( block );
    while( !tagname ) {
        line = Stream_Gets( stream, is_io );
        if( NIL_P( line ) ) {
            if( !is_root(block) )
                rb_raise_notfound_closetag( get_block_tagname(block) );
            return Qnil; /* [gpPbgȊO EOF ꍇ̓G[ */
        }
        Check_Type( line, T_STRING );
        
        if( is_tagclose_line( RSTRING(line)->ptr, parentname ) ) {
            #if PRINT_UNLOCK
            fprintf( stderr, "block_open_child()/block_unlock(): %s's parent\n", STR2CSTR(rb_String(block->nameobj)) );
            #endif
            block->locked = -1;
            block_unlock_parent( block );
            return Qnil; /* eƂȂubNꂽƂ nil Ԃ */
        }
        
        n = is_tagopen_line( RSTRING(line)->ptr );
        if( n > 0 ) {
            tagname = &RSTRING(line)->ptr[1];
        }
        else {
            if( !NIL_P( block->out ) )
                rb_io_write( block->out, line );
        }
    }
    #if PRINT_LOCK
    fprintf( stderr, "block_open_child()/block_lock(): parent=%s\n", STR2CSTR(rb_String(block->nameobj)) );
    #endif
    block_lock( block );
    
    return block_create( parent, tagname, n );
}

static VALUE block_open_root ( VALUE stream )
{
    Block  *block;
    VALUE   obj = block_create( stream, NULL, 0 );
    
    Data_Get_Struct( obj, Block, block );
    
    return obj;
}

static VALUE block_open ( VALUE klass, VALUE parent )
{
    return (rb_obj_class(parent)==rb_cPacketBlock)? block_open_child(parent): block_open_root(parent);
}


/*   ubNXLbv
----------------------------------------------------------------------
*/

static void block_root_close_internal ( Block * block )
{
    VALUE  line, stream, out;
    int    is_io;
    
    stream = block->stream;
    out    = block->out;
    is_io  = is_io_object( block->stream );
    if( NIL_P( out ) ) {
        while( 1 ) {
            line = Stream_Gets( stream, is_io );
            if( NIL_P( line ) )
                break;
            Check_Type( line, T_STRING );
        }
    }
    else {
        while( 1 ) {
            line = Stream_Gets( stream, is_io );
            if( NIL_P( line ) )
                break;
            Check_Type( line, T_STRING );
            rb_io_write( out, line );
        }
    }
    rb_funcall( stream, id_close, 0 );
    block->locked = -1;
}

static void block_data_close_internal ( int level, Block * block, const char * name )
{
    VALUE  line, stream, out;
    int    is_io;
    
    stream = block->stream;
    out    = block->out;
    is_io  = is_io_object( stream );
    
    /* </...> T */
    if( NIL_P( out ) ) {
        while( 1 ) {
            line = Stream_Gets( stream, is_io );
            if( NIL_P(line) )
                rb_raise_notfound_closetag( name );
            Check_Type( line, T_STRING );
            if( is_tagclose_line( RSTRING(line)->ptr, name ) )
                break;
        }
    }
    else {
        while( 1 ) {
            line = Stream_Gets( stream, is_io );
            if( NIL_P(line) )
                rb_raise_notfound_closetag( name );
            Check_Type( line, T_STRING );
            if( is_tagclose_line( RSTRING(line)->ptr, name ) ) {
                if( level > 0 )
                    rb_io_write( out, line );
                break;
            }
            rb_io_write( out, line );
        }
    }
}

static void block_packet_close_internal ( int level, Block * block, const char * name )
{
    VALUE  line, stream, out;
    char  *lineptr;
    int    n, is_io;
    
    stream = block->stream;
    out    = block->out;
    is_io  = is_io_object( stream );
    
    /* </...> T */
    if( NIL_P( out ) ) {
        while( 1 ) {
            line = Stream_Gets( stream, is_io );
            if( NIL_P( line ) )
                rb_raise_notfound_closetag( name );
            Check_Type( line, T_STRING );
            lineptr = RSTRING(line)->ptr;
            if( is_tagclose_line( lineptr, name ) )
                break;
            
            n = is_tagopen_line( lineptr );
            if( n > 0 ) {
                lineptr[1 + n] = '\0'; /* 삵ĂvH */
                switch( lineptr[1] ) {
                case 'p': case 'P':  block_packet_close_internal( 1, block, &lineptr[1] ); break;
                case 'd': case 'D':  block_data_close_internal( 1, block, &lineptr[1] ); break;
                }
            }
        }
    }
    else {
        while( 1 ) {
            line = Stream_Gets( stream, is_io );
            if( NIL_P( line ) )
                rb_raise_notfound_closetag( name );
            Check_Type( line, T_STRING );
            lineptr = RSTRING(line)->ptr;
            if( is_tagclose_line( lineptr, name ) ) {
                if( level > 0 )
                    rb_io_write( out, line );
                break;
            }
            
            rb_io_write( out, line );
            n = is_tagopen_line( lineptr );
            if( n > 0 ) {
                lineptr[1 + n] = '\0';
                switch( lineptr[1] ) {
                case 'p': case 'P':  block_packet_close_internal( 1, block, &lineptr[1] ); break;
                case 'd': case 'D':  block_data_close_internal( 1, block, &lineptr[1] ); break;
                }
                lineptr[1 + n] = '>'; /* line QƂĂ邱Ƃ̂Ōɖ߂ */
            }
        }
    }
}


/*   ubN
----------------------------------------------------------------------
*/

static VALUE block_data_close ( VALUE self )
{
    Block  *block;
    
    Data_Get_Struct( self, Block, block );
    if( is_closed( block ) )
        return Qnil; /* łɕĂꍇ͉Ȃ */
    
    #if PRINT_CLOSE
    fprintf( stderr, "block_data_close(): %s\n", STR2CSTR(block->nameobj) );
    #endif
    if( is_root( block ) )
        block_root_close_internal( block );
    else {
        block_data_close_internal( 0, block, RSTRING(block->nameobj)->ptr );
        
        block->locked = -1;
        #if PRINT_UNLOCK
        fprintf( stderr, "block_data_close()/block_unlock(): %s's parent\n", STR2CSTR(rb_String(block->nameobj)) );
        #endif
        block_unlock_parent( block );
    }
    
    return self;
}

static VALUE block_packet_close ( VALUE self )
{
    Block  *block;
    
    Data_Get_Struct( self, Block, block );
    if( is_closed( block ) )
        return Qnil; /* łɕĂꍇ͉Ȃ */
    
    #if PRINT_CLOSE
    fprintf( stderr, "block_packet_close(): %s\n", STR2CSTR(block->nameobj) );
    #endif
    if( is_root( block ) )
        block_root_close_internal( block );
    else {
        block_packet_close_internal( 0, block, RSTRING(block->nameobj)->ptr );
        
        block->locked = -1;
        #if PRINT_UNLOCK
        fprintf( stderr, "block_packet_close()/block_unlock(): %s's parent\n", STR2CSTR(rb_String(block->nameobj)) );
        #endif
        block_unlock_parent( block );
    }
    
    return self;
}

static VALUE block_data_skip ( VALUE self )
{
    Block  *block;
    VALUE   outbuf;
    
    Data_Get_Struct( self, Block, block );
    outbuf = block->out;
    block->out = Qnil;
    
    block_data_close( self );
    
    block->out = outbuf;
    return self;
}

static VALUE block_packet_skip ( VALUE self )
{
    Block  *block;
    VALUE   outbuf;
    
    Data_Get_Struct( self, Block, block );
    outbuf = block->out;
    block->out = Qnil;
    
    block_packet_close( self );
    
    block->out = outbuf;
    return self;
}


/*   ̎擾
----------------------------------------------------------------------
*/

static VALUE block_data_gets ( VALUE self )
{
    Block  *block;
    VALUE   line;
    
    Data_Get_Struct( self, Block, block );
    if( is_closed( block ) )
        return Qnil; /* łɃubNĂꍇ */
    
    line = Stream_Gets( block->stream, is_io_object( block->stream ) );
    if( NIL_P( line ) ) {
        if( is_root( block ) )
            return Qnil;
        rb_raise_notfound_closetag( get_block_tagname(block) );
    }
    Check_Type( line, T_STRING );
    
    if( is_tagclose_line( RSTRING(line)->ptr, get_block_tagname(block) ) ) {
        #if PRINT_UNLOCK
        fprintf( stderr, "block_data_gets()/block_unlock(): %s's parent\n", STR2CSTR(rb_String(block->nameobj)) );
        #endif
        block->locked = -1;
        block_unlock_parent( block );
        return Qnil;
    }
    if( block->out )
        rb_io_write( block->out, line );
    
    return line;
}

static VALUE block_data_readlines ( VALUE self )
{
    Block  *block;
    VALUE   ary, line, stream, out;
    int     is_io;
    
    Data_Get_Struct( self, Block, block );
    if( is_closed( block ) )
        return Qnil; /* łɃubNĂꍇ */
    
    ary    = rb_ary_new();
    stream = block->stream;
    out    = block->out;
    is_io  = is_io_object( stream );
    if( NIL_P( out ) ) {
        while( 1 ) {
            line = Stream_Gets( block->stream, is_io );
            if( NIL_P( line ) ) {
                if( is_root( block ) )
                    break;
                rb_raise_notfound_closetag( get_block_tagname(block) );
            }
            Check_Type( line, T_STRING );
            if( is_tagclose_line( RSTRING(line)->ptr, get_block_tagname(block) ) )
                break;
            
            rb_ary_push( ary, line );
        }
    }
    else {
        while( 1 ) {
            line = Stream_Gets( block->stream, is_io );
            if( NIL_P( line ) ) {
                if( is_root( block ) )
                    break;
                rb_raise_notfound_closetag( get_block_tagname(block) );
            }
            Check_Type( line, T_STRING );
            if( is_tagclose_line( RSTRING(line)->ptr, get_block_tagname(block) ) )
                break;
            
            rb_io_write( out, line );
            rb_ary_push( ary, line );
        }
    }
    #if PRINT_UNLOCK
    fprintf( stderr, "block_data_readlines()/block_unlock(): %s's parent\n", STR2CSTR(rb_String(block->nameobj)) );
    #endif
    block->locked = -1;
    block_unlock_parent( block );
    
    return ary;
}

static VALUE block_data_read ( VALUE self )
{
    VALUE  ary = block_data_readlines( self );
    
    return (NIL_P(ary))? Qnil: rb_ary_join( ary, Qnil );
}

static VALUE block_data_each_line ( VALUE self )
{
    Block  *block;
    VALUE   stream, out, line;
    int     is_io;
    
    Data_Get_Struct( self, Block, block );
    if( is_closed( block ) )
        return Qnil; /* łɃubNĂꍇ */
    
    stream = block->stream;
    out    = block->out;
    is_io  = is_io_object( stream );
    if( NIL_P( out ) ) {
        while( 1 ) {
            line = Stream_Gets( stream, is_io );
            if( NIL_P( line ) ) {
                if( NIL_P(block->parent) )
                    break;
                rb_raise_notfound_closetag( get_block_tagname(block) );
            }
            Check_Type( line, T_STRING );
            if( is_tagclose_line( RSTRING(line)->ptr, get_block_tagname(block) ) )
                break;
            
            rb_yield( line );
        }
    }
    else {
        while( 1 ) {
            line = Stream_Gets( stream, is_io );
            if( NIL_P( line ) ) {
                if( NIL_P(block->parent) )
                    break;
                rb_raise_notfound_closetag( get_block_tagname(block) );
            }
            Check_Type( line, T_STRING );
            if( is_tagclose_line( RSTRING(line)->ptr, get_block_tagname(block) ) )
                break;
            
            rb_io_write( out, line );
            rb_yield( line );
        }
    }
    
    #if PRINT_UNLOCK
    fprintf( stderr, "block_data_each_line()/block_unlock(): %s's parent\n", STR2CSTR(rb_String(block->nameobj)) );
    #endif
    block->locked = -1;
    block_unlock_parent( block );
    return self;
}


/*   qubN̎擾
----------------------------------------------------------------------
*/

static VALUE block_packet_getb ( VALUE self )
{
    return block_open_child( self );
}

static VALUE block_packet_each_block ( VALUE self )
{
    VALUE  child;
    
    while( 1 ) {
        child = block_open_child( self );
        if( NIL_P( child ) )
            break;
        rb_yield( child );
        rb_funcall( child, id_close, 0 );
    }
    
    return self;
}


/*   ubN̎擾
----------------------------------------------------------------------
*/

static VALUE block_name ( VALUE self )
{
    Block  *block;
    
    Data_Get_Struct( self, Block, block );
    
    return block->nameobj;
}

static VALUE block_path ( VALUE self )
{
    VALUE   ary = rb_ary_new();
    VALUE   parent;
    Block  *block;
    
    Data_Get_Struct( self, Block, block );
    if( NIL_P( block->parent ) )
        return rb_str_new2( "/" );
    
    for( parent = self; !NIL_P( parent ); parent = block->parent ) {
        Data_Get_Struct( parent, Block, block );
        rb_ary_push( ary, block->nameobj );
    }
    rb_ary_reverse( ary );
    
    return rb_ary_join( ary, rb_str_new2("/") );
}


/*   ubN̏o
----------------------------------------------------------------------
*/

static VALUE block_display_main ( VALUE arg )
{
    VALUE * ptr = RARRAY(arg)->ptr;
    
    if( !NIL_P( ptr[0] ) )
        rb_write_opentag_internal( rb_gv_get("$defout"), ptr[0] );
    rb_yield( Qnil );
    
    return ptr[2];
}

static VALUE block_display_ensure ( VALUE arg )
{
    VALUE * ptr = RARRAY(arg)->ptr;
    
    if( !NIL_P( ptr[0] ) )
        rb_write_closetag_internal( rb_gv_get("$defout"), ptr[0] );
    rb_gv_set( "$defout", ptr[1] );
    
    return ptr[2];
}

static VALUE block_display ( int argc, VALUE * argv, VALUE self )
{
    const ID id_lineno = rb_intern("lineno");
    VALUE   out, ary, origout;
    Block  *block;
    
    if( !rb_block_given_p() )
        rb_raise( rb_eRuntimeError, "no block given" );
    
    origout = rb_gv_get( "$defout" );
    rb_scan_args( argc, argv, "01", &out );
    if( NIL_P( out ) )
        out = origout;
    
    Data_Get_Struct( self, Block, block );
    ary = rb_ary_new3( 3, block->nameobj, origout, self );
    
    rb_gv_set( "$defout", out );
    rb_ensure( block_display_main, ary, block_display_ensure, ary );
    
    return self;
}

static VALUE block_display_s ( int argc, VALUE * argv, VALUE self )
{
    VALUE  out, ary, name, origout;
    
    if( !rb_block_given_p() )
        rb_raise( rb_eRuntimeError, "no block given" );
    
    origout = rb_gv_get( "$defout" );
    rb_scan_args( argc, argv, "11", &name, &out );
    if( NIL_P( out ) )
        out = origout;
    
    switch( RSTRING(name)->ptr[0] ) {
    case 'p': case 'P':
    case 'd': case 'D':
        break;
    default:
        rb_raise( rb_eRuntimeError, "``%s'' is invalid block name.", STR2CSTR(name) );
    }
    ary = rb_ary_new3( 3, name, origout, self );
    
    rb_gv_set( "$defout", out );
    rb_ensure( block_display_main, ary, block_display_ensure, ary );
    
    return self;
}


/*   ubNŜ̃_v
----------------------------------------------------------------------
*/

static VALUE block_dump ( int argc, VALUE * argv, VALUE self )
{
    const ID id_lineno = rb_intern("lineno");
    VALUE   out, outtmp;
    Block  *block;
    
    Data_Get_Struct( self, Block, block );
    outtmp = block->out;
    rb_scan_args( argc, argv, "01", &out );
    if( NIL_P( out ) )
        out = rb_gv_get( "$defout" );
    block->out = out;
    
    /* Xg[Ă܂Ăꍇ͗O */
    if( block->lineno != NUM2INT( rb_funcall( block->stream, id_lineno, 0 ) ) )
        rb_raise( rb_eRuntimeError, "Cannot call Block(<%s>).dump() after reading." );
    
    if( is_root( block ) ) {
        rb_funcall( self, id_close, 0 );
    }
    else {
        rb_write_opentag( out, block );
        rb_funcall( self, id_close, 0 );
        rb_write_closetag( out, block );
    }
    
    block->out = outtmp;
    return self;
}


/*   W[̏R[h
----------------------------------------------------------------------
*/

void Init_bstream ( void )
{
    VALUE  rb_mExt;
    
    rb_mExt   = rb_define_module( "Ext" );
    rb_cBlock = rb_define_class_under( rb_mExt, "Block", rb_cData );
    
    rb_cDataBlock   = rb_define_class_under( rb_mExt, "DataBlock", rb_cBlock );
    rb_cPacketBlock = rb_define_class_under( rb_mExt, "PacketBlock", rb_cBlock );
    
    /* O[of[^ */
    rbstr_nl            = rb_str_new2( "\n" );
    rbstr_opentag_head  = rb_str_new2( "<" );
    rbstr_closetag_head = rb_str_new2( "</" );
    rbstr_tag_tail      = rb_str_new2( ">\n" );
    
    global_object_cache = rb_ary_new3( 4, rbstr_nl, rbstr_opentag_head, rbstr_closetag_head, rbstr_tag_tail );
    rb_global_variable( &global_object_cache );
    
    id_gets  = rb_intern("gets");
    id_close = rb_intern("close");
    
    /* Block x[XNX */
    rb_define_singleton_method( rb_cBlock, "open", block_open, 1 );
    rb_define_singleton_method( rb_cBlock, "display", block_display_s, -1 );
    rb_define_method( rb_cBlock, "out", block_get_out, 0 );
    rb_define_method( rb_cBlock, "out=", block_set_out, 1 );
    rb_define_method( rb_cBlock, "name", block_name, 0 );
    rb_define_method( rb_cBlock, "path", block_path, 0 );
    rb_define_method( rb_cBlock, "dump", block_dump, -1 );
    rb_define_method( rb_cBlock, "stream", block_get_stream, 0 );
    rb_define_method( rb_cBlock, "closed?", block_is_closed, 0 );
    rb_define_method( rb_cBlock, "display", block_display, -1 );
    rb_define_method( rb_cBlock, "gets", return_nil, 0 );
    rb_define_method( rb_cBlock, "getb", return_nil, 0 );
    rb_define_method( rb_cBlock, "read", return_nil, 0 );
    rb_define_method( rb_cBlock, "readlines", return_nil, 0 );
    rb_define_alias( rb_cBlock, "monitor", "out" );
    rb_define_alias( rb_cBlock, "monitor=", "out=" );
    
    /* f[^ubN */
    rb_define_method( rb_cDataBlock, "close", block_data_close, 0 );
    rb_define_method( rb_cDataBlock, "skip", block_data_skip, 0 );
    rb_define_method( rb_cDataBlock, "read", block_data_read, 0 );
    rb_define_method( rb_cDataBlock, "readlines", block_data_readlines, 0 );
    rb_define_method( rb_cDataBlock, "each_line", block_data_each_line, 0 );
    rb_define_alias( rb_cDataBlock, "each", "each_line" );
    
    /* pPbgubN */
    rb_define_method( rb_cPacketBlock, "close", block_packet_close, 0 );
    rb_define_method( rb_cPacketBlock, "getb", block_packet_getb, 0 );
    rb_define_method( rb_cPacketBlock, "skip", block_packet_skip, 0 );
    rb_define_method( rb_cPacketBlock, "each_block", block_packet_each_block, 0 );
    rb_define_alias( rb_cPacketBlock, "each", "each_block" );
}

