/*                                                     project Gemini 
----------------------------------------------------------------------
    `[g\
      $Id: chart.c,v 1.11 2002/02/06 12:44:00 Black Exp $

        Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp> 
----------------------------------------------------------------------
*/
#define DEBUGSW 0
#include "nlputil.h"


/* `[g\̃RXgN^EfXgN^
----------------------------------------------------------------------
*/
void chart_free_internal ( Chart * chart )
{
    Item  *p, *q;
    int    i;
    
    if( chart ) {
        rb_gc_unregister_address( &chart->object_cache );
        if( chart->hash_dot ) /* Item  */
        {
            for( i = 0; i < chart->len; ++i ) {
                for( p = chart_dot( chart, i ); p; p = q ) {
                    q = p->next_dot;
                    ruby_xfree( p );
                }
            }
        }
        ruby_xfree_heaps( 4, chart->hash_dot, chart->hash_lwait, chart->hash_rwait, chart->hash_comp );
    }
}

static void chart_free ( Chart * chart )
{
    if( chart ) {
        chart_free_internal( chart );
        ruby_xfree( chart );
    }
}

VALUE chart_new ( Grammar * grammar, VALUE poslist )
{
    VALUE volatile chartobj;
    Chart  *chart;
    
    /* f[^𑵂
    ---------------------------------------- */
    chartobj = Data_Make_Struct( rb_cChart, Chart, NULL, chart_free, chart );
    chart_init( chart, grammar, poslist );
    
    return chartobj;
}

Chart * chart_init ( Chart * chart, Grammar * grammar, VALUE poslist )
{
    chart->grammar = grammar;
    chart->poslist = poslist;
    
    chart->len    = ru_length( poslist ) + 1; /* ̕Kv (0..12 Ȃ 12 Kv) */
    chart->symbol = grammar->size_lchash;
    
    chart->hash_dot   = rb_alloc_zero( Itemptr, chart->len );
    chart->hash_comp  = rb_alloc_zero( Itemptr, chart->len * chart->len );
    chart->hash_rwait = rb_alloc_zero( Itemptr, chart->len * chart->symbol );
    chart->hash_lwait = rb_alloc_zero( Itemptr, chart->len * chart->symbol );
    
    chart->count_item      = 0;
    chart->count_deleted   = 0;
    chart->count_cancelled = 0;
    
    chart->object_cache = rb_ary_new3( 2, chart->poslist, chart->grammar->orig );
    rb_gc_register_address( &chart->object_cache );
    
    return chart;
}


/* ⏕֐
----------------------------------------------------------------------
*/
static Chart * chart_clear_hash ( Chart * chart )
{
    rb_zero_memory( chart->hash_comp, Itemptr, chart->len * chart->len );
    
    if( chart->grammar )
    {
        rb_zero_memory( chart->hash_lwait, Itemptr, chart->len * chart->grammar->size_lchash );
        rb_zero_memory( chart->hash_rwait, Itemptr, chart->len * chart->grammar->size_lchash );
    }
    
    return chart;
}

static Chart * chart_clear_item ( Chart * chart )
{
    chart_clear_hash( chart );
    
    rb_zero_memory( chart->hash_dot, Itemptr, chart->len );
    
    return  chart;
}


/* `[gɍǉ
----------------------------------------------------------------------
*/
static void chart_addhash ( Chart * chart, Item * item )
{
    Itemptr  *ptrp;
    int       wait;
    
    if( item->rule->rhslen <= item->dot ) { /*  */
        ptrp = &chart_comp( chart, item->left, item->right );
        item->next_comp = *ptrp;
        *ptrp = item;
    }
    else { /*  */
        wait = item->rule->rhs[ item->dot ];
        
        ptrp = &chart_rwait( chart, wait, item->right );
        item->next_rwait = *ptrp;
        *ptrp = item;
        
        ptrp = &chart_lwait( chart, wait, item->left );
        item->next_lwait = *ptrp;
        *ptrp = item;
    }
}
void chart_additem ( Chart * chart, Item * item )
{
    Itemptr *ptrp;
    
    /* S */
    ptrp = &chart_dot( chart, item->right );
    item->next_dot = *ptrp;
    *ptrp = item;
    
    chart_addhash( chart, item );
}


/*   `[ggĂȂ폜
----------------------------------------------------------------------
*/
Chart * chart_pruning ( Chart * chart )
{
    Itemptr  item, *prev;
    int  deleted = 0;
    int  i;
    
    /* ̃|C^ׂč폜 */
    chart_clear_hash( chart );
    
    /* svȍ̍폜 */
    for( i = 0; i < chart->len; ++i ) {
        prev = &chart_dot( chart, i );
        item = *prev;
        while( item ) {
            if( item->beta == 0.0 ) { /* Om 0  ǂłgȂ */
                *prev = item->next_dot;
                ruby_xfree( item );
                ++deleted;
            }
            else { /* |C^֐i߂ */
                prev = &item->next_dot;
            }
            item = *prev;
        }
    }
    chart->count_deleted = deleted;
    
    /* hbgʒuNHȂnbVɓo^ */
    for( i = 0; i < chart->len; ++i ) {
        for( item = chart_dot( chart, i ); item; item = item->next_dot ) {
            chart_addhash( chart, item );
        }
    }
    
    return chart;
}


/*  `[gw肳ꂽ̂Ɠ
----------------------------------------------------------------------
*/
Item * chart_find ( Chart * chart, Item const * item )
{
    Rule * rule  = item->rule;
    int    left  = item->left;
    int    right = item->right;
    int    dot   = item->dot;
    int    top   = item->top;
    Item * p;
    
    if( dot >= rule->rhslen ) { /*  */
        for( p = chart_comp( chart, left, right ); p; p = p->next_comp )
        {
            /* rule Ȃ dot rKv͂Ȃ */
            if( rule == p->rule && top == p->top )
                return p;
        }
    }
    else { /*  */
        for( p = chart_lwait( chart, rule->rhs[ dot ], left ); p; p = p->next_lwait )
        {
            /* lwait ŌĂ̂ left ͕K */
            if( rule == p->rule && right == p->right && dot == p->dot )
                return p;
        }
    }
    
    return NULL;
}


/* `[g̍
----------------------------------------------------------------------
 - Ȃ΂ȂȂB[ȂH@ꂾł͑ȂBE[鍀͍[ĂȂ΂ȂȂB
*/
Chart * chart_merge ( Chart * dest, Chart * src, int left, int right )
{
    Item  *p, *q;
    int  i, j;
    
    if( dest->grammar != src->grammar || dest->poslist != src->poslist )
        rb_raise( rb_eNLPException, "chart_merge(): Cannot merge different-type charts." );
    
    /* SĂ̍oĒǉ */
    for( i = 0; i < dest->len; ++i ) {
        if( i == right ) {
            for( p = chart_dot( src, i ); p; p = q ) {
                q = p->next_dot;
                if( p->dot >= p->rule->rhslen || p->left == left )
                    chart_additem( dest, p );
                else
                    ruby_xfree( p );
            }
        }
        else {
            for( p = chart_dot( src, i ); p; p = q ) {
                q = p->next_dot;
                chart_additem( dest, p );
            }
        }
    }
    
    /* src ɂ */
    chart_clear_item( src );
    
    return dest;
}


/* e`[gɂ鍀q`[gֈړ
----------------------------------------------------------------------
 - left Ŏw肳ꂽʒuɃhbgړ
 - ꎞIɐe`[g͔j󂳂ꂽ`ɂȂ̂Œ
*/
Chart * chart_move_parentitem ( Chart * chart, Chart * parentchart, int left )
{
    Item  *p, *next;
    int  i;
    
    if( !parentchart )
        return chart;
    
    /* parentchart  chart ֈړ */
    for( p = chart_dot( parentchart, left ); p; p = next ) {
        next = p->next_dot;
        chart_additem( chart, p );
    }
    
    /* parentchart č\z */
    chart_dot( parentchart, left ) = NULL;
    chart_clear_hash( parentchart );
    for( i = 0; i < parentchart->len; ++i ) {
        for( p = chart_dot( parentchart, i ); p; p = p->next_dot )
            chart_addhash( parentchart, p );
    }
    
    return chart;
}


/*  `[g̍̊mp[^
----------------------------------------------------------------------
*/
void chart_reset_probability ( Chart * chart, GenericStolcke * stolcke )
{
    Item  *item;
    int    position;
    
    for( position = 0; position < chart->len; ++position ) {
        for( item = chart_dot( chart, position ); item; item = item->next_dot )
        {
            /* hbgʒu 0 ̂̂Ƀp[^ݒ肷 */
            item->gamma = (item->dot > 0)? 0.0: stolcke->probtbl[ item->rule->id ];
            item->beta  = 0.0;
            item->tag   = 0;
        }
    }
}


/*  ̐m (̑ΐޓx) ߂
----------------------------------------------------------------------
*/
double chart_likelihood_internal ( Chart * chart )
{
    double   sentprob = 0.0;
    Item    *p;
    int      start = NUM2INT( rb_gv_get( "$Start" ) );
    int      itemcount = 0;
    
    for( p = chart_comp( chart, 0, chart->len-1 ); p; p = p->next_comp ) {
        if( p->top != start )
            continue;
        sentprob += p->gamma;
        ++itemcount;
    }
    
    if( itemcount == 0 )
        return Qnil;
    
    if( sentprob == 0.0 )
        rb_raise( rb_eNLPException, "chart_likelihood(): Sentence probability equal zero!!" );
    
    return log( sentprob );
}
static VALUE chart_likelihood ( VALUE self )
{
    Chart  *chart;
    
    Data_Get_Struct( self, Chart, chart );
    
    return rb_float_new( chart_likelihood_internal( chart ) );
}


/*  `[g\̍𓾂
----------------------------------------------------------------------
*/
static VALUE chart_count_item ( VALUE self )
{
    Chart  *chart;
    Item   *item;
    int     i, count;
    
    Data_Get_Struct( self, Chart, chart );
    
    if( chart->count_item == 0 ) {
        count = 0;
        for( i = 0; i < chart->len; ++i ) {
            for( item = chart_dot( chart, i ); item; item = item->next_dot )
                ++count;
        }
        chart->count_item = count;
    }
    
    return INT2NUM(chart->count_item);
}


/*  OmvZɍ폜ꂽ̐𓾂
----------------------------------------------------------------------
*/
static  VALUE  chart_count_deleted  ( VALUE self )
{
    Chart  *chart;
    
    Data_Get_Struct( self, Chart, chart );
    
    return INT2NUM(chart->count_deleted);
}


/*  `[gɓo^悤ƂĎꂽ̐𓾂
----------------------------------------------------------------------
*/
static  VALUE  chart_count_cancelled  ( VALUE self )
{
    Chart  *chart;
    
    Data_Get_Struct( self, Chart, chart );
    
    return INT2NUM(chart->count_cancelled);
}


/* `[gmۂĂ郁̏𓾂
----------------------------------------------------------------------
*/
struct tagChartMemoryInfo
{
    double  total;
    double  tblsize;
    double  itemsize;
};
typedef  struct tagChartMemoryInfo  ChartMemoryInfo;

static ChartMemoryInfo chart_check_memory ( VALUE self )
{
    ChartMemoryInfo  info;
    Chart   *chart;
    Itemptr  p;
    int  length, symbol;
    int  i;
    
    Data_Get_Struct( self, Chart, chart );
    
    info.tblsize = (chart->len*chart->len + 2*chart->len*chart->symbol + chart->len) * sizeof(Itemptr) + (chart->len-1) * sizeof(int);
    info.itemsize = 0;
    for( i = 0; i < chart->len; ++i ) {
        for( p = chart_dot( chart, i ); p; p = p->next_dot ) {
            info.itemsize += sizeof(Item);
        }
    }
    info.total = sizeof(Chart) + info.itemsize + info.tblsize;
    
    return info;
}

static VALUE chart_meminfo ( VALUE self )
{
    ChartMemoryInfo  info;
    char  buffer [1024];
    
    info = chart_check_memory( self );
    
    sprintf( buffer, "%.1f KB (tbl=%.1f;item=%.1f)", info.total/1024, info.tblsize/1024, info.itemsize/1024 );
    
    return rb_str_new2( buffer );
}

static VALUE chart_memsize ( VALUE self )
{
    ChartMemoryInfo  info;
    
    info = chart_check_memory( self );
    
    return INT2NUM( (int)info.total );
}


/*   `[gf[^Ƃďo͂
----------------------------------------------------------------------
*/
static VALUE chart_display ( VALUE self, VALUE out )
{
    Chart  *chart;
    Item   *p;
    int     i;
    
    VALUE volatile str_nl = rb_str_new2( "\n" );
    
    Data_Get_Struct( self, Chart, chart );
    for( i = 0; i < chart->len; ++i ) {
        for( p = chart_dot( chart, i ); p; p = p->next_dot ) {
            rb_io_write( out, item_inspect( p ) );
            rb_io_write( out, str_nl );
        }
    }
    
    return self;
}


/*   fobOpɃ`[g\
----------------------------------------------------------------------
*/
void chart_debug_display ( Chart * chart )
{
    int  i;
    Item  *p;
    
    for( i = 0; i < chart->len; ++i ) {
        for( p = chart_dot( chart, i ); p; p = p->next_dot ) {
            item_display( p );
        }
    }
    fprintf( stderr, "----\n" );
}


/*   \͌ʂ̕\Ɏgpf[^\
----------------------------------------------------------------------
*/
struct  tagExtractData
{
    Chart    *chart;
    Itemptr  *ptr;
    int       cap, len;
};
typedef  struct tagExtractData  ExtractData;

static void extract_data_free ( ExtractData * dat )
{
    if( dat ) {
        ruby_xfree_heaps( 1, dat->ptr );
        ruby_xfree( dat );
    }
}

struct  tagExtractList
{
    struct tagExtractList  *next;
    Item  *item;
};
typedef  struct tagExtractList  ExtractList;


/*   `[g̏
----------------------------------------------------------------------
*/
static void extract_visit_item ( ExtractData * data, Item * p, ExtractList * prev )
{
    Chart  *chart = data->chart;
    
    ExtractList  unit;
    Rule  *r = p->rule;
    Item  *i, *j;
    int    n, wantdot, wantcat, processed;
    int    nonterm = chart->grammar->size_lhshash;
    
    if( p->tag & TAG_Visited )
        return; /* łɖKς݂ȂΖⓚp */
    
    if( p->dot == 0 ) {
        /* ׂ̉͂ďÎŃX^bNɍ̃Xg */
        if( data->cap <= data->len + 1 + r->rhslen ) {
            data->cap += 128 * (1 + r->rhslen);
            REALLOC_N( data->ptr, Itemptr, data->cap );
        }
        data->ptr[ data->len++ ] = p;
        for( n = r->rhslen; n > 0; --n ) {
            data->ptr[ data->len++ ] = prev->item;
            prev = prev->next;
        }
        return;
    }
    
    if( p->dot >= r->rhslen )
        p->tag |= TAG_Visited;  /* ɂ͏dď񂵂Ȃ悤ɔԕu */
    
    unit.next = prev;
    unit.item = p;
    
    wantdot = p->dot - 1;
    wantcat = r->rhs[ wantdot ];
    for( i = chart_lwait( chart, wantcat, p->left ); i; i = i->next_lwait )
    {
        if( i->right >= p->right || i->rule != p->rule || i->dot != wantdot )
            continue;
        /* ̎_ i  p ̖ЂƂƂȂ */
        
        if( wantcat < nonterm ) {
            /* I[L҂Ăꍇ̂ݖLɊւďs */
            processed = 0;
            for( j = chart_comp( chart, i->right, p->right ); j; j = j->next_comp ) {
                if( j->top != wantcat )
                    continue;
                extract_visit_item( data, j, NULL );  /* LɊւď񂷂 */
                ++processed;
            }
            if( processed == 0 )
                continue; /* \zɎgȂ̂̏ꍇ̓LZ */
        }
        else {
            /* I[L҂Ăꍇ͐Ă邩𒲂ׂ */
            if( i->right+1 != p->right )
                continue;
        }
        
        extract_visit_item( data, i, &unit ); /* c񂷂 */
    }
}


/*  \͌ʂ\
----------------------------------------------------------------------
    X^bN̓X^bNq[vɂƂB
    ŐV̕擪ɂȂB
    (l,r) X --> A B * C D
    ƂAl n܂ B ҂B
    AKƒĂ͈͂ŁAhbgʒu 1 Ȃ̂B
    
     static ȃO[oϐgĂ̂ŒӁB
    xOt쐬Ƀ\[ĝō̏͊֌WȂB
    
    obt@̍\
     [0] : X --> * A B ...
     [1] : X --> A * B ...
     ...
*/
static VALUE chart_display_itemset ( VALUE self, VALUE io )
{
    VALUE const volatile str_nl = rb_str_new2( "\n" );
    const int start = NUM2INT( rb_gv_get( "$Start" ) );
    
    ExtractData  *data;
    Grammar      *gram;
    Chart        *chart;
    
    VALUE volatile strobj, lineary, data_obj;
    Item  *item;
    Rule  *rule;
    char  *p, buf[1024];
    int    i, len, pos, offset;
    int   *rulemap;
    
    /* f[^̏ */
    data_obj = Data_Make_Struct( rb_cData, ExtractData, NULL, extract_data_free, data );
    rb_zero_memory( data, ExtractData, 1 );
    
    /* `[g̏ƃX^bN̍\z (data->ptr \z) */
    Data_Get_Struct( self, Chart, chart );
    data->chart = chart;
    for( item = chart_comp( chart, 0, chart->len-1 ); item; item = item->next_comp ) {
        if( item->top != start )
            continue;
        extract_visit_item( data, item, NULL );
    }
    #if DEBUGSW
    for( i = 0; i < data->len; ++i ) /* ͌ʂ̃x݂̂\ */
        fprintf( stderr, "[%d] (%d,%d) %s\n", i, data->ptr[i]->left, data->ptr[i]->right,
                 STR2CSTR(rb_funcall(data->ptr[i]->rule->orig, id_typestring, 1, INT2FIX(data->ptr[i]->dot))) );
    #endif
    
    gram = chart->grammar;
    lineary = rb_ary_new2( (chart->count_item > 0)? chart->count_item: 1024 );
    
    /* oK}bv̏o */
    rulemap = ALLOCA_N( int, gram->size_rule ); /* rulemap: rule-id --> no. */
    rb_zero_memory( rulemap, int, gram->size_rule );
    len = 1;
    for( offset = 0; offset < data->len; ++offset ) {
        i = data->ptr[offset]->rule->id;
        if( rulemap[i] == 0 )
            rulemap[i] = len++;
    }
    for( i = 0; i < gram->size_rule; ++i ) {
        if( rulemap[i] == 0 )
            continue;
        sprintf( buf, "#%d %s", rulemap[i], STR2CSTR(rb_funcall(gram->rules[i].orig, id_typestring, 0)) );
        strobj = rb_str_new2( buf );
        rb_ary_push( lineary, strobj );
    }
    
    /* obt@tɏoĂ */
    offset = data->len - 1;
    while( offset >= 0 ) {
        /* ŏ̃ItZbg̓hbg̍ (A --> X Y Z * Ȃ) */
        rule = data->ptr[offset]->rule;
        len  = rule->rhslen;
        offset -= len; /* hbg擪̍ (A --> * X Y Z Ȃ) ɃItZbg킹 */
        
        pos = data->ptr[offset]->left;
        p = buf;
        sprintf( p, "[%d", pos ); p += strlen( p );
        for( i = 1; i <= len; ++i ) {
            item = data->ptr[ offset + i ];
            if( item->rule != rule )
                rb_raise( rb_eNLPException, "chart_display_itemset(): Stack is broken." );
            if( item->right <= pos )
                rb_raise( rb_eNLPException, "chart_display_itemset(): Position list is broken." );
            pos = item->right;
            sprintf( p, ",%d", pos ); p += strlen( p );
        }
        sprintf( p, "] %d", rulemap[rule->id] );
        strobj = rb_str_new2( buf );
        rb_ary_push( lineary, strobj );
        
        /* ̃ubNɈڂ */
        --offset;
    }
    rb_ary_push( lineary, Qnil );
    strobj = rb_ary_join( lineary, str_nl );
    rb_io_write( io, strobj );
    
    return Qtrue;
}


/*  Chart NX̍\z
----------------------------------------------------------------------
*/
void init_chart ( void )
{
    /* lCeBu `[gNX̒` */
    rb_cChart = rb_define_class_under( rb_mExt, "Chart", rb_cData );
    
    /* \bh */
    rb_define_methods(
        rb_cChart,
        
        "likelihood"		, chart_likelihood, 0,

        "count_item"		, chart_count_item, 0,
        
        "count_deleted"		, chart_count_deleted, 0,
        
        "count_cancelled"	, chart_count_cancelled, 0,
        
        "meminfo"		, chart_meminfo, 0,
        
        "memsize"		, chart_memsize, 0,
        
        "display"		, chart_display, 1,
        
        "display_itemset"	, chart_display_itemset, 1,
        
        NULL );
}


