/*                                                     project Gemini 
----------------------------------------------------------------------
    graphical EM algorithm
      $Id: gem.c,v 1.12 2002/02/08 04:33:55 Black Exp $

        Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp> 
----------------------------------------------------------------------
*/
#define DEBUGSW 0
#include "nlputil.h"


/*   IuWFNg̍\z
----------------------------------------------------------------------
*/
static void gem_free ( GraphicalEM * gem )
{
    if( gem ) {
        rb_gc_unregister_address( &gem->obj_cache );
        ruby_xfree_heaps( 2, gem->probtbl, gem->xnumtbl );
        ruby_xfree( gem );
    }
}

static VALUE gem_new ( VALUE klass, VALUE gramobj )
{
    GraphicalEM  *gem;
    VALUE volatile obj;
    
    ru_is_instance_of( gramobj, rb_cGrammar );
    
    obj = Data_Make_Struct( klass, GraphicalEM, NULL, gem_free, gem );
    rb_zero_memory( gem, GraphicalEM, 1 );
    gem->sgary = Qnil;
    
    rb_obj_call_init( obj, 1, &gramobj );
    
    return obj;
}

static VALUE gem_initialize ( VALUE self, VALUE gramobj )
{
    GraphicalEM  *gem;
    int  rulesize;
    
    Data_Get_Struct( self, GraphicalEM, gem );
    gem->loopcount = 1;
    
    Data_Get_Struct( gramobj, Grammar, gem->grammar );
    rulesize = gem->grammar->size_rule;
    gem->probtbl = ALLOC_N( real_t, rulesize );
    gem->xnumtbl = ALLOC_N( real_t, rulesize );
    gem->sgary = rb_ary_new();
    
    rb_zero_memory( gem->probtbl, real_t, rulesize );
    rb_zero_memory( gem->xnumtbl, real_t, rulesize );
    gem->probtbl[0] = 1.0;
    
    gem->obj_cache = rb_ary_new3( 2, gem->grammar->orig, gem->sgary );
    rb_gc_register_address( &gem->obj_cache );
    
    return self;
}


/*   xOt̒ǉ
----------------------------------------------------------------------
*/
static VALUE gem_store ( VALUE self, VALUE sgobj )
{
    GraphicalEM   *gem;
    SupportGraph  *sg;
    
    Data_Get_Struct( self, GraphicalEM, gem );
    
    /* `FbN */
    ru_is_instance_of( sgobj, rb_cSupportGraph );
    Data_Get_Struct( sgobj, SupportGraph, sg );
    /* $$ `FbN $$ */
    
    sgraph_compile( sg, gem->grammar->size_rule, gem->probtbl, gem->xnumtbl );
    rb_ary_push( gem->sgary, sgobj );
    
    return self;
}


/*   xOtXg̎擾
----------------------------------------------------------------------
*/
static VALUE gem_sglist ( VALUE self )
{
    GraphicalEM  *gem;
    
    Data_Get_Struct( self, GraphicalEM, gem );
    
    return rb_obj_dup( gem->sgary );
}


/*   xOtXg̃TCY擾
----------------------------------------------------------------------
*/
static VALUE gem_size ( VALUE self )
{
    GraphicalEM  *gem;
    
    Data_Get_Struct( self, GraphicalEM, gem );
    
    return rb_funcall( gem->sgary, id_size, 0 );
}


/*   p[^XV
----------------------------------------------------------------------
*/
static real_t learn_param ( GraphicalEM * gem, SupportGraph * sgraph, int num )
{
    real_t     w, sum, sentprob;
    SG_Node   *node;
    SG_Path   *path;
    SG_Graph  *graph;
    int  i, j, k;
    
    /*  Inside
    ---------------------------------------- */
    graph = &sgraph->buffer_graph[ sgraph->size_graph-1 ];
    for( i = sgraph->size_graph; i > 0; --i )
    {
        sum = 0.0;
        path = graph->pathary;
        for( j = graph->pathlen; j > 0; --j )
        {
            w = 1.0;
            node = path->nodeary;
            for( k = path->nodelen; k > 0; --k )
            {
                switch( node->d.ind.type ) {
                case 2:
                    w *= node->d.ind.data.graph->inside;
                    break;
                case 1:
                    w *= gem->probtbl[ node->d.ind.data.sw ];
                    break;
                default:
                    rb_raise( rb_eNLPException, "ext/nlputil/gem.c: learn_param(): Support graph is broken." );
                }
                ++node;
            }
            sum += w;
            path->inside = w;
            ++path;
        }
        graph->inside = sum;
        --graph;
    }
    sentprob = sgraph->buffer_graph[0].inside;
    
    if( sentprob == 0.0 )
        rb_raise( rb_eRuntimeError, "Sentence[%d]'s probability is zero.", num );
    
    /*  Outside
    ---------------------------------------- */
    graph = sgraph->buffer_graph;
    for( i = 0; i < sgraph->size_graph; ++i ) {
        graph->outside = 0;
        ++graph;
    }
    sgraph->buffer_graph[0].outside = 1.0;

    graph = sgraph->buffer_graph;
    for( i = sgraph->size_graph; i > 0; --i )
    {
        path = graph->pathary;
        for( j = graph->pathlen; j > 0; --j )
        {
            sum = graph->outside * path->inside;
            node = path->nodeary;
            for( k = path->nodelen; k > 0; --k )
            {
                if( node->d.ind.type == 1 )
                    gem->xnumtbl[ node->d.ind.data.sw ] += sum / sentprob;
                else {
                    if( node->d.ind.data.graph->inside > 0 )
                        node->d.ind.data.graph->outside += sum / node->d.ind.data.graph->inside;
                }
                ++node;
            }
            ++path;
        }
        ++graph;
    }

    return log( sentprob );
}

static real_t learn_param_fast ( GraphicalEM * gem, SupportGraph * sgraph, int num )
{
    real_t     w, sum, sentprob;
    SG_Node   *node;
    SG_Path   *path;
    SG_Graph  *graph;
    int  i, j, k;
    
    /*  Inside */
    graph = &sgraph->buffer_graph[ sgraph->size_graph-1 ];
    for( i = sgraph->size_graph; i > 0; --i ) {
        sum = 0.0;
        path = graph->pathary;
        for( j = graph->pathlen; j > 0; --j ) {
            w = 1.0;
            node = path->nodeary;
            for( k = path->nodelen; k > 0; --k ) {
                w *= *node->d.dir.i_ptr;
                ++node;
            }
            sum += w;
            path->inside = w;
            ++path;
        }
        graph->inside = sum;
        --graph;
    }
    sentprob = sgraph->buffer_graph[0].inside;
    
    if( sentprob == 0.0 )
        rb_raise( rb_eRuntimeError, "Sentence[%d]'s probability is zero.", num );
    
    /*  Outside */
    graph = sgraph->buffer_graph;
    for( i = 0; i < sgraph->size_graph; ++i ) {
        graph->outside = 0;
        ++graph;
    }
    sgraph->buffer_graph[0].outside = 1.0;

    graph = sgraph->buffer_graph;
    for( i = sgraph->size_graph; i > 0; --i ) {
        path = graph->pathary;
        for( j = graph->pathlen; j > 0; --j ) {
            sum = graph->outside * path->inside;
            node = path->nodeary;
            for( k = path->nodelen; k > 0; --k ) {
                if( *node->d.dir.o_inside > 0 )
                    *node->d.dir.o_ptr += sum / *node->d.dir.o_inside;
                ++node;
            }
            ++path;
        }
        ++graph;
    }

    return log( sentprob );
}

static VALUE gem_update ( VALUE self )
{
    const ID  id_get_pr = rb_intern( "pr" );
    const ID  id_set_pr = rb_intern( "pr=" );
    
    SupportGraph  *sgraph;
    GraphicalEM   *gem;
    RArrayPtr      sgary;
    Grammar       *gram;
    
    real_t   sum, likelihood;
    VALUE    prob;
    Rule    *r;
    int      i, loopcount;
    
    Data_Get_Struct( self, GraphicalEM, gem );
    gram = gem->grammar;
    
    /* p[^̏
    ---------------------------------------- */
    gem->probtbl[ 0 ] = 0.0;
    for( i = 1; i < gram->size_rule; ++i ) {
        prob = rb_funcall( gram->rules[i].orig, id_get_pr, 0 );
        if( NIL_P( prob ) )
            rb_raise( rb_eRuntimeError, "ERROR: probability of rule[ %d ] is not assigned", i );
        gem->probtbl[i] = NUM2DBL( rb_Float( prob ) );
    }
    
    /* p[^
    ---------------------------------------- */
    loopcount = (gem->loopcount <= 0)? 1: gem->loopcount;
    for( ; loopcount > 0; --loopcount ) {
        rb_zero_memory( gem->xnumtbl, real_t, gram->size_rule );
        
        sgary = RARRAY( gem->sgary );
        likelihood = 0.0;
        if( rb_block_given_p() ) {
            for( i = 0; i < sgary->len; ++i ) {
                rb_yield( sgary->ptr[i] );
                Data_Get_Struct( sgary->ptr[i], SupportGraph, sgraph );
                if( sgraph->probtbl )
                    likelihood += learn_param_fast( gem, sgraph, i+1 );
                else
                    likelihood += learn_param( gem, sgraph, i+1 );
            }
        }
        else {
            for( i = 0; i < sgary->len; ++i ) {
                Data_Get_Struct( sgary->ptr[i], SupportGraph, sgraph );
                if( sgraph->probtbl )
                    likelihood += learn_param_fast( gem, sgraph, i+1 );
                else
                    likelihood += learn_param( gem, sgraph, i+1 );
            }
        }
        
        /* p[^̍XV
        ---------------------------------------- */
        for( i = 0; i < gram->size_lhshash; ++i ) {
            sum = 0.0;
            for( r = gram->hash_lhs[ i ]; r; r = r->next_lhs ) {
                sum += gem->xnumtbl[ r->id ];
            }
            if( sum > 0.0 ) {
                for( r = gram->hash_lhs[ i ]; r; r = r->next_lhs ) {
                    gem->probtbl[ r->id ] = gem->xnumtbl[ r->id ] / sum;
                    /* $$$ KvȂH $$$
                    if( gem->xnumtbl[r->id] > 0 && gem->probtbl[r->id] == 0.0 )
                        rb_raise( rb_eNLPException, "Warning: Precision is lost." );
                    */
                }
            }
        }
    }
    
    /* p[^̐ݒ
    ---------------------------------------- */
    for( i = 1; i < gram->size_rule; ++i )
        rb_funcall( gram->rules[i].orig, id_set_pr, 1, rb_float_new( (double)gem->probtbl[i] ) );
    
    return rb_float_new( likelihood );
}


/*   x`}[Np̃p[^XV
----------------------------------------------------------------------
*/
static VALUE gem_bench_update ( VALUE self )
{
    const ID  id_get_pr = rb_intern( "pr" );
    
    SupportGraph  *sgraph;
    GraphicalEM   *gem;
    RArrayPtr      sgary;
    Grammar       *gram;
    
    VALUE   prob;
    real_t  sum, likelihood;
    int     i, loopcount;
    
    Data_Get_Struct( self, GraphicalEM, gem );
    gram = gem->grammar;
    
    /* p[^̏
    ---------------------------------------- */
    gem->probtbl[ 0 ] = 0.0;
    for( i = 1; i < gram->size_rule; ++i ) {
        prob = rb_funcall( gram->rules[i].orig, id_get_pr, 0 );
        if( NIL_P( prob ) )
            rb_raise( rb_eRuntimeError, "ERROR: probability of rule[ %d ] is not assigned", i );
        sum = NUM2DBL( rb_Float( prob ) );;
        gem->probtbl[i] = sum;
    }
    
    /* p[^
    ---------------------------------------- */
    loopcount = (gem->loopcount <= 0)? 1: gem->loopcount;
    for( ; loopcount > 0; --loopcount ) {
        rb_zero_memory( gem->xnumtbl, real_t, gram->size_rule );
        sgary = RARRAY( gem->sgary );
        likelihood = 0.0;
        for( i = 0; i < sgary->len; ++i ) {
            Data_Get_Struct( sgary->ptr[i], SupportGraph, sgraph );
            if( sgraph->probtbl )
                likelihood += learn_param_fast( gem, sgraph, i+1 );
            else
                likelihood += learn_param( gem, sgraph, i+1 );
        }
    }
    
    return INT2FIX( 0 );
}


/*   ftHg̃[v񐔂ݒE擾
----------------------------------------------------------------------
*/
static VALUE gem_set_loopcount ( VALUE self, VALUE count )
{
    GraphicalEM  *gem;
    
    Data_Get_Struct( self, GraphicalEM, gem );
    gem->loopcount = NUM2INT(count);
    
    return count;
}

static VALUE gem_get_loopcount ( VALUE self )
{
    GraphicalEM  *gem;
    
    Data_Get_Struct( self, GraphicalEM, gem );
    
    return INT2FIX(gem->loopcount);
}


/*   Ғl̃Xg𓾂
----------------------------------------------------------------------
*/
static VALUE gem_xnumtbl ( VALUE self )
{
    VALUE volatile ret;
    GraphicalEM  *gem;
    Grammar  *gram;
    int  i;
    
    Data_Get_Struct( self, GraphicalEM, gem );
    gram = gem->grammar;
    
    ret = rb_ary_new2( gram->size_rule );
    for( i = 1; i < gram->size_rule; ++i ) {
        rb_ary_push( ret, rb_float_new( gem->xnumtbl[i] ) );
    }
    
    return ret;
}


/*   OtBJ EM NX̍\z
----------------------------------------------------------------------
*/
void init_gem ( void )
{
    VALUE  rb_cBench;
    
    /* xOt̃NX` */
    rb_cGraphicalEM = rb_define_class_under( rb_mExt, "GraphicalEM", rb_cData );
    rb_cBench       = rb_define_class_under( rb_cGraphicalEM, "BenchMark", rb_cGraphicalEM );
    
    /* NX\bh */
    rb_define_singleton_methods(
        rb_cGraphicalEM,
        
        "new"			, gem_new, 1,
        
        NULL );
    
    /* \bh */
    rb_define_methods(
        rb_cGraphicalEM,
        
        "initialize"		, gem_initialize, 1,
        
        "store"			, gem_store, 1,
        
        "sglist"		, gem_sglist, 0,
        
        "update"		, gem_update, 0,
        
        "loopcount="		, gem_set_loopcount, 1,
        
        "loopcount"		, gem_get_loopcount, 0,
        
        "size"			, gem_size, 0,
        
        "xnumtbl"		, gem_xnumtbl, 0,
        
        NULL );
    
    rb_define_methods(
        rb_cBench,
        
        "update"		, gem_bench_update, 0,
        
        NULL );
    
    /* GCAX */
    rb_define_aliases(
        rb_cGraphicalEM,
        
        "<<"   , "store",
        
        "to_a" , "sglist",
        
        NULL );
    
}


