/*                                                     project Gemini 
----------------------------------------------------------------------
    @NX
      $Id: grammar.c,v 1.5 2002/01/20 05:10:01 Black Exp $

        Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp> 
----------------------------------------------------------------------
*/
#include "nlputil.h"
#include <time.h>


/*   @̏
----------------------------------------------------------------------
*/
static void cfg_free ( Grammar * g )
{
    if( g ) {
        rb_gc_unregister_address( &g->object_cache );
        ruby_xfree_heaps( 4, g->buffer_rhs, g->hash_lc, g->hash_lhs, g->rules );
        ruby_xfree( g );
    }
}

static VALUE cfg_new ( VALUE klass, VALUE origrule )
{
    VALUE volatile obj;
    Grammar  * gram;
    
    gram = rb_alloc_zero( Grammar, 1 );
    gram->orig = Qnil;
    
    obj = Data_Wrap_Struct( klass, NULL, cfg_free, gram );
    rb_obj_call_init( obj, 1, &origrule );
    
    return obj;
}

static void init_alloc ( Grammar * gram, VALUE origgram )
{
    int  i;
    
    gram->size_rule    = ru_getnum( origgram, "size_rule" );
    gram->size_lhshash = ru_getnum( origgram, "size_nonterm" );
    gram->size_lchash  = ru_getnum( origgram, "size_symbol" );
    
    gram->rules = ALLOC_N( Rule, gram->size_rule );
    for( i = 0; i < gram->size_rule; ++i )
        gram->rules[i].orig = Qnil;
    
    gram->hash_lhs = ALLOC_N( RulePtr, gram->size_lhshash );
    for( i = 0; i < gram->size_lhshash; ++i )
        gram->hash_lhs[i] = NULL;
    
    gram->hash_lc = ALLOC_N( RulePtr, gram->size_lchash );
    for( i = 0; i < gram->size_lchash; ++i )
        gram->hash_lc[i] = NULL;
}

static void init_measure ( Grammar * gram, VALUE origgram )
{
    RArrayPtr  ruleary;
    int  size = 0;
    int  i;
    
    ruleary = RARRAY( rb_Array( origgram ) );
    for( i = 1; i < ruleary->len; ++i )
        size += ru_length( ru_rhs( ruleary->ptr[i] ) );
    
    gram->buffer_rhs_size = size;
    gram->buffer_rhs      = ALLOC_N( int, size );
}

static void init_set ( Grammar * gram, VALUE origgram )
{
    RArrayPtr  ruleary;
    VALUE  ary;
    Rule  *r;
    int    i, j, rhsbufsize, *p;
    
    ruleary = RARRAY( rb_Array( origgram ) );
    Check_Type( ruleary, T_ARRAY );
    
    /* ID=0 ɊւĂ͓ʈ */
    r = &gram->rules[0];
    r->lhs     =  0;
    r->rhs     =  NULL;
    r->rhslen  =  0;
    
    rhsbufsize = 0;
    for( i = 1; i < ruleary->len; ++i ) {
        r = &gram->rules[i];
        r->id = i;
        r->next_lc = r->next_lhs = NULL;
        
        /* KIuWFNgEĂ */
        r->orig = ruleary->ptr[i];
        if( NIL_P( r->orig ) )
            rb_raise( rb_eNLPException, "Rule-ID [%d] is nil.", i );
        
        /* ӂ̐ݒ */
        r->lhs = ru_id( ru_lhs( r->orig ) ); /* lhs.id */
        
        /* Eӂ̐ݒ */
        ary = ru_rhs( r->orig );
        r->rhslen = ru_length( ary );
        if( r->rhslen == 0 )
            r->rhs = NULL;
        else {
            /* obt@EӒ̗̈Ă */
            r->rhs = &gram->buffer_rhs[ rhsbufsize ];
            p = r->rhs;
            for( j = 0; j < r->rhslen; ++j )
                p[j] = ru_id( ru_at(ary, j) ); /* rhs[j].id */
            rhsbufsize += r->rhslen;
        }
    }
}

static void init_rehash ( Grammar * gram )
{
    Rule  *r;
    int    i, j;
    
    for( i = 1; i < gram->size_rule; ++i ) {
        r = &gram->rules[i];
        
        j = r->lhs;
        if( gram->hash_lhs[j] )
            r->next_lhs = gram->hash_lhs[j];
        gram->hash_lhs[j] = r;
        
        if( r->rhslen > 0 ) {
            j = r->rhs[0];
            if( gram->hash_lc[j] )
                r->next_lc = gram->hash_lc[j];
            gram->hash_lc[j] = r;
        }
    }
}

static VALUE cfg_initialize ( VALUE self, VALUE origgram )
{
    Grammar  *gram;
    int  i;
    
    Data_Get_Struct( self, Grammar, gram );
    gram->orig = Qnil;
    
    /* f[^̈̊m */
    init_alloc( gram, origgram );
    
    /* @̒ */
    init_measure( gram, origgram );
    
    /* @IuWFNg̐ݒ */
    init_set( gram, origgram );
    
    /* nbV̍쐬 */
    init_rehash( gram );
    
    /* }[NpXg̍쐬 */
    gram->orig = origgram;
    gram->object_cache = rb_ary_new2( gram->size_rule );
    rb_ary_push( gram->object_cache, gram->orig );
    for( i = 1; i < gram->size_rule; ++i )
        rb_ary_push( gram->object_cache, gram->rules[i].orig );
    rb_gc_register_address( &gram->object_cache );
    
    return self;
}



/*  L擾
---------------------------------------------------------------------- */
VALUE cfg_get_catname ( Grammar * g, int id )
{
    VALUE gramobj, cat;
    
    if( !g ) { /* @w肳ĂȂꍇ͕Ŵ̂T */
        gramobj = rb_gv_get( "$grammar" );
        if( rb_obj_is_instance_of( gramobj, rb_cGrammar ) ) {
            Data_Get_Struct( gramobj, Grammar, g );
            gramobj = g->orig;
        }
    }
    else {
        gramobj = g->orig;
    }
    
    cat = Qnil;
    if( !NIL_P( gramobj ) ) {
        cat = rb_funcall( gramobj, rb_intern( "getcat" ), 1, INT2FIX(id) );
    }
    if( NIL_P( cat ) )
        cat = INT2FIX( id );
    
    return rb_String( cat );
}


/*   PCFG mF
----------------------------------------------------------------------
*/
static VALUE cfg_is_pcfg ( VALUE self )
{
    const double  Tolerance = 0.000001;
    const ID      id_prob = rb_intern( "pr" );
    
    Grammar  *g;
    RulePtr  r;
    double   sum;
    VALUE    prob;
    char     buf[ 1024 ];
    int      i;
    
    Data_Get_Struct( self, Grammar, g );
    
    for( i = 0; i < g->size_lhshash; ++i ) {
        r = g->hash_lhs[ i ];
        if( !r )
            continue;
        
        sum = 0.0;
        for( ; r; r = r->next_lhs ) {
            prob = rb_funcall( r->orig, id_prob, 0 );
            if( NIL_P( prob ) ) {
                sum = 0.0;
                break;
            }
            sum += NUM2DBL( rb_Float( prob ) );
        }
        if( !(sum + Tolerance > 1.0  &&  sum - Tolerance < 1.0) ) {
            sprintf( buf, "Sum of rule's propabilities is not 1. (LHS=%s)", STR2CSTR(cfg_get_catname( g, i )) );
            rb_gv_set( "$!", rb_exc_new2( rb_eNLPException, buf ) );
            return Qfalse;
        }
    }
    return Qtrue;
}


/*  _ PCFG p[^蓖Ă
----------------------------------------------------------------------
*/
static  VALUE  cfg_set_random_probability  ( VALUE self )
{
    const int  SeedRange = 100;
    
    ID  id_sprob = rb_intern( "pr=" );
    ID  id_rprob = rb_intern( "pr" );
    
    Grammar  *g;
    RulePtr  rp;
    double   prob;
    time_t   t;
    int      i, j, seed, sum;
    
    Data_Get_Struct( self, Grammar, g );
    srand( (unsigned)time( &t ) );
    
    for( i = 0; i < g->size_lhshash; ++i ) {
        sum = 0;
        for( rp = g->hash_lhs[ i ]; rp; rp = rp->next_lhs ) {
            sum += (seed = 1 + (rand() % SeedRange));
            rb_funcall( rp->orig, id_sprob, 1, INT2FIX( seed ) );
        }
        for( rp = g->hash_lhs[ i ]; rp; rp = rp->next_lhs ) {
            seed = FIX2INT( rb_funcall( rp->orig, id_rprob, 0 ) );
            rb_funcall( rp->orig, id_sprob, 1, rb_float_new( (double)seed / sum ) );
        }
    }
    
    return self;
}


/*   p[^𒲂ׂ
----------------------------------------------------------------------
*/
static VALUE cfg_size_rule ( VALUE self )
{
    Grammar  *g;
    
    Data_Get_Struct( self, Grammar, g );
    
    return INT2FIX( g->size_rule );
}
static VALUE cfg_size_nonterm ( VALUE self )
{
    Grammar  *g;
    
    Data_Get_Struct( self, Grammar, g );
    
    return INT2FIX( g->size_lhshash );
}
static VALUE cfg_size_symbol ( VALUE self )
{
    Grammar  *g;
    
    Data_Get_Struct( self, Grammar, g );
    
    return INT2FIX( g->size_lchash );
}


/*  Ruby ̕@IuWFNg𓾂
----------------------------------------------------------------------
*/
static VALUE cfg_grammar ( VALUE self )
{
    Grammar  *gram;
    
    Data_Get_Struct( self, Grammar, gram );
    
    return  gram->orig;
}


/*  FIRST W Ruby ̕@IuWFNg擾
----------------------------------------------------------------------
*/
static VALUE cfg_FIRST ( VALUE self )
{
    return rb_funcall( cfg_grammar( self ), rb_intern( "FIRST" ), 0 );
}


/*  W[gpĂ郁̃TCY𓾂
----------------------------------------------------------------------
*/
static VALUE cfg_meminfo ( VALUE self )
{
    Grammar *gram;
    char buffer [ 256 ];
    double rulesize, hashsize, total;
    
    Data_Get_Struct( self, Grammar, gram );
    
    rulesize = (sizeof(Rule) * gram->size_rule)
             + (sizeof(int) * gram->buffer_rhs_size);
    hashsize = (sizeof(RulePtr) * (gram->size_lhshash + gram->size_lchash));
    total    = rulesize + hashsize + sizeof(Grammar);
    
    rulesize /= 1024;
    hashsize /= 1024;
    total    /= 1024;
    
    sprintf( buffer, "%.1f KB (rule=%.1f;hash=%.1f)", total, rulesize, hashsize );
    return rb_str_new2( buffer );
}


/* FIRST e[u
----------------------------------------------------------------------
*/
void first_tbl_free ( int * *ptr )
{
    int * *p;
    
    if( ptr )
    {
        for( p = ptr; *p; ++p )
            ruby_xfree( *p );
        ruby_xfree( ptr );
    }
}


/* FIRST e[u𓾂
----------------------------------------------------------------------
쐬Ɏgp Ruby f[^̃e[u̓nbVz肵ĂB
to_a \bhȂǂŔzɕϊꂽ̌`A
  [ [L, [LXg...]], [L, [LXg...]], ... ]
ƂȂĂȂ΂ȂȂB
ŁAL String łȂ΂ȂȂBRuby ł̕@IuWFNg
猟ꂽʂgpB
܂ALXg Category IuWFNg̔zɂ邱ƁB
*/
int **  first_tbl_get  ( Grammar * gram )
{
    const ID id_getcat = rb_intern( "getcat" );
    VALUE const orig = gram->orig;
    
    RArrayPtr  datary, ary;
    VALUE  tbl, key, catname;
    int  **ret;
    int    i, j;
    int    nonterm, symbols;
    int    keycat, datcat;
    
    tbl = rb_funcall( orig, rb_intern("FIRST"), 0 );
    if( NIL_P( tbl ) )
        return NULL;
    
    nonterm = gram->size_lhshash;
    symbols = gram->size_lchash;
    
    /* e[ů蓖
    ---------------------------------------- */
    ret = ALLOC_N( int *, nonterm + 1 );
    for( i = 0; i < nonterm; ++i ) {
        ret[i] = ALLOC_N( int, symbols );
        for( j = 0; j < symbols; ++j )
            ret[i][j] = 0;
    }
    ret[ nonterm ] = NULL; /* e[uɒ񂪖ߔԕő */
    
    
    /* e[u̍쐬
    ---------------------------------------- */
    ary = RARRAY( rb_Array( tbl ) );
    for( i = 0; i < ary->len; ++i )
    {
        catname = rb_String( ru_first( ary->ptr[i] ) );
        
        key = rb_funcall( orig, id_getcat, 1, catname );
        if( NIL_P( key ) ) {
            first_tbl_free( ret );
            rb_raise( rb_eNLPException, "`%s' is unknown PoS.", STR2CSTR(catname) );
        }
        
        keycat = ru_id(key);
        datary = RARRAY( rb_Array( ru_last( ary->ptr[i] ) ) );
        
        for( j = 0; j < datary->len; ++j )
            ret[ keycat ][ ru_id(datary->ptr[j]) ] = 1;
    }
    
    return ret;
}


/*  ̊gW[̏R[h
----------------------------------------------------------------------
*/
void init_grammar ( void )
{
    /* lCeBu Rule NX̒`
    ---------------------------------------- */
    rb_cGrammar = rb_define_class_under( rb_mExt, "Grammar", rb_cData );
    
    /* NX\bh
    ---------------------------------------- */
    rb_define_singleton_methods(
        rb_cGrammar,
        
        "new"				, cfg_new, 1,
        
        NULL );
    
    /* \bh
    ---------------------------------------- */
    rb_define_methods(
        rb_cGrammar,
        
        "initialize"			, cfg_initialize, 1,
        
        "FIRST"				, cfg_FIRST, 0,
        
        "isPCFG?"			, cfg_is_pcfg, 0,
        
        "setRandomProbability"		, cfg_set_random_probability, 0,
        
        "size_rule"			, cfg_size_rule, 0,
        
        "size_nonterm"			, cfg_size_nonterm, 0,
        
        "size_symbol"			, cfg_size_symbol, 0,
        
        "grammar"			, cfg_grammar, 0,
        
        "meminfo"			, cfg_meminfo, 0,
        
        NULL );
}


