/*                                                     project Gemini 
----------------------------------------------------------------------
    \͌ʑp[eBeB
      $Id: itemset.c,v 1.7 2002/01/22 06:10:17 Black Exp $

        Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp> 
----------------------------------------------------------------------
*/
#define DEBUGSW 0
#include "nlputil.h"

/* ł֐̓}NŒ`邩ǂ */
#define USE_MACRO 1


/* ItemSet ̓e\
----------------------------------------------------------------------
*/
static VALUE itemset_display ( VALUE self, VALUE out )
{
    ItemSet     *itemset;
    ParsedItem  *item, *child;
    const char * catname;
    VALUE  rule;
    char   buf[ 1024 ];
    int    i, j;
    
    Data_Get_Struct( self, ItemSet, itemset );
    for( i = 0; i < itemset->itembuf_size; ++i ) {
        item = &itemset->itembuf[i];
        rule = item->rule;
        
        catname = STR2CSTR( rb_String( ru_lhs(rule) ) );
        sprintf( buf, "%s(%d,%d) : %s\n", catname, item->left, item->right, STR2CSTR(rb_funcall(rule, id_typestring, 0)) );
        rb_io_write( out, rb_str_new2( buf ) );
        
        for( j = 0; j < item->child_len; ++j ) {
            child = item->child[j];
            if( !child )
                continue;
            catname = STR2CSTR( rb_String( ru_lhs(child->rule) ) );
            sprintf( buf, " [%d] ==> %s(%d,%d)\n", j+1, catname, child->left, child->right );
            rb_io_write( out, rb_str_new2( buf ) );
        }
    }
    rb_io_write( out, rb_str_new2( "****************************************\n" ) );
    for( i = 0; i < itemset->headbuf_size; ++i ) {
        item = itemset->headbuf[i];
        catname = STR2CSTR( rb_String( ru_lhs( item->rule ) ) );
        sprintf( buf, "%s(%d,%d)\n", catname, item->left, item->right );
        rb_io_write( out, rb_str_new2( buf ) );
    }
    
    return self;
}


/* RXgN^EfXgN^
----------------------------------------------------------------------
*/
static void itemset_free ( ItemSet * is )
{
    if( is ) {
        rb_gc_unregister_address( &is->obj_cache );
        ruby_xfree_heaps( 3, is->headbuf, is->childbuf, is->itembuf );
        ruby_xfree( is );
    }
}

static VALUE itemset_new ( VALUE klass, VALUE itemsetobj )
{
    ItemSet  *itemset;
    VALUE volatile obj;
    
    itemset = rb_alloc_zero( ItemSet, 1 );
    itemset->rulemap = Qnil;
    
    obj = Data_Wrap_Struct( klass, NULL, itemset_free, itemset );
    rb_obj_call_init( obj, 1, &itemsetobj );
    
    return obj;
}

static int count_child_size ( RArrayPtr itemlist )
{
    RArrayPtr  posary;
    int  i;
    int  count = 0;
    
    for( i = 0; i < itemlist->len; ++i ) {
        posary = RARRAY( rb_Array(itemlist->ptr[i]) );
        count += posary->len;
    }
    
    return count;
}

static void create_hash ( ItemSet * itemset, ParsedItemptr * hash )
{
    /* hash ͍ŏIIɔ͈݂͂̂ɂnbVƂȂgbvL̂͂ЂƂ */
    ParsedItemptr  ptr;
    ParsedItem    *item;
    int  i, addr;
    
    rb_zero_memory( hash, ParsedItemptr, itemset->sent_len * itemset->sent_len );
    
    for( i = 0; i < itemset->itembuf_size; ++i ) {
        item = &itemset->itembuf[ i ];
        addr = item->left + item->right * itemset->sent_len;
        
        /* hash[addr]̃Xgɓgbv̍΂̉ɗƂ */
        for( ptr = hash[addr]; ptr; ptr = ptr->next_hash ) {
            if( ptr->top == item->top )
                break; /* ptr gbv̐擪 */
        }
        if( ptr ) {
            item->next_hash = NULL;
            item->next_item = ptr->next_item;
            ptr->next_item = item;
        }
        else {
            item->next_hash = hash[addr];
            hash[addr] = item;
            item->next_item = NULL;
        }
    }
}

static ParsedItemptr find_item ( ItemSet * itemset, ParsedItemptr * hash, int left, int right, int top )
{
    ParsedItemptr  ptr;
    
    for( ptr = hash[ left + right*itemset->sent_len ]; ptr; ptr = ptr->next_hash ) {
        if( ptr->top == top )
            return ptr;
    }
    
    rb_raise( rb_eNLPException, "find_item(): Item not found. (#%d(%d,%d))", top, left, right );
    return NULL; /* _~[ */
}

static void display_hash ( ItemSet * itemset, ParsedItemptr * hash )
{
    ParsedItem  *item, *ptr;
    int  left, right, i;
    
    for( right = 0; right < itemset->sent_len; ++right ) {
        for( left = 0; left < itemset->sent_len; ++left ) {
            for( item = hash[ left + right * itemset->sent_len ]; item; item = item->next_hash ) {
                fprintf( stderr, "(%d,%d) %s", item->left, item->right, STR2CSTR(rb_String(ru_lhs(item->rule))) );
                i = 0;
                for( ptr = item; ptr; ptr = ptr->next_item )
                    ++i;
                fprintf( stderr, " - %d\n", i );
            }
        }
    }
    fprintf( stderr, "--------\n" );
}

static void set_item ( ItemSet * itemset, RArrayPtr itemlist, VALUE startitem )
{
    ID const id_nonterm_p = rb_intern("nonterm?");
    ID const id_right     = rb_intern("right");
    ID const id_left      = rb_intern("left");
    
    ParsedItem     *item;
    ParsedItemptr  *hash, ptr;
    RArrayPtr  posary, rhsary;
    VALUE      itemobj, catobj;
    int  i, j, childptr;
    int  left, right, sentlen;
    
    /* ܂͊Tς */
    childptr = 0;
    itemset->sent_len = 0;
    for( i = 0; i < itemlist->len; ++i ) {
        item    = &itemset->itembuf[i];
        itemobj = itemlist->ptr[i];
        
        item->rule  = rb_funcall( itemobj, id_to_r, 0 );
        item->top   = ru_id( ru_lhs( item->rule ) );
        item->left  = NUM2INT( rb_funcall( itemobj, id_left, 0 ) );
        item->right = NUM2INT( rb_funcall( itemobj, id_right, 0 ) );
        item->child = &itemset->childbuf[ childptr ];
        
        posary = RARRAY( rb_Array( itemobj ) );
        item->child_len = posary->len - 1; /* -1: E[(or[) */
        childptr += item->child_len + 1; /* +1: NULL */
        if( item->right > itemset->sent_len )
            itemset->sent_len = item->right;
    }
    itemset->sent_len += 1; /* itemset->sent_len ͎ۂ̒ 1 傫 */
    sentlen = itemset->sent_len;
    
    /* nbV */
    hash = ALLOCA_N( ParsedItemptr, itemset->sent_len * itemset->sent_len );
    create_hash( itemset, hash );
    
    /* NqĂ */
    for( i = 0; i < itemlist->len; ++i ) {
        item    = &itemset->itembuf[i];
        itemobj = itemlist->ptr[i];
        
        posary = RARRAY( rb_Array( itemobj ) );
        rhsary = RARRAY( ru_rhs( item->rule ) );
        if( posary->len != rhsary->len + 1 )
            rb_raise( rb_eNLPException, "set_item(): Position list or Rule is broken. (item #%d)", i+1 );
        for( j = 0; j < rhsary->len; ++j ) {
            catobj = rhsary->ptr[j];
            if( RTEST( rb_funcall( catobj, id_nonterm_p, 0 ) ) ) {
                left  = NUM2INT(posary->ptr[j]);
                right = NUM2INT(posary->ptr[j+1]);
                if( left >= sentlen || right >= sentlen )
                    rb_raise( rb_eNLPException, "set_item(): Position list is broken. (item #%d)", i+1 );
                item->child[j] = find_item( itemset, hash, left, right, ru_id(catobj) );
            }
            else {
                item->child[j] = NULL;
            }
        }
        item->child[ rhsary->len ] = NULL;
    }
    left  = NUM2INT( rb_funcall( startitem, id_left, 0 ) );
    right = NUM2INT( rb_funcall( startitem, id_right, 0 ) );
    itemobj = rb_funcall( startitem, id_to_r, 0 );
    itemset->start = find_item( itemset, hash, left, right, ru_id( ru_lhs( itemobj ) ) );
    
    /* lރXg쐬 */
    i = 0;
    for( j = itemset->sent_len * itemset->sent_len - 1; j >= 0; --j ) {
        for( item = hash[j]; item; item = item->next_hash )
            ++i;
    }
    itemset->headbuf_size = i;
    itemset->headbuf = ALLOC_N( ParsedItemptr, itemset->headbuf_size );
    i = 0;
    for( j = itemset->sent_len * itemset->sent_len - 1; j >= 0; --j ) {
        for( item = hash[j]; item; item = item->next_hash )
            itemset->headbuf[ i++ ] = item;
    }
}

static VALUE itemset_initialize ( VALUE self, VALUE itemsetobj )
{
    ItemSet   *itemset;
    RArrayPtr  itemlist;
    VALUE volatile itemlist_obj, startary_obj, rulemap_obj;
    
    /* f[^̐ݒ */
    Data_Get_Struct( self, ItemSet, itemset );
    rulemap_obj = rb_funcall( itemsetobj, rb_intern("rulemap"), 0 );
    Check_Type( rulemap_obj, T_ARRAY );
    
    itemlist_obj = rb_funcall( itemsetobj, rb_intern("itemlist"), 0 );
    Check_Type( itemlist_obj, T_ARRAY );
    
    startary_obj = rb_funcall( itemsetobj, rb_intern("startary"), 0 );
    Check_Type( startary_obj, T_ARRAY );
    
    itemlist = RARRAY(itemlist_obj);
    
    if( RARRAY(startary_obj)->len <= 0 )
        rb_raise( rb_eNLPException, "itemset_initialize(): Start item is not found in ItemSet." );
    
    /* ̃obt@쐬 */
    itemset->itembuf_size = itemlist->len;
    itemset->itembuf = ALLOC_N( ParsedItem, itemset->itembuf_size );
    
    /* q̃obt@ɕKvȗʂZo */
    itemset->childbuf_size = count_child_size( itemlist );
    itemset->childbuf = ALLOC_N( ParsedItem *, itemset->childbuf_size );
    
    /* f[^ݒ肷 */
    set_item( itemset, itemlist, RARRAY(startary_obj)->ptr[0] );
    
    /* ֘Af[^̐ݒ */
    itemset->rulemap = rulemap_obj;
    itemset->obj_cache = rb_ary_new3( 1, itemset->rulemap );
    rb_gc_register_address( &itemset->obj_cache );
    
    return self;
}


/* ParsedItem ̓e\
----------------------------------------------------------------------
*/
static void parseditem_display ( ParsedItem * item )
{
    ParsedItem  *c;
    int  i;
    
    if( !item ) {
        fprintf( stderr, "NULL(ParsedItem)\n" );
        return;
    }
    
    fprintf( stderr, "(%d,%d) %s ---", item->left, item->right, STR2CSTR(rb_funcall(item->rule, id_typestring, 0)) );
    for( i = 0; i < item->child_len; ++i ) {
        c = item->child[i];
        if( !c )
            continue;
        fprintf( stderr, " %s(%d,%d)", STR2CSTR(rb_String(ru_lhs(c->rule))), c->left, c->right );
    }
    fprintf( stderr, "\n" );
}


/* CYK ^Cv̎xOto͂
----------------------------------------------------------------------
*/
struct tagCYKData
{
    int  id; /* xOtp̃io[ */
};
typedef  struct tagCYKData  CYKData;

struct tagCYKDispData
{
    ItemSet        *itemset;
    CYKData        *databuf;
    ParsedItemptr  *itemlist;
    int  id;
};
typedef  struct tagCYKDispData  CYKDispData;

static void cyk_dispdat_free ( CYKDispData * d )
{
    if( d ) {
        ruby_xfree_heaps( 2, d->databuf, d->itemlist );
        ruby_xfree( d );
    }
}

static const char * getstr_rule ( VALUE rule )
{
    return STR2CSTR( rb_String( rb_funcall(rule, id_typestring, 0) ) );
}

static const char * getstr_dotted_rule ( VALUE rule, int dot )
{
    return STR2CSTR( rb_String( rb_funcall(rule, id_typestring, 1, INT2FIX(dot)) ) );
}

static void visit_cyk ( CYKDispData * dispdat, ParsedItem * item )
{
    CYKData  *data;
    int       i;
    
    if( !item )
        return;
    
    data = (CYKData *)item->data;
    if( !data )
        rb_raise( rb_eNLPException, "visit_cyk(): ItemSet is broken. (id=%d)", dispdat->id );
    if( data->id >= 0 )
        return; /* łɏ񂵂Ă */
    
    for( ; item; item = item->next_item ) {
        for( i = 0; i < item->child_len; ++i ) {
            visit_cyk( dispdat, item->child[i] );
        }
    }
    data->id = dispdat->id;
    dispdat->id += 1;
}

static void itemset_display_cyk_output ( CYKDispData * dispdat, VALUE out )
{
    ParsedItemptr  ptr, child;
    CYKData  *data;
    VALUE volatile  strobj;
    char  buf[1024], *p;
    int   i, j;
    
    for( i = 0; i < dispdat->id; ++i ) {
        ptr = dispdat->itemlist[i];
        
        strobj = rb_String( ru_lhs(ptr->rule) );
        sprintf( buf, "[ %d ]\t\t# %s(%d,%d)\n", i, STR2CSTR(strobj), ptr->left, ptr->right );
        strobj = rb_str_new2( buf );
        rb_io_write( out, strobj );
        
        for( ; ptr; ptr = ptr->next_item ) {
            p = buf;
            sprintf( p, " +- (:%s:)", getstr_rule( ptr->rule ) ); p += strlen( p );
            for( j = 0; j < ptr->child_len; ++j ) {
                child = ptr->child[j];
                if( !child )
                    continue;
                sprintf( p, " @%d", ((CYKData *)child->data)->id ); p += strlen( p );
            }
            sprintf( p, "\n" );
            strobj = rb_str_new2( buf );
            rb_io_write( out, strobj );
        }
    }
}

static VALUE itemset_display_cyk ( VALUE self, VALUE out )
{
    CYKDispData    *dispdat;
    ItemSet        *itemset;
    ParsedItemptr  *headbuf;
    CYKData        *data;
    
    VALUE volatile dispdat_obj;
    int  i, addr, sgsize;
    
    Data_Get_Struct( self, ItemSet, itemset );
    headbuf = itemset->headbuf;
    
    dispdat_obj = Data_Make_Struct( rb_cData, CYKDispData, NULL, cyk_dispdat_free, dispdat );
    dispdat->itemset = itemset;
    
    /* ܂͏ */
    sgsize = itemset->headbuf_size; /* xOt̃TCY͍Zbg炷ɔ */
    for( i = 0; i < itemset->itembuf_size; ++i )
        itemset->itembuf[i].data = NULL;
    dispdat->databuf = ALLOC_N( CYKData, sgsize );
    for( i = 0; i < sgsize; ++i ) {
        headbuf[i]->data = &dispdat->databuf[i];
        dispdat->databuf[i].id = -1;
    }
    dispdat->id = 0;
    visit_cyk( dispdat, itemset->start );
    if( dispdat->id != sgsize )
        rb_raise( rb_eNLPException, "ItemSet is broken. (id=%d, sgsize=%d)", dispdat->id, sgsize );
    
    /* Otёւ */
    dispdat->itemlist = ALLOC_N( ParsedItemptr, sgsize );
    for( i = 0; i < sgsize; ++i ) {
        data = (CYKData *)headbuf[i]->data;
        addr = (sgsize-1) - data->id;
        if( addr >= sgsize )
            rb_raise( rb_eNLPException, "Found unvisited item." );
        dispdat->itemlist[addr] = headbuf[i];
        data->id = addr;
    }
    
    /* o */
    itemset_display_cyk_output( dispdat, out );
    
    return self;
}


/* Earley ^Cv̎xOto͂
----------------------------------------------------------------------
 - ParsedItem H
    * EarleyItem: ParsedItem Ɋ֘AÂꂽ left,right,top Ɠ̂ЂƂB
 - H
    * EarleyItem: ParsedItem  left,right,top Ɠ̂̊B
    * EarleyItem: hbgЂƂ߂B
    * EarleyItem: hbgʒu (LJEg) ̍B

 - next_right 邽߂:
    * (l,r)[<X> --> <A> * <B>] Ƃ (l,*)[<X> --> <A> * <B>] ׂĂHȂ΂ȂȂB
    * (l,*)[<X> --> <A> * <B>] ŃOAgɖ߂܂Ń[v΂悢B
    * (l,r)... Cӂ (l,r2)... TB
      (l,r2)  next_right ɑ}B(l,r2) ΎgɃNB

 - (l,r)[<X> --> <A> a * <B>] ɑΉpX (l,r-1)[<X> --> <A> * a <B>] ЂƂ݂ȂB
   (l,l)[<X> --> * ...] ɑΉpXK݂̂łЂƂ݂ȂB
   (l,r)[<X> --> <A> <B> * <C>] ɑΉpX͕݂B
    (l,i)[<X> --> <A> * <B> <C>] <-- right ݂̂ς
    (i,r)[<B> --> ... *] <-- Eӂ݂̂ς

 - EarleyItem ӂɒ߂vf  left, right, rule, dot (ׂĂvΓ)
*/

/* ---- Earley  ---- */
struct tagEarleyItem
{
    ParsedItem const  *item; /* 쐬Ɏgpf[^ */
    
    int  id;   /* ΉxOt̃CfNX */
    int  flag; /* vpeBtO */
    
    VALUE  rule;  /* ΉK */
    int    dot;   /* hbgʒu */
    int    left;  /* [ʒu */
    int    right; /* E[ʒu */
    
    int  waiting; /* hbǧɂL */
    
    struct tagEarleyItem  *next_hash; /* nbVp̃f[^ */
    
    struct tagEarleyItem  *next_right; /* right ݂̂ςO */
    struct tagEarleyItem  *prev;       /* hbgЂƂ߂ right Oւ̃|C^ */
};
typedef  struct tagEarleyItem    EarleyItem;
typedef  struct tagEarleyItem *  EarleyItemptr;

#if USE_MACRO
#define is_visited(ITEM) ((ITEM)->flag & 0x01)
#define set_visited(ITEM) ((ITEM)->flag |= 0x01)
#define is_comp(ITEM) ((ITEM)->flag & 0x02)
#define set_comp(ITEM) ((ITEM)->flag |= 0x02)
#define is_scan(ITEM) ((ITEM)->flag & 0x04)
#define set_scan(ITEM) ((ITEM)->flag |= 0x04)
#else
static int is_visited ( EarleyItem const * eitem )
{
    return eitem->flag & 0x01;
}
static void set_visited ( EarleyItem const * eitem )
{
    eitem->flag |= 0x01;
}
static int is_comp ( EarleyItem const * eitem )
{
    return eitem->flag & 0x02;
}
static void set_comp ( EarleyItem const * eitem )
{
    eitem->flag |= 0x02;
}
static int is_scan ( EarleyItem const * eitem ) /* hbgO̍I[LH */
{
    return eitem->flag & 0x04;
}
static void set_scan ( EarleyItem const * eitem )
{
    eitem->flag |= 0x04;
}
#endif

static EarleyItem * init_earley_item ( EarleyItem * eitem, ParsedItem const * item, int dot )
{
    eitem->item = item;
    
    eitem->id   = -1;
    eitem->flag = 0;
    
    if( dot >= item->child_len ) {
        eitem->waiting = -1;
        set_comp( eitem );
    }
    else {
        if( item->child[dot] )
            eitem->waiting = item->child[dot]->top;
        else
            eitem->waiting = -1;
    }
    
    eitem->dot  = dot;
    eitem->rule = item->rule;
    eitem->left = item->left;
    if( dot > 0 && !item->child[dot-1] )
        set_scan( eitem );
    
    eitem->next_hash = NULL;
    eitem->next_right = NULL;
    eitem->prev = NULL;
    
    return eitem;
}

static const char * getstr_earley_rule ( EarleyItem * eitem )
{
    return getstr_dotted_rule( eitem->item->rule, eitem->dot );
}


/* ---- nbVe[u ---- */
struct tagEarleyHash
{
    int  sent_len; /* +1 */
    int  cap;
    EarleyItemptr  *ptr;
};
typedef  struct tagEarleyHash  EarleyHash;

static EarleyHash * init_earley_hash ( EarleyHash * hash, int sentlen )
{
    hash->sent_len = sentlen + 1;
    hash->cap = hash->sent_len * hash->sent_len;
    hash->ptr = ALLOC_N( EarleyItemptr, hash->cap );
    
    return hash;
}

#if USE_MACRO
#define  earley_hash_at(HASH,LEFT,RIGHT)  ((HASH)->ptr[(LEFT) + (RIGHT)*(HASH)->sent_len])
#else
static EarleyItemptr earley_hash_at ( EarleyHash * hash, int left, int right )
{
    return hash->ptr[ left + right * hash->sent_len ];
}
#endif

static EarleyHash * setitem_earley_hash ( EarleyHash * hash, EarleyItem * eitem )
{
    hash->ptr[ eitem->left + eitem->right * hash->sent_len ] = eitem;
    return hash;
}

static EarleyHash * clear_earley_hash ( EarleyHash * hash )
{
    rb_zero_memory( hash->ptr, EarleyItemptr, hash->cap );
    return hash;
}

static void delete_earley_hash ( EarleyHash * hash )
{
    if( hash ) {
        ruby_xfree( hash->ptr );
    }
}

static EarleyItem * add_earley_hash ( EarleyHash * hash, EarleyItem * eitem )
{
    VALUE  rule  = eitem->rule;
    int    dot   = eitem->dot;
    int    left  = eitem->left;
    int    right = eitem->right;
    
    EarleyItemptr  ptr;
    int  i;
    
    /* Ô߂̃`FbN (Ȃf[^邱Ƃ̂) */
    if( left < 0 || left > hash->sent_len || right < 0 || right > hash->sent_len )
        rb_raise( rb_eNLPException, "add_earley_hash(): Broken item." );
    
    /* ܂͊Svf[^ */
    for( ptr = earley_hash_at( hash, left, right ); ptr; ptr = ptr->next_hash ) {
        if( ptr->rule == rule && ptr->dot == dot )
            return ptr;
    }
    
    /* right ݂̂ςOɒǉ */
    for( i = left; !eitem->next_right && i < hash->sent_len; ++i ) {
        for( ptr = earley_hash_at( hash, left, i ); ptr; ptr = ptr->next_hash ) {
            if( ptr->rule == rule && ptr->dot == dot ) {
                eitem->next_right = ptr->next_right;
                ptr->next_right = eitem;
                break;
            }
        }
    }
    if( !eitem->next_right )
        eitem->next_right = eitem;
    
    /* hbgЂƂ߂O */
    if( dot > 0 ) {
        --dot;
        if( dot == 0 ) {
            for( ptr = earley_hash_at( hash, left, left ); ptr; ptr = ptr->next_hash ) {
                if( ptr->rule == rule && ptr->dot == 0 ) {
                    eitem->prev = ptr;
                    break;
                }
            }
        }
        else {
            for( i = left; i < right; ++i ) {
                for( ptr = earley_hash_at( hash, left, i ); ptr; ptr = ptr->next_hash ) {
                    if( ptr->rule == rule && ptr->dot == dot ) {
                        eitem->prev = ptr;
                        break;
                    }
                }
            }
        }
        if( !eitem->prev )
            rb_raise( rb_eNLPException, "add_earley_hash(): (%d,*)[%s] is not found.", left, getstr_dotted_rule( rule, dot ) );
        ++dot;
    }
    
    /* nbVɒǉ */
    eitem->next_hash = earley_hash_at( hash, eitem->left, eitem->right );
    setitem_earley_hash( hash, eitem );
    
    return eitem;
}


/* ---- display \bhŎgp鋤ʃf[^ ---- */
struct tagEarleyDispData
{
    ItemSet        *itemset;
    EarleyItem     *itembuf;
    EarleyItemptr  *itemlist;
    EarleyHash      hash;
    int             sgnum;
    int             itembuf_len;
    char            strbuf[ 512 ];
};
typedef  struct tagEarleyDispData  EarleyDispData;

static void earley_dispdat_free ( EarleyDispData * data )
{
    if( data ) {
        delete_earley_hash( &data->hash );
        ruby_xfree_heaps( 2, data->itembuf, data->itemlist );
        ruby_xfree( data );
    }
}


/* ---- xOt̏o ---- */
static void construct_earley_hash ( EarleyDispData * dispdat )
{
    EarleyHash  *hash = &dispdat->hash;
    int          itembuf_size = dispdat->itemset->itembuf_size;
    
    ParsedItem const *item;
    EarleyItem       *eitem;
    int               i, j, right;
    
    item  = dispdat->itemset->itembuf;
    eitem = dispdat->itembuf;
    
    clear_earley_hash( hash );
    for( i = 0; i < itembuf_size; ++i ) {
        right = item->left;
        for( j = 0; j <= item->child_len; ++j ) {
            init_earley_item( eitem, item, j );
            eitem->right = right;
            add_earley_hash( hash, eitem ); /* Ă̂̂ꍇ̓LZ */
            right = (item->child[j])? item->child[j]->right: right+1;
            ++eitem;
        }
        ++item;
    }
}
static void earley_visit_item ( EarleyDispData * dispdat, EarleyItem * eitem )
{
    EarleyItem  *prev, *child;
    int  flag;
    
    if( is_visited(eitem) )
        return;
    set_visited( eitem );
    
    if( eitem->dot > 0 ) { /* hbgЂƂO̍To */
        if( !is_scan( eitem ) ) { /* eitem: ... <A> * ... ̏ꍇ */
            for( prev = eitem->prev; prev; prev = (prev->next_right == eitem->prev)? NULL: prev->next_right ) {
                if( prev->right >= eitem->right )
                    continue;
                flag = 0;
                for( child = earley_hash_at( &dispdat->hash, prev->right, eitem->right ); child; child = child->next_hash ) {
                    if( !is_comp( child ) || child->item->top != prev->waiting )
                        continue;
                    earley_visit_item( dispdat, child );
                    ++flag;
                }
                if( flag > 0 )
                    earley_visit_item( dispdat, prev );
            }
        }
        else { /* eitem: ... a * ... ̏ꍇ */
            flag = 0;
            for( prev = eitem->prev; prev; prev = (prev->next_right == eitem->prev)? NULL: prev->next_right ) {
                if( prev->right + 1 != eitem->right )
                    continue;
                if( flag > 0 )
                    rb_raise( rb_eNLPException, "earley_visit_item(): Found too many (%d,%d)[%s].", prev->left, prev->right, getstr_earley_rule(prev) );
                earley_visit_item( dispdat, prev );
            }
        }
    }
    
    eitem->id = dispdat->sgnum;
    ++(dispdat->sgnum);
}
static void earley_visit ( EarleyDispData * dispdat )
{
    ParsedItem const  *item;
    EarleyItem  *eitem;
    
    /* Jnɑ鍀ڂKJn */
    dispdat->sgnum = 0;
    for( item = dispdat->itemset->start; item; item = item->next_item ) {
        for( eitem = earley_hash_at( &dispdat->hash, item->left, item->right ); eitem; eitem = eitem->next_hash ) {
            if( is_comp(eitem) && eitem->rule == item->rule )
                earley_visit_item( dispdat, eitem );
        }
    }
}
static void construct_earley_list ( EarleyDispData * dispdat )
{
    EarleyHash * hash = &dispdat->hash;
    EarleyItem * eitem;
    int  i, addr;
    
    dispdat->itemlist = rb_alloc_zero( EarleyItemptr, dispdat->sgnum );
    for( i = 0; i < hash->cap; ++i ) {
        for( eitem = hash->ptr[i]; eitem; eitem = eitem->next_hash ) {
            if( eitem->id < 0 )
                rb_raise( rb_eNLPException, "itemset_display_earley(): (%d,%d)[%s] is not used.", eitem->left, eitem->right, getstr_earley_rule(eitem) );
            addr = (dispdat->sgnum - 1) - eitem->id;
            eitem->id = addr;
            if( dispdat->itemlist[addr] )
                rb_raise( rb_eNLPException, "itemset_display_earley(): Found too many ID #%d.", addr );
            dispdat->itemlist[addr] = eitem;
        }
    }
}
static void earley_output_sgraph ( EarleyDispData * dispdat, VALUE out )
{
    EarleyItem  *eitem, *prev, *child;
    VALUE volatile strobj;
    int  id, flag, right;
    
    for( id = 0; id < dispdat->sgnum; ++id ) {
        eitem = dispdat->itemlist[id];
        sprintf( dispdat->strbuf, "[ %d ]\t\t# (%d,%d) %s\n", id, eitem->left, eitem->right, getstr_earley_rule(eitem) );
        strobj = rb_str_new2( dispdat->strbuf );
        rb_io_write( out, strobj );
        
        /* visit Ɠ */
        if( eitem->dot > 0 ) {
            if( !is_scan( eitem ) ) {
                for( prev = eitem->prev; prev; prev = (prev->next_right == eitem->prev)? NULL: prev->next_right ) {
                    if( prev->right >= eitem->right )
                        continue;
                    /*  */
                    for( child = earley_hash_at( &dispdat->hash, prev->right, eitem->right ); child; child = child->next_hash ) {
                        if( !is_comp( child ) || child->item->top != prev->waiting )
                            continue;
                        sprintf( dispdat->strbuf, " +- @%d @%d\n", prev->id, child->id );
                        strobj = rb_str_new2( dispdat->strbuf );
                        rb_io_write( out, strobj );
                    }
                }
            }
            else {
                for( prev = eitem->prev; prev; prev = (prev->next_right == eitem->prev)? NULL: prev->next_right ) {
                    if( prev->right + 1 != eitem->right )
                        continue;
                    sprintf( dispdat->strbuf, " +- @%d\n", prev->id );
                    strobj = rb_str_new2( dispdat->strbuf );
                    rb_io_write( out, strobj );
                }
            }
        }
        else {
            sprintf( dispdat->strbuf, " +- (:%s:)\n", getstr_rule( eitem->item->rule ) );
            strobj = rb_str_new2( dispdat->strbuf );
            rb_io_write( out, strobj );
        }
    }
}
static VALUE itemset_display_earley ( VALUE self, VALUE out )
{
    ItemSet         *itemset;
    EarleyDispData  *dispdat;
    VALUE volatile   dispdat_obj;
    int              i;
    
    /* ʃf[^̐ */
    Data_Get_Struct( self, ItemSet, itemset );
    dispdat_obj = Data_Make_Struct( rb_cData, EarleyDispData, NULL, earley_dispdat_free, dispdat );
    dispdat->itemset = itemset;
    
    dispdat->itembuf_len = 0;
    for( i = 0; i < dispdat->itemset->itembuf_size; ++i )
        dispdat->itembuf_len += dispdat->itemset->itembuf[i].child_len + 1; /* +1: hbgʒũp^[ɂ */
    dispdat->itembuf = ALLOC_N( EarleyItem, dispdat->itembuf_len );
    
    init_earley_hash( &dispdat->hash, itemset->sent_len - 1 );
    construct_earley_hash( dispdat );
    
    earley_visit( dispdat );
    
    construct_earley_list( dispdat );
    earley_output_sgraph( dispdat, out );
    
    return self;
}


/* ItemSet NX̍\z
----------------------------------------------------------------------
*/
void init_itemset ( void )
{
    /* lCeBu `[gNX̒` */
    rb_cItemSet = rb_define_class_under( rb_mExt, "ItemSet", rb_cData );
    
    /* NX\bh */
    rb_define_singleton_methods(
        rb_cItemSet,
        
        "new"			, itemset_new, 1,
        
        NULL );
    
    /* \bh */
    rb_define_methods(
        rb_cItemSet,
        
        "initialize"		, itemset_initialize, 1,
        
        "display"		, itemset_display, 1,
        
        "display_cyk"		, itemset_display_cyk, 1,
        
        "display_earley"	, itemset_display_earley, 1,
        
        NULL );
}


