/*                                                     gEM extentions 
----------------------------------------------------------------------
    L[
      $Id: queue.c,v 1.3 2002/01/03 05:41:41 Black Exp $

        Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp> 
----------------------------------------------------------------------
*/
#define DEBUGSW 0
#include "nlputil.h"


/* PL[
----------------------------------------------------------------------
 * L[̍Ōvf shift pɊmۂĂB
*/
SimpleQueue * squeue_new ( void )
{
    return  squeue_init( ALLOC( SimpleQueue ) );
}

SimpleQueue * squeue_init ( SimpleQueue * sq )
{
    sq->buffer = NULL;
    sq->capacity = sq->size = 0;
    
    return  sq;
}

SimpleQueue * squeue_clear ( SimpleQueue * sq )
{
    sq->size = 0;
    
    return sq;
}

void squeue_free ( SimpleQueue * sq )
{
    ruby_xfree( sq->buffer );
    sq->buffer = NULL;
}

SimpleQueue * squeue_push ( SimpleQueue * sq, Item * item )
{
    if( sq->size + 1 >= sq->capacity ) {
        sq->capacity = (sq->capacity < 128)? 128: sq->capacity * 2;
        REALLOC_N( sq->buffer, Item, sq->capacity );
    }
    sq->buffer[ sq->size ] = *item;
    sq->size++;
    
    return sq;
}

int squeue_is_empty ( SimpleQueue * sq )
{
    return  (sq->size == 0);
}

Item * squeue_shift ( SimpleQueue * sq )
{
    if( sq->size > 0 ) {
        sq->buffer[ sq->capacity-1 ] = sq->buffer[ 0 ];
        --sq->size;
        memmove( &sq->buffer[0], &sq->buffer[1], sizeof(Item) * sq->size );
    }
    
    return &sq->buffer[ sq->capacity-1 ];
}

Item * squeue_find ( SimpleQueue * sq, Item * item )
{
    Itemptr  p;
    
    Rule  *rule = item->rule;
    int    left = item->left;
    int    right = item->right;
    int    i;
    
    for( i = 0; i < sq->size; ++i ) {
        p = &sq->buffer[ i ];
        if( rule == p->rule && left == p->left && right == p->right ) {
            return p;
        }
    }
    
    return NULL;
}

void squeue_inspect ( SimpleQueue * sq )
{
    int  i;
    
    fprintf( stderr, "ext/gemstolcke/queue.c: simple queue\n" );
    for( i = 0; i < sq->size; ++i ) {
        item_display( &sq->buffer[ i ] );
    }
    fprintf( stderr, "-- end --\n" );
}


/*  tL[
----------------------------------------------------------------------
*/

PrioritizedQueue * pqueue_new ( PrioritizedQueue * queue )
{
    if( !queue )
        queue = ALLOC( PrioritizedQueue );
    
    queue->ptr = NULL;
    queue->len = 0;
    queue->cap = 0;
    queue->head = queue->tail = -1;
    
    return queue;
}

PrioritizedQueue * pqueue_init ( PrioritizedQueue * pq, int length )
{
    int  i;
    
    if( length == 0 ) {
        pq->ptr = NULL;
        pq->len = 0;
        pq->cap = 0;
        
        pq->head = pq->tail = -1;
    }
    else {
        pq->len = length + 1;
        if( pq->cap < pq->len ) {
            REALLOC_N( pq->ptr, PrioritizedQueueUnit, pq->len );
            for( i = pq->cap; i < pq->len; ++i ) {
                pq->ptr[i].cap = 0;
                pq->ptr[i].ptr = NULL;
            }
            pq->cap = pq->len;
        }
        pq->head = pq->len;
        pq->tail = -1;
        
        for( i = 0; i < pq->cap; ++i )
            pq->ptr[i].len = 0;
    }
    
    return pq;
}

void pqueue_free ( PrioritizedQueue * pq )
{
    int  i;
    
    for( i = 0; i < pq->cap; ++i )
        ruby_xfree( pq->ptr[i].ptr );
    ruby_xfree( pq->ptr );
}

PrioritizedQueue * pqueue_clear ( PrioritizedQueue * pq )
{
    int i;
    
    for( i = 0; i < pq->cap; ++i )
        pq->ptr[i].len = 0;
    pq->len = 0;
    
    return  pq;
}

/* L[̐擪獀Ă = leftŏ, ntorderŏ(\؃gbv) */
Item * pqueue_shift ( PrioritizedQueue * q )
{
    PrioritizedQueueUnit  *p;
    Item  *ret;
    
    if( q->head >= q->len )
        return NULL;
    
    while( q->ptr[ q->head ].len == 0 ) {
        ++(q->head);
        if( q->head >= q->len )
            return NULL;
    }
    p = &q->ptr[ q->head ];
    
    ret = p->ptr[ 0 ];
    --(p->len);
    memmove( &p->ptr[0], &p->ptr[1], sizeof(p[0]) * p->len );
    
    return  ret;
}

/* L[̖獀Ă = leftő, ntorderő(\؍ŉ) */
Item * pqueue_pop ( PrioritizedQueue * q )
{
    PrioritizedQueueUnit  *p;
    Item  *ret;
    
    if( q->tail < 0 )
        return NULL;
    
    /* osL[̌ */
    while( q->ptr[ q->tail ].len == 0 ) {
        --(q->tail);
        if( q->tail < 0 )
            return NULL;
    }
    p = &q->ptr[ q->tail ];
    
    /* L[獀ڂo */
    ret = p->ptr[ p->len - 1 ];
    --(p->len);
    
    return  ret;
}

static void _pqueue_enlarge_buffer ( PrioritizedQueueUnit * u )
{
    Itemptr  *p;
    
    /* size == cap ̏ꍇɊg (ꂩf[^ЂƂƊm肵Ă̂) */
    if( u->len < u->cap )
        return;
    
    u->cap = (u->cap < 128)? 128: u->cap * 2;
    REALLOC_N( u->ptr, Itemptr, u->cap );
}

/* obt@ u ̈ʒu index(<= u->size)  item } */
static void _pqueue_insert ( PrioritizedQueueUnit * u, Item * item, int index )
{
    Itemptr  *p = u->ptr;
    
    if( index < u->len )
        memmove( &p[ index+1 ], &p[ index ], sizeof(p[0]) * (u->len - index) );
    p[ index ] = item;
}

static int _pqueue_index ( PrioritizedQueueUnit * u, Item * item, int * order )
{
    Itemptr  p;
    int  i, top, itemorder;
    
    itemorder = order[ item->rule->lhs ];
    for( i = 0; i < u->len; ++i ) {
        /* (l,r)[A -> ...] ͏ɗ (l,r)[X -> Y] ̒Oɑ} */
        p = u->ptr[i];
        top = p->rule->lhs;
        if( p->right != item->right || order[ top ] < 0 )
            continue;
        
        if( order[ top ] < itemorder )
            return i;
    }
    
    return u->len;
}

static int _pqueue_rindex ( PrioritizedQueueUnit * u, Item * item, int * order )
{
    Itemptr  p;
    int  i, top, itemorder;
    
    itemorder = order[ item->rule->lhs ];
    for( i = u->len - 1; i >= 0; --i ) {
        /* (l,r)[A -> ...] ͏ɗ (l,r)[X -> Y] ̒Oɑ} */
        p = u->ptr[i];
        top = p->rule->lhs;
        if( p->right != item->right || order[ top ] < 0 )
            continue;
        
        if( order[ top ] < itemorder )
            return i+1;
    }
    
    return 0;
}

Item * pqueue_push ( PrioritizedQueue * q, Item * item, int * order )
{
    PrioritizedQueueUnit  *u = q->ptr + item->left;
    
    /* obt@̑傫𒲐߂ */
    _pqueue_enlarge_buffer( u );
    
    if( order && order[ item->top ] >= 0 ) {
        /* tsꍇ */
        _pqueue_insert( u, item, _pqueue_index( u, item, order ) );
    }
    else {
        /* ts킸ɕ荞 */
        u->ptr[ u->len ] = item;
    }
    ++(u->len);
    
    if( item->left > q->tail )
        q->tail = item->left;
    
    if( item->left < q->head )
        q->head = item->left;
    
    return  item;
}

Item * pqueue_unshift ( PrioritizedQueue * q, Item * item, int * order )
{
    PrioritizedQueueUnit  *u = q->ptr + item->left;
    
    /* obt@̑傫𒲐߂ */
    _pqueue_enlarge_buffer( u );
    
    if( order ) {
        /* tsꍇ */
        _pqueue_insert( u, item, _pqueue_rindex( u, item, order ) );
    }
    else {
        /* ts킸ɕ荞 */
        memmove( &u->ptr[1], &u->ptr[0], sizeof(u->ptr[0]) * u->len );
        u->ptr[ 0 ] = item;
    }
    ++(u->len);
    
    if( item->left > q->tail )
        q->tail = item->left;
    
    if( item->left < q->head )
        q->head = item->left;
    
    return  item;
}

Item * pqueue_find ( PrioritizedQueue * q, Item * item, int * order )
{
    PrioritizedQueueUnit  *u = &q->ptr[ item->left ];
    Item  *d;
    int    i, itemorder;
    
    if( order ) {
        /* order ̂T */
        itemorder = order[ item->top ];
        for( i = 0; i < u->len; ++i ) {
            d = u->ptr[i];
            if( itemorder < order[ d->top ] )
                return  NULL;
            
            if( itemorder == order[ d->top ] ) {
                if( item->rule == d->rule && item->left == d->left && item->right == d->right )
                    return d;
            }
        }
    }
    else {
        /* 肪肪̂őŜT */
        for( i = 0; i < u->len; ++i ) {
            d = u->ptr[i];
            if( item->rule == d->rule && item->left == d->left && item->right == d->right )
                return d;
        }
    }
    
    return NULL;
}

Item * pqueue_is_include ( PrioritizedQueue * q, Item * item )
{
    PrioritizedQueueUnit  *u = &q->ptr[ item->left ];
    Item  *d;
    int    i;
    
    for( i = 0; i < u->len; ++i ) {
        if( u->ptr[i] == item )
            return item;
    }
    
    return NULL;
}

void pqueue_inspect ( PrioritizedQueue * q )
{
    PrioritizedQueueUnit  *unit;
    int  i, j, processed;
    
    processed = 0;
    for( i = 0; i < q->len; ++i ) {
        unit = &q->ptr[ i ];
        if( unit->len == 0 )
            continue;
        
        fprintf( stderr, "pqueue#%p: [@%d]\n", q, i );
        for( j = 0; j < unit->len; ++j )
            item_display( unit->ptr[j] );
        
        ++processed;
    }
    if( processed == 0 )
        fprintf( stderr, "pqueue#%p: empty (len=%d)\n", q, q->len );
}

