/*                                                     gEM extentions 
----------------------------------------------------------------------
    Ruby p
      $Id: rubytool.c,v 1.4 2002/01/03 05:41:41 Black Exp $

        Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp> 
----------------------------------------------------------------------
*/
#include "rubytool.h"
#include <stdarg.h>


/*  Ruby NX\z̃[eBeB
----------------------------------------------------------------------
*/

void  rb_p_stderr  ( VALUE v )
{
    fprintf( stderr, STR2CSTR( rb_inspect( v ) ) );
}

void  rb_define_consts  ( VALUE klass, ... )
{
    va_list  ap;
    const char *  name;
    VALUE  value;
    
    va_start( ap, klass );
    while( name = va_arg( ap, const char * ) ) {
        value = va_arg( ap, VALUE );
        rb_define_const( klass, name, value );
    }
    va_end( ap );
}

void  rb_define_singleton_methods  ( VALUE klass, ... )
{
    va_list  ap;
    const char *name;
    void  *func;
    int  argnum;
    
    va_start( ap, klass );
    while( name = va_arg( ap, const char * ) ) {
        func = va_arg( ap, void * );
        argnum = va_arg( ap, int );
        rb_define_singleton_method( klass, name, func, argnum );
    }
    va_end( ap );
}

void  rb_define_module_functions  ( VALUE module, ... )
{
    va_list  ap;
    const char *name;
    void  *func;
    int  argnum;
    
    va_start( ap, module );
    while( name = va_arg( ap, const char * ) ) {
        func = va_arg( ap, void * );
        argnum = va_arg( ap, int );
        rb_define_module_function( module, name, func, argnum );
    }
    va_end( ap );
}

void  rb_define_methods  ( VALUE klass, ... )
{
    va_list  ap;
    const char *name;
    void  *func;
    int  argnum;
    
    va_start( ap, klass );
    while( name = va_arg( ap, const char * ) ) {
        func = va_arg( ap, void * );
        argnum = va_arg( ap, int );
        rb_define_method( klass, name, func, argnum );
    }
    va_end( ap );
}

void  rb_define_aliases  ( VALUE klass, ... )
{
    va_list  ap;
    const char *oldname;
    const char *newname;
    
    va_start( ap, klass );
    while( newname = va_arg( ap, const char * ) ) {
        oldname = va_arg( ap, const char * );
        rb_define_alias( klass, newname, oldname );
    }
    va_end( ap );
}

void  ruby_xfree_heaps  ( int n, ... )
{
    va_list  ap;
    void    *ptr;
    
    va_start( ap, n );
    for( ; n > 0; --n ) {
        ptr = va_arg( ap, void * );
        ruby_xfree( ptr );
    }
    va_end( ap );
}

void  rb_gc_mark_values  ( int n, ... )
{
    va_list  ap;
    VALUE    v;
    
    va_start( ap, n );
    for( ; n > 0; --n ) {
        v = va_arg( ap, VALUE );
        rb_gc_mark( v );
    }
    va_end( ap );
}



/*   Ruby [eBeB
----------------------------------------------------------------------
*/

void * rb_alloc_zero_func ( int size, int n )
{
    return rb_zero_memory_func( ruby_xmalloc( size * n ), size, n );
}

void * rb_zero_memory_func ( void * p, int size, int n )
{
    memset( p, 0, size * n );
    
    return p;
}

