/*                                                     project Gemini 
----------------------------------------------------------------------
    xOtIuWFNg
      $Id: sgraph.c,v 1.12 2002/02/08 04:42:30 Black Exp $

        Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp> 
----------------------------------------------------------------------
*/
#define DEBUGSW 0
#include "nlputil.h"


/*   IuWFNg̍\z
----------------------------------------------------------------------
*/
static void sgraph_free ( SupportGraph * sgraph )
{
    if( sgraph ) {
        ruby_xfree_heaps( 4, sgraph->buffer_node, sgraph->buffer_path, sgraph->buffer_graph, sgraph->ext_graph );
        ruby_xfree( sgraph );
    }
}

static VALUE sgraph_new ( VALUE klass, VALUE sg )
{
    VALUE volatile obj;
    SupportGraph  *sgraph;
    
    sgraph = rb_alloc_zero( SupportGraph, 1 );
    
    obj = Data_Wrap_Struct( klass, NULL, sgraph_free, sgraph );
    rb_obj_call_init( obj, 1, &sg );
    
    return obj;
}

static VALUE sgraph_initialize ( int argc, VALUE * argv, VALUE self )
{
    const ID id_call = rb_intern( "[]" );
    
    SupportGraph  *sgraph;
    SG_Graph      *graph;
    SG_Path       *path;
    SG_Node       *node;
    
    RArrayPtr  gary, pary, nary;
    VALUE      sg, gramobj, nodeval;
    int        pindex, nindex, i, j, k;
    
    rb_scan_args( argc, argv, "11", &sg, &gramobj );
    if( NIL_P( gramobj ) )
        gramobj = rb_gv_get( "$grammar" );
    
    Data_Get_Struct( self, SupportGraph, sgraph );
    sgraph->probtbl = NULL;
    
    /* Kvȃf[^ʂ̒Ɗm
    ---------------------------------------- */
    sgraph->size_graph = NUM2INT( rb_funcall( sg, rb_intern("count_graph"), 0 ) );
    sgraph->size_path  = NUM2INT( rb_funcall( sg, rb_intern("count_path"), 0 ) );
    sgraph->size_node  = NUM2INT( rb_funcall( sg, rb_intern("count_node"), 0 ) );
    
    sgraph->buffer_graph = ALLOC_N( SG_Graph, sgraph->size_graph );
    sgraph->buffer_path  = ALLOC_N( SG_Path, sgraph->size_path );
    sgraph->buffer_node  = ALLOC_N( SG_Node, sgraph->size_node );
    
    /* IuWFNg̏
    ---------------------------------------- */
    pindex = nindex = 0;
    gary = RARRAY( rb_Array( sg ) );
    for( i = 0; i < gary->len; ++i ) {
        if( NIL_P(gary->ptr[i]) )
            rb_raise( rb_eNLPException, "Uninitialized subgraph #%d.", i );
        
        graph = &sgraph->buffer_graph[i];
        pary = RARRAY( rb_Array( gary->ptr[i] ) );
        
        graph->pathary = &sgraph->buffer_path[ pindex ];
        graph->pathlen = pary->len;
        for( j = 0; j < pary->len; ++j ) {
            if( NIL_P(pary->ptr[j]) )
                rb_raise( rb_eNLPException, "Uninitialized path #%d in subgraph #%d.", j, i );
            
            path = &graph->pathary[j];
            nary = RARRAY( rb_Array( pary->ptr[j] ) );
            
            path->nodeary = &sgraph->buffer_node[ nindex ];
            path->nodelen = nary->len;
            for( k = 0; k < nary->len; ++k ) {
                node = &path->nodeary[k];
                nodeval = nary->ptr[k];
                
                switch( TYPE( nodeval ) ) {
                case T_FIXNUM:
                    node->d.ind.type = 2;
                    node->d.ind.data.graph = &sgraph->buffer_graph[ FIX2INT(nodeval) ];
                    break;
                case T_STRING:
                    node->d.ind.type = 1;
                    node->d.ind.data.sw = ru_id( rb_funcall( gramobj, id_call, 1, nodeval ) );
                    break;
                default:
                    node->d.ind.type = 1;
                    node->d.ind.data.sw = ru_id( nodeval );
                    break;
                }
            }
            nindex += nary->len;
        }
        pindex += pary->len;
    }
    
    return self;
}


/*   ȅo
----------------------------------------------------------------------
*/
static VALUE sgraph_display ( VALUE self, VALUE out )
{
    SupportGraph  *sgraph;
    SG_Graph      *graph;
    SG_Path       *path;
    SG_Node       *node;
    
    VALUE volatile lineary, strobj;
    char   buf[1024];
    char  *p;
    int    i, j, k;
    
    Data_Get_Struct( self, SupportGraph, sgraph );
    if( sgraph->probtbl ) {
        sprintf( buf, "<SupportGraph#%p>", sgraph );
        strobj = rb_str_new2( buf );
        rb_io_write( out, strobj );
    }
    else {
        strobj  = Qnil;
        lineary = rb_ary_new2( sgraph->size_path + sgraph->size_graph );
        for( i = 0; i < sgraph->size_graph; ++i ) {
            sprintf( buf, "[ %d ]", i );
            strobj = rb_str_new2( buf );
            rb_ary_push( lineary, strobj );

            graph = &sgraph->buffer_graph[ i ];
            for( j = 0; j < graph->pathlen; ++j ) {
                p = buf;
                sprintf( p, " +-" ); p += strlen( p );

                path = &graph->pathary[ j ];
                for( k = 0; k < path->nodelen; ++k ) {
                    node = &path->nodeary[ k ];

                    switch( node->d.ind.type ) {
                    case 1:
                        sprintf( p, " (:%d:)", node->d.ind.data.sw );
                        break;
                    case 2:
                        sprintf( p, " @%d", node->d.ind.data.graph - sgraph->buffer_graph );
                        break;
                    default:
                        sprintf( p, " (:!BROKEN!:)" );
                        break;
                    }
                    p += strlen( p );
                }
                strobj = rb_str_new2( buf );
                rb_ary_push( lineary, strobj );
            }
        }
        strobj = rb_str_new2("\n");
        strobj = rb_ary_join( lineary, strobj );
        rb_io_write( out, strobj );
    }
    
    return self;
}


/*   vZpɃOtRpC
----------------------------------------------------------------------
*/

SupportGraph * sgraph_compile ( SupportGraph * sg, int rulesize, real_t * probtbl, real_t * xnumtbl )
{
    int        i, index;
    SG_Node   *node = sg->buffer_node;
    SG_Graph  *g;
    
    if( sg->probtbl )
        rb_raise( rb_eNLPException, "Support graph is already compiled." );
    
    /* CfbNX  ڃN */
    for( i = 0; i < sg->size_node; ++i, ++node ) {
        switch( node->d.ind.type ) {
        case 1:
            index = node->d.ind.data.sw;
            if( index < 0 || index >= rulesize )
                rb_raise( rb_eNLPException, "SW index is out of range." );
            node->d.dir.i_ptr = &probtbl[ index ];
            node->d.dir.o_ptr = &xnumtbl[ index ];
            node->d.dir.o_inside = &sg->buffer_graph[0].inside;
            break;
        case 2:
            g = node->d.ind.data.graph;
            node->d.dir.i_ptr = &g->inside;
            node->d.dir.o_ptr = &g->outside;
            node->d.dir.o_inside = &g->inside;
            break;
        default:
            rb_raise( rb_eNLPException, "Support graph is broken." );
        }
    }
    sg->probtbl = probtbl;
    sg->xnumtbl = xnumtbl;
    sg->size_rule = rulesize;
    
    return sg;
}


/*   gpĂ郁̏𓾂
----------------------------------------------------------------------
*/
struct tagSupportGraphMemoryInfo
{
    int  total;
    int  bufsize;
};
typedef  struct tagSupportGraphMemoryInfo  SupportGraphMemoryInfo;

static SupportGraphMemoryInfo sgraph_check_memory ( VALUE self )
{
    SupportGraphMemoryInfo  info;
    SupportGraph  *sg;
    
    Data_Get_Struct( self, SupportGraph, sg );
    info.bufsize = (sg->size_graph * sizeof(SG_Graph)) + (sg->size_path * sizeof(SG_Path)) + (sg->size_node * sizeof(SG_Node));
    info.total = info.bufsize + sizeof(SupportGraph);
    
    return info;
}

static VALUE sgraph_meminfo ( VALUE self )
{
    SupportGraphMemoryInfo  info;
    char  buffer[256];
    
    info = sgraph_check_memory( self );
    
    sprintf( buffer, "%.1f KB (buf=%.1f)", (double)info.total/1024, (double)info.bufsize/1024 );
    
    return rb_str_new2( buffer );
}

static VALUE sgraph_memsize ( VALUE self )
{
    SupportGraphMemoryInfo  info;
    
    info = sgraph_check_memory( self );
    
    return INT2NUM( info.total );
}


/*   m𓾂
----------------------------------------------------------------------
*/
static VALUE sgraph_inside ( VALUE self, VALUE n )
{
    SupportGraph  *sg;
    int  addr;
    
    Data_Get_Struct( self, SupportGraph, sg );
    addr = NUM2INT( rb_Integer(n) );
    if( addr < 0 || addr >= sg->size_graph )
        return Qnil;
    
    return rb_float_new( sg->buffer_graph[addr].inside );
}

static VALUE sgraph_outside ( VALUE self, VALUE n )
{
    SupportGraph  *sg;
    int  addr;
    
    Data_Get_Struct( self, SupportGraph, sg );
    addr = NUM2INT( rb_Integer(n) );
    if( addr < 0 || addr >= sg->size_graph )
        return Qnil;
    
    return rb_float_new( sg->buffer_graph[addr].outside );
}


/*   Ot̐𓾂
----------------------------------------------------------------------
*/
static VALUE sgraph_size_graph ( VALUE self )
{
    SupportGraph  *sg;
    
    Data_Get_Struct( self, SupportGraph, sg );
    
    return INT2NUM( sg->size_graph );
}


/*   ǉf[^蓖Ă
----------------------------------------------------------------------
 - byte ^̂ŁAł߂TCY int zmۂāAgƂɁB
 - xOť`Ă犄蓖Ă邱ƁB
 - ǉf[^͕̏ʂɍsB
*/
SupportGraph * sgraph_alloc_ext ( SupportGraph * sg, int unit_size )
{
    int  i, intnum;
    
    if( !sg )
        return NULL;
    
    if( sg->ext_graph )
        ruby_xfree( sg->ext_graph );
    
    intnum = (unit_size-1)/sizeof(int) + 1;
    sg->ext_graph = ruby_xmalloc( intnum * sizeof(int) * sg->size_graph );
    for( i = 0; i < sg->size_graph; ++i ) {
        sg->buffer_graph[i].ext = &((int *)sg->ext_graph)[ i * intnum ];
    }
    
    return sg;
}


/*   xOtNX̍\z
----------------------------------------------------------------------
*/
void init_sgraph ( void )
{
    /* xOt̃NX` */
    rb_cSupportGraph = rb_define_class_under( rb_mExt, "SupportGraph", rb_cObject );
    
    /* NX\bh */
    rb_define_singleton_methods(
        rb_cSupportGraph,
        
        "new"			, sgraph_new, 1,
        
        NULL );
    
    /* \bh */
    rb_define_methods(
        rb_cSupportGraph,
        
        "initialize"		, sgraph_initialize, -1,
        
        "display"		, sgraph_display, 1,
        
        "meminfo"		, sgraph_meminfo, 0,
        
        "memsize"		, sgraph_memsize, 0,
        
        "inside"		, sgraph_inside, 1,
        
        "outside"		, sgraph_outside, 1,
        
        "size_graph"		, sgraph_size_graph, 0,
        
        NULL );
    
    /* GCAX */
    rb_define_aliases(
        rb_cSupportGraph,
        
        "size"   , "size_graph",
        
        "length" , "size_graph",
        
        NULL );
}


