/*                                                     project Gemini 
----------------------------------------------------------------------
    Stolcke  Earley p[UIuWFNg
      $Id: stolcke.c,v 1.10 2002/01/23 15:25:21 Black Exp $

        Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp> 
----------------------------------------------------------------------
*/
#define DEBUGSW 0
#include "nlputil.h"


/* stolcke_matrix.c */
VALUE stolcke_matrix_new ( VALUE klass, VALUE gramobj );
VALUE stolcke_matrix_initialize ( VALUE self, VALUE gramobj );
VALUE stolcke_matrix_update ( VALUE self );
VALUE stolcke_matrix_parse ( VALUE self, VALUE poslist );
VALUE stolcke_matrix_visit ( VALUE self, VALUE chart );
VALUE stolcke_matrix_set_pruning ( VALUE self, VALUE flag );
VALUE stolcke_matrix_get_pruning ( VALUE self );
VALUE stolcke_matrix_test_matrix ( VALUE klass, VALUE size );
VALUE stolcke_matrix_meminfo ( VALUE self );
VALUE stolcke_matrix_xnumtbl ( VALUE self );
VALUE stolcke_matrix_compute_unit_matrix ( VALUE self );
VALUE stolcke_matrix_store ( VALUE self, VALUE chart );
VALUE stolcke_matrix_chartlist ( VALUE self );
VALUE stolcke_matrix_get_loopcount ( VALUE self );
VALUE stolcke_matrix_set_loopcount ( VALUE self, VALUE count );

VALUE stolcke_matrix_bench_update ( VALUE self );


/*   I[L̕ёւ
----------------------------------------------------------------------
    A --> B ƂKĂAe[u쐬B
*/
static int * ntlist_sort_sub ( Grammar * gram, int * tbl, int top, int toplevel )
{
    int  nonterm = gram->size_lhshash;
    int  symbol  = gram->size_lchash;
    
    RulePtr  r;
    int  i;
    
    /* K葽HƂƂ̓[v */
    if( toplevel > gram->size_rule )
        return NULL;
    
    /* top I[LłȂΏI */
    if( top >= nonterm )
        return tbl;
    
    /* łɑ傫ԍ蓖ĂĂΏI */
    if( tbl[ top ] >= toplevel )
        return tbl;
    
    /* top --> something ƂKĂ */
    tbl[ top ] = toplevel;
    for( r = gram->hash_lhs[ top ]; r; r = r->next_lhs ) {
        if( r->rhslen != 1 || r->rhs[0] >= nonterm )
            continue;
        ntlist_sort_sub( gram, tbl, r->rhs[0], toplevel+1 );
    }
    
    return tbl;
}

static int * ntlist_sort ( Grammar * gram )
{
    int  nonterm = gram->size_lhshash;
    int  symbol  = gram->size_lchash;
    
    Rule  *r, *p;
    int   *tbl;
    int    i, j, k, count;
    
    tbl = ALLOC_N( int, symbol );
    for( i = 0; i < symbol; ++i )
        tbl[ i ] = -1;
    
    /* gbv̋L݂̂HĂ */
    for( i = 1; i < gram->size_rule; ++i ) {
        /* A --> B `̋K݂̂𒲂ׂĂ */
        r = &gram->rules[ i ];
        if( r->rhslen != 1 || r->rhs[0] >= nonterm )
            continue;
        
        /* X --> A ƂK݂邩𒲂ׂ (݂Ȃ΃gbvłȂ߃LZ) */
        j = 0;
        for( p = gram->hash_lc[ r->lhs ]; p && j == 0; p = p->next_lc ) {
            if( p->rhslen != 1 )
                continue;
            ++j;
        }
        if( j > 0 )
            continue;
        
        /* [gƂȂLɂ͕K 0 蓖Ă */
        tbl[ r->lhs ] = 0;
        if( !ntlist_sort_sub( gram, tbl, r->rhs[0], 1 ) ) {
            ruby_xfree( tbl );
            rb_raise( rb_eNLPException, "detected an unary loop." );
        }
    }
    
    return tbl;
}


/*   ʕf[^p̃R[h
----------------------------------------------------------------------
*/
GenericStolcke * stolcke_generic_new ( GenericStolcke * stolcke )
{
    rb_zero_memory( stolcke, GenericStolcke, 1 );
    
    return stolcke;
}

GenericStolcke * stolcke_generic_initialize ( GenericStolcke * stolcke, Grammar * grammar )
{
    int  size_rule;
    
    if( !stolcke )
        stolcke = ALLOC( GenericStolcke );
    
    stolcke->grammar = grammar;
    stolcke->chart_ary = rb_ary_new();
    stolcke->loopcount = 1;
    
    size_rule = grammar->size_rule;
    stolcke->probtbl = ALLOC_N( double, size_rule );
    stolcke->xnumtbl = ALLOC_N( double, size_rule );
    stolcke->xnumbuf = ALLOC_N( double, size_rule );
    
    stolcke->predtbl = ALLOC_N( int, grammar->size_lhshash );
    stolcke_generic_clear_predtbl( stolcke );
    
    /* FIRST W̎擾 */
    stolcke->first_tbl = first_tbl_get( grammar );
    
    stolcke->object_cache = rb_ary_new3( 2, stolcke->grammar->orig, stolcke->chart_ary );
    rb_gc_register_address( &stolcke->object_cache );
    
    return stolcke;
}

void stolcke_generic_delete ( GenericStolcke * stolcke )
{
    if( stolcke ) {
        rb_gc_unregister_address( &stolcke->object_cache );
        ruby_xfree_heaps( 5, stolcke->probtbl, stolcke->xnumtbl, stolcke->xnumbuf, stolcke->predtbl, stolcke->first_tbl );
    }
}

void stolcke_generic_clear_predtbl ( GenericStolcke * stolcke )
{
    rb_zero_memory( stolcke->predtbl, int, stolcke->grammar->size_lhshash );
}


/*   ʕ⏕֐
----------------------------------------------------------------------
*/
int stolcke_reachable_p ( GenericStolcke * stolcke, int keycat, VALUE cat )
{
    /* keycat  cat ֓Bł邩𒲂ׂ */
    RArrayPtr  catary;
    int  *tbl;
    int   n;
    
    if( !stolcke->first_tbl )
        return 1;
    
    /* keycat I[L̏ꍇ cat Ɠꍇ̂ */
    if( keycat >= stolcke->grammar->size_lhshash ) {
        switch( TYPE(cat) )
        {
        case T_ARRAY: /* ЂƂ̒Pɕ̉\ꍇ */
            catary = RARRAY( cat );
            for( n = 0; n < catary->len; ++n ) {
                if( keycat == NUM2INT( catary->ptr[n] ) )
                    return 1;
            }
            break;
        default: /* P͂ЂƂ */
            if( keycat == NUM2INT( cat ) )
                return 1;
            break;
        }
        return 0;
    }
    
    /* e[u𒲂ׂ */
    tbl = stolcke->first_tbl[ keycat ];
    switch( TYPE(cat) )
    {
    case T_ARRAY: /* ЂƂ̒Pɕ̉\ꍇ */
        catary = RARRAY( cat );
        for( n = 0; n < catary->len; ++n ) {
            if( tbl[ NUM2INT( catary->ptr[n] ) ] )
                return 1;
        }
        break;
    default: /* P͂ЂƂ */
        if( tbl[ NUM2INT( cat ) ] )
            return 1;
        break;
    }
    return 0;
}

VALUE stolcke_solve_poslist_nest ( VALUE pos, RArrayPtr posary_list )
{
    VALUE  labelarray;
    int    n;
    
    while( FIXNUM_P(pos) && (n = FIX2INT(pos)) < 0 ) { /* n ̏ꍇ͑̕Xgɑ΂郊N */
        n = -n;
        
        if( n >= posary_list->len )
            rb_raise( rb_eNLPException, "ext/nlputil/stolcke.c: stolcke_solve_poslist_nest(): PoSList is broken." );
        
        labelarray = rb_Array( posary_list->ptr[ n ] );
        pos = RARRAY(labelarray)->ptr[0];
    }
    
    return pos;
}


/*   Prediction
----------------------------------------------------------------------
*/
void stolcke_prediction ( GenericStolcke * stolcke, Chart * chart, int position, VALUE prefetch )
{
    Grammar  *gram;
    int      *tbl;
    
    Rule  *r;
    Item  *p, *newitem;
    int    waiting;
    int    updated = 0, cancelled = 0;
    
    gram = stolcke->grammar;
    tbl  = stolcke->predtbl;
    
    /* ǂ݂PȊȌꍇ̓LZ */
    if( FIXNUM_P(prefetch) && FIX2INT(prefetch) < 0 )
        return;
    
    do {
        updated = 0;
        for( p = chart_dot( chart, position ); p; p = p->next_dot ) {
            /* position ܂ŉ͂ꂽ̍Ɋւāc */
            if( p->dot >= p->rule->rhslen )
                continue;
            
            waiting = p->rule->rhs[ p->dot ]; /* waiting: KŃhbg̎ɂL */
            if( waiting >= gram->size_lhshash )
                continue; /* I[L琶K݂͑Ȃ */
            
            if( tbl[ waiting ] )
                continue; /* ҋ@̌ꂩ͂łpredictionsĂ */
            
            if( stolcke_reachable_p( stolcke, waiting, prefetch ) ) {
                /* ҋ@L (waiting) ǂ (prefetch) ɓBłꍇ */
                for( r = gram->hash_lhs[waiting]; r; r = r->next_lhs ) {
                    if( stolcke_reachable_p( stolcke, r->rhs[0], prefetch ) ) {
                        /* ҋ@L (waiting) Ŏn܂K̂EӍ[ǂ (prefetch) ɓBłꍇ (KB킯ł͂Ȃ) */
                        newitem = item_new( r, position, position, 0 );
                        newitem->gamma = stolcke->probtbl[ r->id ];
                        chart_additem( chart, newitem );
                        ++updated;
                    }
                    else
                        ++cancelled;
                }
            }
            else
                ++cancelled;
            tbl[waiting] = 1;
        }
    } while( updated > 0 );
    chart->count_cancelled += cancelled;
    
    stolcke_generic_clear_predtbl( stolcke ); /* gÎŃNA */
}

void stolcke_prediction_toplevel ( GenericStolcke * stolcke, RArrayPtr posary_list, Chart * chart, VALUE prefetch )
{   /* [$Start --> * ...] ̍𐶐  */
    Grammar  *gram;
    int      *tbl;
    
    VALUE  labelarray;
    Rule  *r;
    Item  *p;
    int    cancelled = 0;
    int    keycat;
    
    gram = stolcke->grammar;
    tbl  = stolcke->predtbl;
    prefetch = stolcke_solve_poslist_nest( prefetch, posary_list ); /* ǂ݂͕KP */
    keycat = NUM2INT( rb_gv_get( "$Start" ) ); /* JnLg */
    
    if( !stolcke_reachable_p( stolcke, keycat, prefetch ) ) /* keycat ǂ݂ɓBł邩𒲂ׂ */
        rb_raise( rb_eNLPException, "Start symbol cannot reach a prefetch symbol." );
    
    for( r = gram->hash_lhs[keycat]; r; r = r->next_lhs ) {
        p = item_new( r, 0, 0, 0 );
        p->gamma = stolcke->probtbl[ r->id ];
        
        if( r->rhslen == 0 ) /* ÐK̏ꍇ */
            rb_raise( rb_eNLPException, "Found an epsilon rule. (RuleID: %d)", r->id );
        
        chart_additem( chart, p );
    }
    tbl[ keycat ] = 1; /*  Predictioin Ȃ̂ŃtOݒ */
    chart->count_cancelled += cancelled;
}


/*   gĂȂ̍폜
----------------------------------------------------------------------
*/
void stolcke_pruning ( Chart * chart )
{
    chart_pruning( chart );
}


/*   p[UNX̍\z
----------------------------------------------------------------------
*/
void init_stolcke ( void )
{
    VALUE  rb_cBench;
    
    /* Stolcke  Earley p[TNX` */
    rb_mStolckeParser = rb_define_module_under( rb_mExt, "StolckeParser" );
    rb_cStolckeParser_matrix = rb_define_class_under( rb_mStolckeParser, "Matrix", rb_cObject );
    rb_cBench = rb_define_class_under( rb_cStolckeParser_matrix, "BenchMark", rb_cStolckeParser_matrix );
    
    /* NX\bh */
    rb_define_singleton_methods(
        rb_cStolckeParser_matrix,
        
        "new"			, stolcke_matrix_new, 1,
        
        "test_matrix"		, stolcke_matrix_test_matrix, 1,
        
        NULL );
    
    /* \bh */
    rb_define_methods(
        rb_cStolckeParser_matrix,
        
        "initialize"		, stolcke_matrix_initialize, 1,
        
        "meminfo"		, stolcke_matrix_meminfo, 0,
        
        "update"		, stolcke_matrix_update, 0,
        
        "parse"			, stolcke_matrix_parse, 1,
        
        "visit"			, stolcke_matrix_visit, 1,
        
        "store"			, stolcke_matrix_store, 1,
        
        "chartlist"		, stolcke_matrix_chartlist, 0,
        
        "pruning="		, stolcke_matrix_set_pruning, 1,
        
        "pruning"		, stolcke_matrix_get_pruning, 0,
        
        "xnumtbl"		, stolcke_matrix_xnumtbl, 0,
        
        "compute_unit_matrix"	, stolcke_matrix_compute_unit_matrix, 0,
        
        "loopcount"		, stolcke_matrix_get_loopcount, 0,
        
        "loopcount="		, stolcke_matrix_set_loopcount, 1,
        
        NULL );
        
    rb_define_methods(
        rb_cBench,
        
        "update"		, stolcke_matrix_bench_update, 0,
        
        NULL );
    
    /* GCAX */
    rb_define_aliases(
        rb_cStolckeParser_matrix,
        
        "<<"   , "store",
        
        "to_a" , "chartlist",
        
        NULL );
}


