/*                                                     project Gemini 
----------------------------------------------------------------------
    Stolcke  Earley p[U (s)
      $Id: stolcke_matrix.c,v 1.11 2002/01/30 09:52:16 Black Exp $

        Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp> 
----------------------------------------------------------------------
*/
#define DEBUGSW 0
#include "nlputil.h"


/*   ͓ȑ
----------------------------------------------------------------------
*/
struct tagParseVar
{
    MatrixStolcke     *stolcke;
    Chart             *chart;
    PrioritizedQueue  *queue;
    
    RArrayPtr  posary_list;
    VALUE      posary_object;
};
typedef  struct tagParseVar  ParseVar;


static VALUE matrix_parse ( ParseVar * var, Chart * parentchart, int posline );


/*   RXgN^Ȃ
----------------------------------------------------------------------
*/
void stolcke_matrix_free ( MatrixStolcke * stolcke )
{
    if( stolcke ) {
        matrix_free( &stolcke->ru );
        stolcke_generic_delete( &stolcke->gen );
        ruby_xfree( stolcke );
    }
}

VALUE stolcke_matrix_new ( VALUE klass, VALUE gramobj )
{
    MatrixStolcke  *stolcke;
    VALUE volatile obj;
    
    stolcke = rb_alloc_zero( MatrixStolcke, 1 );
    stolcke_generic_new( &stolcke->gen );
    matrix_new( &stolcke->ru );
    
    obj = Data_Wrap_Struct( klass, NULL, stolcke_matrix_free, stolcke );
    rb_obj_call_init( obj, 1, &gramobj );
    
    return obj;
}

VALUE stolcke_matrix_initialize ( VALUE self, VALUE gramobj )
{
    MatrixStolcke  *stolcke;
    Grammar  *gram;
    
    ru_is_instance_of( gramobj, rb_cGrammar );
    Data_Get_Struct( gramobj, Grammar, gram );
    
    Data_Get_Struct( self, MatrixStolcke, stolcke );
    stolcke_generic_initialize( &stolcke->gen, gram );
    matrix_init( &stolcke->ru, gram->size_lhshash, gram->size_lhshash );
    
    return self;
}


/*   R_U ̎Zo
----------------------------------------------------------------------
 - P ̍\
  - col=x, row=y ̈ʒu P_U( y --> x ) B
    cɍӁAɉEӂƂB
*/
static int fill_exist_left ( Grammar * gram, int * buf )
{
    /* eL unit production ɂčӂɌ񐔂L^ */
    int  i, count, cat;
    Rule  *r;
    
    cat = 0;
    for( i = 0; i < gram->size_lhshash; ++i ) {
        count = 0;
        for( r = gram->hash_lc[i]; r; r = r->next_lc ) {
            if( r->rhslen != 1 )
                continue;
            ++count;
        }
        buf[i] = count;
        cat += (count > 0)? 1: 0;
    }
    
    return cat;
}

static int fill_exist_right ( Grammar * gram, int * buf )
{
    /* eL unit production ɂĉEӂɌ񐔂L^ */
    int  i, count, cat;
    Rule  *r;
    
    cat = 0;
    for( i = 0; i < gram->size_lhshash; ++i ) {
        count = 0;
        for( r = gram->hash_lhs[i]; r; r = r->next_lhs ) {
            if( r->rhslen != 1 )
                continue;
            ++count;
        }
        buf[i] = count;
        cat += (count > 0)? 1: 0;
    }
    
    return cat;
}

static Matrix compute_inverse ( MatrixStolcke * stolcke, Matrix const * pu )
{
    Matrix  ret;
    
    ret = matrix_dup( pu );
    matrix_inverse( &ret );
    
    return ret;
}

static void compute_unit_matrix ( MatrixStolcke * stolcke )
{   /* stolcke->ru  */
    ID  id_prob = rb_intern( "pr" );

    GenericStolcke  *gen;
    Grammar         *gram;
    
    Matrix   pu;
    Rule    *r;
    int      i, ntsize;
    
    gen    = &stolcke->gen;
    gram   = stolcke->gen.grammar;
    ntsize = gram->size_lhshash;
    
    /* pu  I - P_U ɂ */
    pu.rows = ntsize;
    pu.cols = ntsize;
    pu.buf  = ALLOCA_N( double, pu.rows * pu.cols );
    pu.m    = ALLOCA_N( double *, pu.rows );
    matrix_set_anchor( &pu );
    memset( pu.buf, 0, sizeof(double) * pu.rows * pu.cols );
    for( i = 0; i < ntsize; ++i ) {
        for( r = gram->hash_lhs[i]; r; r = r->next_lhs ) {
            if( r->rhslen != 1 || r->rhs[0] >= ntsize )
                continue;
            pu.m[r->lhs][r->rhs[0]] = -gen->probtbl[r->id];
        }
        pu.m[i][i] += 1.0;
    }
    
    /* 炩̎i R_U ߂ */
    if( stolcke->ru.rows != 0 && stolcke->ru.cols != 0 )
        matrix_free( &stolcke->ru );
    stolcke->ru = compute_inverse( stolcke, &pu );
    #if 0
    { /* R_U ̂ non-zero ȕ\ */
        int j;
        double *m;
        ID g = rb_intern("getcat");
        for( i = 0; i < ntsize; ++i ) {
            for( j = 0; j < ntsize; ++j ) {
                m = stolcke->ru.m[i];
                if( m[j] != 0 ) {
                    fprintf( stderr, "%s -U-> %s : %f\n",
                             STR2CSTR( rb_String( rb_funcall(gram->orig, g, 1, INT2FIX(i)) ) ),
                             STR2CSTR( rb_String( rb_funcall(gram->orig, g, 1, INT2FIX(j)) ) ),
                             m[j] );
                }
            }
        }
    }
    #endif
}


/*   Om̌vZ
----------------------------------------------------------------------
 - (i,j) X --> ...
   ɊւāAi  Aj   ̏ŕёւB(Ɏ擾ł)
 - item->data L[̃NɎg
*/
struct tagItemQueue
{
    Itemptr  *ptr;
    int  len, cap;
};
typedef  struct tagItemQueue  ItemQueue;

static void iqueue_display_line ( ItemQueue * queue, int left )
{
    Itemptr  p;
    
    fprintf( stderr, "Q[%d]****\n", left );
    for( p = queue->ptr[ left ]; p; p = p->data )
        item_display( p );
    fprintf( stderr, "----\n" );
}

/*
[a](l,r) <X> --> <A> <B> * <C>  [b](l,r2) <X> --> <A> * <B> <C>  [c](r2,r) <B> --> ... * B
Al <= r2 < rB̂߁Al != r2 ̏ꍇ [b]  [c] ̏ŃL[ɕԁB
l == r2 ɂȂ̂̓hbgʒu 1 ̂ƂŁÂƂ [b]  [c] ̏Ԃ͒`łȂA[b] ̓hbgʒuK 0 ɂȂ̂ŖłB

ǂ̍A]̂͑ĂxłȂ΂ȂȂB
L[̓ł́A\؂ŏɗ̂Ɏo悤ɂȂĂȂ΂ȂȂB
̐ɋCȂ΂ȂȂ̂́A(l,r) ̍瓯͈͂ (l,r) ƂB
Ȃ킿Ahbgʒu 1 ŁA擪̋LI[L̂ƂB
[a] <X> --> <A> * ...
[b] <Y> --> ... * ...
[a]  [b] ̂ǂ炪ɗ邩𒲂ׂꍇÂ悤ɂȂB

܂A[a]  unit production ̏ꍇAȂ킿
[a1] <X> --> <A> *
̏ꍇAR_U(<X> => <Y>) > 0 Ȃ [a] ̕ɂȂB

ɁA[a] ȊȌꍇAȂ킿
[a2] <X> --> <A> * <B> ...
̏ꍇAR_U(<A> => <Y>) > 0 Ȃ [a] ̕ɂȂB

܂A[a1]  [a2] 邱Ƃ͂蓾Ȃ̂ŁA[a1] ̕KɔzuB
*/
static void iqueue_push ( ItemQueue * queue, ParseVar * var, Itemptr item )
{   /* 1: left ̏, 2: right ̑傫, 3: L̏㉺, Ń\[g */
    MatrixStolcke  *matstolcke = var->stolcke;
    Itemptr  next, *prev;
    double  **mat;
    int  cat, ntsize;
    
    if( item->dot == 0 )
        return; /* hbgʒu 0 ͓̂̂o^łȂ */
    
    mat    = matstolcke->ru.m;
    ntsize = matstolcke->gen.grammar->size_lhshash;
    if( item->left < queue->len )
        queue->len = item->left; /* left ̂擾Ă悤ɂ */
    
    prev = &(queue->ptr[item->left]);
    next = *prev;
    
    cat = item->rule->rhs[0];
    if( item->dot == 1 && cat < ntsize ) {
        /* X --> A * B C ̏ꍇ (l,r) ̓o邩Ȃ̂Ń\[gKv */
        if( item->rule->rhslen == 1 ) {
            while( next ) {
                if( item->right == next->right && next->rule->rhslen == 1 ) { /* next ͓ */
                    if( mat[ cat ][ next->top ] > 0 )
                        break;
                }
                if( item->right > next->right ) break; /* next ͋E */
                prev = &((*prev)->data);
                next = *prev;
            }
        }
        else {
            while( next ) {
                if( item->right == next->right ) { /* next ͓ */
                    if( mat[ cat ][ next->top ] > 0 || next->rule->rhslen == 1 )
                        break;
                }
                if( item->right > next->right ) break; /* next ͋E */
                prev = &((*prev)->data);
                next = *prev;
            }
        }
    }
    else {
        while( next ) {
            if( item->right > next->right ) break; /* next E */
            prev = &((*prev)->data);
            next = *prev;
        }
    }
    item->data = *prev;
    *prev = item;
}

static Itemptr iqueue_shift ( ItemQueue * queue )
{
    Itemptr  p;
    
    while( queue->len < queue->cap ) {
        p = queue->ptr[ queue->len ];
        if( p ) {
            queue->ptr[ queue->len ] = p->data;
            return p;
        }
        ++queue->len;
    }
    return NULL;
}

static Itemptr iqueue_include_p ( ItemQueue * queue, Itemptr i )
{
    Itemptr  p;
    
    for( p = queue->ptr[i->left]; p; p = p->data ) {
        if( p == i )
            return i;
    }
    return NULL;
}

static double matrix_outside_push_start ( ParseVar * var, ItemQueue * queue )
{
    const int start = NUM2INT( rb_gv_get( "$Start" ) );
    GenericStolcke  *stolcke  = &var->stolcke->gen;
    Chart           *chart    =  var->chart;
    
    Itemptr  item;
    double   sentprob = 0.0;
    
    for( item = chart_comp( chart, 0, chart->len-1 ); item; item = item->next_comp ) {
        if( item->top != start )
            continue;
        sentprob += item->gamma;
        item->beta = 1.0;
        item->tag |= TAG_StartItem;
        iqueue_push( queue, var, item );
    }
    
    return sentprob;
}

static void matrix_outside_chain ( Chart * chart, Itemptr item )
{
    Itemptr  prev, comp;
    double   gamma;
    Rule    *rule = item->rule;
    int      cat;
    
    /* item: [<X> --> <A> *] */
    cat = rule->rhs[0];
    for( prev = chart_lwait( chart, cat, item->left ); prev; prev = prev->next_lwait ) {
        if( prev->dot != 0 || prev->rule != rule )
            continue; /* prev: [<X> --> * <A>] */
        
        gamma = 0.0;
        for( comp = chart_comp( chart, item->left, item->right ); comp; comp = comp->next_comp ) {
            if( comp->top != cat )
                continue; /* comp: [<A> --> ... *] */
            gamma += comp->gamma;
        }
        
        prev->beta += item->beta * gamma;
    }
}

static void matrix_outside ( ParseVar * var )
{
    MatrixStolcke   *matstolcke  =  var->stolcke;
    GenericStolcke  *stolcke     = &var->stolcke->gen;
    Chart           *chart       =  var->chart;
    
    ItemQueue  queue;
    Itemptr  item, pre, comp;
    double  sentprob, gamma, r, **ru;
    Rule  *rule;
    int  i, cat, ntsize, wantdot;
    
    queue.cap = chart->len;
    queue.len = queue.cap + 1;
    queue.ptr = ALLOCA_N( Itemptr, queue.cap );
    rb_zero_memory( queue.ptr, Itemptr, queue.cap );
    
    /* xnumbuf ͂Ŏgp */
    rb_zero_memory( stolcke->xnumbuf, double, stolcke->grammar->size_rule );
    
    /* Jn */
    sentprob = matrix_outside_push_start( var, &queue );
    if( sentprob == 0.0 )
        rb_raise( rb_eNLPException, "ext/nlputil/stolcke_matrix.c: matrix_outside(): Sentence prob. is zero." );
    
    /*  */
    ntsize = stolcke->grammar->size_lhshash;
    ru     = matstolcke->ru.m;
    
    while( item = iqueue_shift( &queue ) ) { /* item: <X> --> (1) (A) * (2) */
        if( item->dot == 0 )
            continue; /* hbgʒu 0 ̂̂͏łȂ */
        
        /* (1), (2)  (A)  <A>(I[L) ̏ꍇ */
        if( (item->tag & TAG_StartItem) == 0 ) { /* JnL̂ݓʈŉsȂ */
            if( item->rule->rhslen == 1 && item->rule->rhs[0] < ntsize ) {
                matrix_outside_chain( chart, item );
                continue;
            }
        }
        
        wantdot = item->dot - 1;
        rule    = item->rule;
        cat     = rule->rhs[ wantdot ];
        for( pre = chart_lwait( chart, cat, item->left ); pre; pre = pre->next_lwait ) {
            if( pre->rule != rule || pre->dot != wantdot )
                continue; /* pre: <X> --> (1) * (A) (2) */
            
            if( (cat < ntsize && pre->right >= item->right) || (cat >= ntsize && pre->right+1 != item->right) )
                continue; /* pre ̈ʒu`FbN */
            
            gamma = 0.0;
            if( cat < ntsize ) { /* (A)  <A>(I[L) łꍇ */
                for( comp = chart_comp( chart, pre->right, item->right ); comp; comp = comp->next_comp ) {
                    /* <A> Ɠ͈͂̊ <B> ɂāc */
                    r = ru[ cat ][ comp->top ];
                    if( r == 0.0 )
                        continue; /* R_U(<A> ==> <B>) > 0 K{ (comp: <B> --> ... *) */
                    
                    comp->beta += pre->gamma * item->beta * r;
                    if( !iqueue_include_p( &queue, comp ) )
                        iqueue_push( &queue, var, comp );
                    
                    if( comp->rule->rhslen != 1 || comp->rule->rhs[0] >= ntsize ) {
                        /* comp: <B> --> <C> * (unit production) ȊȌꍇ */
                        gamma += comp->gamma * r;
                    }
                }
                if( gamma == 0.0 )
                    continue; /* p[WOłȂ̂Œ~Ď pre  */
            }
            else { /* (A)  a(I[L) ̏ꍇ reverse-scanning */
                gamma = 1.0;
            }
            pre->beta += gamma * item->beta;
            if( !iqueue_include_p( &queue, pre ) )
                iqueue_push( &queue, var, pre );
        }
    }
    
    /* Ғl̍XV */
    for( i = 0; i < chart->len; ++i ) {
        for( item = chart_dot( chart, i ); item; item = item->next_dot ) {
            if( item->dot != 0 )
                continue;
            stolcke->xnumbuf[ item->rule->id ] += (item->gamma * item->beta);
        }
    }
    for( i = 1; i < stolcke->grammar->size_rule; ++i ) { /* obt@̔f */
        if( stolcke->xnumbuf[i] > 0 ) {
            stolcke->xnumtbl[i] += stolcke->xnumbuf[i] / sentprob;
        }
    }
}


/* scanning
----------------------------------------------------------------------
*/
static void scanning_word ( ParseVar * var, int position, int cat )
{
    GenericStolcke  *stolcke = &var->stolcke->gen;
    Chart  *chart = var->chart;
    Item  *p, *i;
    
    /* cat ƏI[LɂȂĂ邩𒲂ׂ */
    if( cat < stolcke->grammar->size_lhshash || cat >= stolcke->grammar->size_lchash )
        rb_raise( rb_eNLPException, "ext/nlputil/stolcke_matrix.c: scanning_word(): Invalid symbol ID, [%d].", cat );
    
    /* 쐬ꂽĂ */
    for( p = chart_rwait( chart, cat, position ); p; p = p->next_rwait ) {
        i = item_new( p->rule, p->left, p->right+1, p->dot+1 );
        i->gamma = p->gamma;
        if( i->dot >= p->rule->rhslen )
            pqueue_unshift( var->queue, i, NULL );
        chart_additem( chart, i );
    }
}

static int scanning ( ParseVar * var, int position, VALUE prefetch )
{
    RArrayPtr  ary;
    ParseVar   tmpvar;
    VALUE  val;
    int    n, len;
    
    switch( TYPE(prefetch) )
    {
    case T_FIXNUM: /* P */
        n = FIX2INT(prefetch);
        if( n < 0 ) { /* lXg͕̃Xgɔ */
            tmpvar = *var;
            val = matrix_parse( &tmpvar, var->chart, -n );
            len = ru_length( var->posary_list->ptr[-n] );
            chart_merge( var->chart, tmpvar.chart, position, position+len ); /* chart ŏC */
            chart_free_internal( tmpvar.chart ); /* Ruby  GC ŉ̂Ńf[^̂ݍ폜 */
            
            return len;
        }
        else {
            scanning_word( var, position, n );
        }
        break;
        
    case T_ARRAY: /* P */
        ary = RARRAY(prefetch);
        for( n = 0; n < ary->len; ++n )
            scanning_word( var, position, NUM2INT(ary->ptr[n]) );
        break;
    }
    
    return 1;
}


/* completion
----------------------------------------------------------------------
*/
static void completion ( ParseVar * var )
{
    MatrixStolcke     *matstolcke  =  var->stolcke;
    GenericStolcke    *stolcke     = &var->stolcke->gen;
    Chart             *chart       =  var->chart;
    PrioritizedQueue  *queue       =  var->queue;
    
    Itemptr   item, sameitem, citem, witem;
    Item      tempitem;
    int       waiting, ntsize;
    double  **ru, r;
    
    ntsize = stolcke->grammar->size_lhshash;
    ru = matstolcke->ru.m;
    while( citem = pqueue_pop( queue ) )
    {
        if( citem->rule->rhslen == 1 && citem->rule->rhs[0] < ntsize )
            continue; /* Unit Production ͎gpȂ */
        
        for( waiting = 0; waiting < ntsize; ++waiting ) {
            r = ru[waiting][citem->top];
            if( r == 0.0 )
                continue; /* sȂ̂Ŗ */
            
            for( witem = chart_rwait( chart, waiting, citem->left ); witem; witem = witem->next_rwait )
            {
                /* tempitem  witem ̃hbgi߂̂ɂ */
                item_reset( &tempitem, witem->rule, witem->left, citem->right, witem->dot + 1 );

                item = chart_find( chart, &tempitem );
                if( !item ) { /* `[gɌȂΐēo^ */
                    item = ALLOC( Item );
                    *item = tempitem;
                    chart_additem( chart, item );
                }

                if( item->dot >= item->rule->rhslen ) {
                    if( !pqueue_find( queue, item, NULL ) ) {
                        pqueue_unshift( queue, item, NULL );
                    }
                }
                item->gamma += witem->gamma * citem->gamma * r;
            }
        }
    }
}


/*   \
----------------------------------------------------------------------
*/
static VALUE matrix_parse ( ParseVar * var, Chart * parentchart, int posline )
{
    MatrixStolcke   *matstolcke = var->stolcke;
    GenericStolcke  *stolcke = &var->stolcke->gen;
    Chart  *chart;
    
    PrioritizedQueue  queue;
    RArrayPtr         posary;
    
    VALUE volatile chartobj, posary_obj;
    VALUE  label, prefetch;
    int    left, position; /* position  posary łǂ܂ŉ͂i񂾂 */
    
    
    /* `[g̏
    ---------------------------------------- */
    Dputs( "ext/gemstolcke/stolcke_matrix.c: matrix_parse()\n" );
    
    label  = var->posary_list->ptr[posline];
    posary_obj = rb_Array(label);
    posary = RARRAY( posary_obj );
    
    chartobj = chart_new( stolcke->grammar, var->posary_object );
    Data_Get_Struct( chartobj, Chart, chart );
    var->chart = chart;
    
    if( posary->len == 0 ) /* ̏ꍇ͂낢눫̂łŉ */
        return chartobj;
    
    
    /* \͂Ɏgpf[^̏
    ---------------------------------------- */
    left = NUM2INT( ru_first( label ) );
    
    chart_move_parentitem( chart, parentchart, left ); /* e`[g̍Ă (ꎞIɐej󂳂̂Œ) */
    
    pqueue_new( &queue );
    pqueue_init( &queue, chart->len-1 );
    var->queue = &queue;
    
    
    /* \
    ---------------------------------------- */
    prefetch = posary->ptr[0];
    if( !parentchart )
        stolcke_prediction_toplevel( stolcke, var->posary_list, chart, prefetch );
    
    for( position = 0; position < posary->len; ++position ) {
        Dprintf( "[%d]", position );
        
        prefetch = posary->ptr[ position ];
        stolcke_prediction( stolcke, chart, left, prefetch );
        
        left += scanning( var, left, prefetch );
        
        completion( var );
        
        Dputs( ">" );
    }
    Dputs( "[.]" );
    
    pqueue_free( &queue );
    var->queue = NULL;
    
    Dputs( "*\n" );
    
    return  chartobj;
}

VALUE stolcke_matrix_parse ( VALUE self, VALUE poslist )
{
    MatrixStolcke  * stolcke;
    ParseVar  var;
    VALUE volatile chartobj;
    
    /* RpCς݂𒲂ׂ
    ---------------------------------------- */
    if( !ru_compiled_p( poslist ) )
        rb_raise( rb_eNLPException, "ext/nlputil/stolcke_matrix.c: Argument (poslist) is not compiled." );
    Data_Get_Struct( self, MatrixStolcke, stolcke );
    
    /* m̌vZ
    ---------------------------------------- */
    var.stolcke       = stolcke;
    var.posary_object = poslist;
    var.posary_list   = RARRAY( rb_Array(poslist) );
    var.chart         = NULL;
    var.queue         = NULL;
    chartobj = matrix_parse( &var, NULL, 0 );
    
    /* Om̌vZ
    ---------------------------------------- */
    matrix_outside( &var );
    
    /* }
    ---------------------------------------- */
    if( stolcke->gen.pruning_flag )
        stolcke_pruning( var.chart );
    
    return  chartobj;
}


/*   em̌vZ
----------------------------------------------------------------------
*/
static void matrix_visit ( ParseVar * var )
{
    MatrixStolcke   *matstolcke  =  var->stolcke;
    GenericStolcke  *stolcke     = &var->stolcke->gen;
    Chart           *chart       =  var->chart;
    
    PrioritizedQueue  queue;
    Grammar          *grammar;
    
    RArrayPtr  catlist;
    Itemptr    item, citem, witem, sameitem;
    Item       tempitem;
    int        ntsize, position, waiting, i;
    double   **ru, r;
    
    VALUE  prefetch, *catary;
    int    catary_len;
    
    
    /* gpf[^̏
    ---------------------------------------- */
    Dputs( "ext/gemstolcke/stolcke_matrix.c: matrix_visit()\n" );
    
    ntsize = stolcke->grammar->size_lhshash;
    pqueue_new( &queue );
    pqueue_init( &queue, chart->len-1 );
    
    catlist = RARRAY( rb_Array(rb_funcall( chart->poslist, id_catlist, 0 )) );
    
    if( catlist->len != chart->len-1 )
        rb_raise( rb_eNLPException, "Informations of catlist and chart are conflict. (catlist=%d, chart=%d)", catlist->len, chart->len-1 );
    
    
    /* \
    ---------------------------------------- */
    ru = matstolcke->ru.m;
    for( position = 0; position < catlist->len; ++position ) {
        Dprintf( "[%d]", position );
        prefetch = catlist->ptr[position];
        if( TYPE( prefetch ) == T_ARRAY ) {
            catary     = RARRAY(prefetch)->ptr;
            catary_len = RARRAY(prefetch)->len;
        }
        else {
            catary     = &prefetch;
            catary_len = 1;
        }
        
        /* scanning */
        for( i = 0; i < catary_len; ++i ) {
            for( witem = chart_rwait( chart, NUM2INT(catary[i]), position ); witem; witem = witem->next_rwait ) {
                for( item = chart_dot( chart, position+1 ); item; item = item->next_dot ) {
                    if( item->rule != witem->rule || item->left != witem->left )
                        continue;
                    item->gamma = witem->gamma;
                    if( item->dot >= item->rule->rhslen )
                        pqueue_unshift( &queue, item, NULL );
                }
            }
        }
        
        /* completion */
        while( citem = pqueue_pop( &queue ) ) {
            if( citem->rule->rhslen == 1 && citem->rule->rhs[0] < ntsize )
                continue;
            for( waiting = 0; waiting < ntsize; ++waiting ) {
                r = ru[waiting][citem->top];
                if( r == 0.0 )
                    continue;
                for( witem = chart_rwait( chart, waiting, citem->left ); witem; witem = witem->next_rwait )
                {
                    item_reset( &tempitem, witem->rule, witem->left, citem->right, witem->dot + 1 );
                    item = chart_find( chart, &tempitem );
                    if( !item )
                        continue; /* gȂ */
                    
                    if( item->dot >= item->rule->rhslen )
                        if( !pqueue_find( &queue, item, NULL ) )
                            pqueue_unshift( &queue, item, NULL );
                    item->gamma += witem->gamma * citem->gamma * r;
                }
            }
        }
        
        Dputs( ">" );
    }
    Dputs( "[.]\n" );
    
    pqueue_free( &queue );
}

VALUE stolcke_matrix_visit ( VALUE self, VALUE chartobj )
{
    MatrixStolcke  * stolcke;
    ParseVar  var;
    Chart  * chart;
    
    ru_is_instance_of( chartobj, rb_cChart );
    Data_Get_Struct( chartobj, Chart, chart );
    Data_Get_Struct( self, MatrixStolcke, stolcke );
    
    /* m̌vZ
    ---------------------------------------- */
    var.stolcke = stolcke;
    var.chart   = chart;
    var.queue   = NULL;
    chart_reset_probability( var.chart, &stolcke->gen ); /* `[g̍VKp[^ŃZbg */
    matrix_visit( &var );
    
    /* Om̌vZ
    ---------------------------------------- */
    matrix_outside( &var );
    
    return  chartobj;
}


/*   p[^XV
----------------------------------------------------------------------
*/
VALUE stolcke_matrix_update ( VALUE self )
{
    ID  id_prob    = rb_intern( "pr" );
    ID  id_probset = rb_intern( "pr=" );
    
    MatrixStolcke   *stolcke;
    GenericStolcke  *gen;
    Grammar         *gram;
    
    double   sum, likelihood;
    Chart   *chart;
    VALUE    prob, *chartary;
    Rule    *r;
    int      i, ntsize, loopcount;
    
    Data_Get_Struct( self, MatrixStolcke, stolcke );
    gen    = &stolcke->gen;
    gram   = gen->grammar;
    ntsize = gram->size_lhshash;
    
    /* p[^̏ */
    gen->probtbl[ 0 ] = 0.0;
    for( i = 1; i < gram->size_rule; ++i ) {
        prob = rb_funcall( gram->rules[i].orig, id_prob, 0 );
        if( NIL_P( prob ) )
            rb_raise( rb_eNLPException, "Probability of Rule-ID [%d] is not assigned.", i );
        
        gen->probtbl[ i ] = NUM2DBL( rb_Float( prob ) );
    }
    
    /* p[^̐ */
    loopcount = (gen->loopcount > 0)? gen->loopcount: 1;
    for( ; loopcount > 0; --loopcount ) {
        rb_zero_memory( gen->xnumtbl, double, gram->size_rule );
        
        /* R_U ̍쐬 */
        compute_unit_matrix( stolcke );
        
        /* nꂽubN̕] (ғKp񐔂̎Zo) */
        likelihood = 0.0;
        if( rb_block_given_p() ) {
            rb_yield( Qnil );
        }
        else {
            chartary = RARRAY(gen->chart_ary)->ptr;
            for( i = 0; i < RARRAY(gen->chart_ary)->len; ++i ) {
                stolcke_matrix_visit( self, chartary[i] );
                Data_Get_Struct( chartary[i], Chart, chart );
                likelihood += chart_likelihood_internal( chart );
            }
        }
    
        /* p[^̍XV */
        for( i = 0; i < ntsize; ++i ) {
            sum = 0.0;
            for( r = gram->hash_lhs[ i ]; r; r = r->next_lhs ) {
                sum += gen->xnumtbl[ r->id ];
            }
            if( sum == 0.0 )
                continue;
            for( r = gram->hash_lhs[ i ]; r; r = r->next_lhs ) {
                gen->probtbl[r->id] = gen->xnumtbl[r->id] / sum;
                if( gen->xnumtbl[r->id] > 0 && gen->probtbl[r->id] == 0.0 )
                    rb_raise( rb_eNLPException, "Warning: Precision is lost." );
            }
        }
    }
    
    /* p[^̐ݒ */
    for( i = 1; i < gram->size_rule; ++i )
        rb_funcall( gram->rules[i].orig, id_probset, 1, rb_float_new( (double)gen->probtbl[i] ) );
    
    return (likelihood == 0.0)? Qtrue: rb_float_new( likelihood );
}


/*   tO̐ݒƎ擾
----------------------------------------------------------------------
*/
VALUE stolcke_matrix_set_pruning ( VALUE self, VALUE flag )
{
    MatrixStolcke  * stolcke;
    
    Data_Get_Struct( self, MatrixStolcke, stolcke );
    stolcke->gen.pruning_flag = (RTEST(flag))? 1: 0;
    
    return flag;
}

VALUE stolcke_matrix_get_pruning ( VALUE self )
{
    MatrixStolcke  * stolcke;
    
    Data_Get_Struct( self, MatrixStolcke, stolcke );
    
    return (stolcke->gen.pruning_flag)? Qtrue: Qfalse;
}


/*   s񉉎Z̓쎎
----------------------------------------------------------------------
*/
VALUE stolcke_matrix_test_matrix ( VALUE klass, VALUE size )
{
    return( (matrix_test_inverse(NUM2INT(size)))? Qtrue: Qfalse );
}


/*   ̎gp󋵂𒲂ׂ
----------------------------------------------------------------------
*/
struct tagStolckeMatrixMemoryInfo
{
    int  mat;
    int  gen;
    int  total;
};
typedef  struct tagStolckeMatrixMemoryInfo  StolckeMatrixMemoryInfo;

static StolckeMatrixMemoryInfo stolcke_matrix_check_memory ( VALUE self )
{
    StolckeMatrixMemoryInfo  info;
    MatrixStolcke   *matstolcke;
    GenericStolcke  *stolcke;
    
    Matrix  *m;
    int  nonterm, symbols;
    
    Data_Get_Struct( self, MatrixStolcke, matstolcke );
    stolcke = &matstolcke->gen;
    
    m = &matstolcke->ru;
    info.mat = sizeof(double) * m->rows * m->cols;
    
    nonterm = stolcke->grammar->size_lhshash;
    symbols = stolcke->grammar->size_lchash;
    info.gen = 3 * sizeof(double) * stolcke->grammar->size_rule + sizeof(int) * nonterm;
    if( stolcke->first_tbl ) {
        info.gen += sizeof(int *) * (nonterm+1) + sizeof(int) * symbols;
    }
    
    info.total = info.mat + info.gen + sizeof(MatrixStolcke);
    
    return info;
}

VALUE stolcke_matrix_meminfo ( VALUE self )
{
    StolckeMatrixMemoryInfo  info;
    char  buffer[ 1024 ];
    
    info = stolcke_matrix_check_memory( self );
    
    sprintf( buffer, "%.1f KB (buf=%.1f)", (double)info.total/1024, (double)(info.gen+info.mat)/1024 );
    
    return rb_str_new2( buffer );
}


/*   Ғl̃Xg𓾂
----------------------------------------------------------------------
*/
VALUE stolcke_matrix_xnumtbl ( VALUE self )
{
    MatrixStolcke  *stolcke;
    Grammar  *gram;
    VALUE volatile ret;
    int  i;
    
    Data_Get_Struct( self, MatrixStolcke, stolcke );
    gram = stolcke->gen.grammar;
    
    ret = rb_ary_new2( gram->size_rule );
    for( i = 1; i < gram->size_rule; ++i ) {
        rb_ary_push( ret, rb_float_new( stolcke->gen.xnumtbl[i] ) );
    }
    
    return ret;
}


/*   R_U vZ
----------------------------------------------------------------------
*/
VALUE stolcke_matrix_compute_unit_matrix ( VALUE self )
{
    ID  id_prob = rb_intern( "pr" );
    
    MatrixStolcke   *stolcke;
    GenericStolcke  *gen;
    Grammar         *gram;
    int    i, ntsize;
    VALUE  prob;
    
    /* stolcke_matrix_update() ̃Rs[ */
    Data_Get_Struct( self, MatrixStolcke, stolcke );
    gen    = &stolcke->gen;
    gram   = gen->grammar;
    ntsize = gram->size_lhshash;
    
    gen->probtbl[ 0 ] = 0.0;
    gen->xnumtbl[ 0 ] = 0.0;
    for( i = 1; i < gram->size_rule; ++i ) {
        prob = rb_funcall( gram->rules[i].orig, id_prob, 0 );
        if( NIL_P( prob ) )
            rb_raise( rb_eNLPException, "Probability of Rule-ID [%d] is not assigned.", i );
        gen->probtbl[ i ] = NUM2DBL( rb_Float( prob ) );
        gen->xnumtbl[ i ] = 0.0;
    }
    
    /* R_U ̍쐬 */
    compute_unit_matrix( stolcke );
    
    return Qtrue;
}


/*   `[gۊǂ
----------------------------------------------------------------------
*/
VALUE stolcke_matrix_store ( VALUE self, VALUE chart )
{
    MatrixStolcke  *stolcke;
    
    Data_Get_Struct( self, MatrixStolcke, stolcke );
    ru_is_instance_of( chart, rb_cChart );
    
    rb_ary_push( stolcke->gen.chart_ary, chart );
    
    return self;
}


/*   o^ꂽ`[gXg𓾂
----------------------------------------------------------------------
*/
VALUE stolcke_matrix_chartlist ( VALUE self )
{
    MatrixStolcke  *stolcke;
    
    Data_Get_Struct( self, MatrixStolcke, stolcke );
    
    return rb_obj_dup( stolcke->gen.chart_ary );
}


/*   [v񐔂ݒE擾
----------------------------------------------------------------------
*/
VALUE stolcke_matrix_get_loopcount ( VALUE self )
{
    MatrixStolcke  *stolcke;
    
    Data_Get_Struct( self, MatrixStolcke, stolcke );
    
    return INT2FIX( stolcke->gen.loopcount );
}

VALUE stolcke_matrix_set_loopcount ( VALUE self, VALUE count )
{
    MatrixStolcke  *stolcke;
    
    Data_Get_Struct( self, MatrixStolcke, stolcke );
    stolcke->gen.loopcount = NUM2INT( count );
    
    return count;
}


/*   x`}[Np̃p[^
----------------------------------------------------------------------
*/
VALUE stolcke_matrix_bench_update ( VALUE self )
{
    ID const id_prob = rb_intern( "pr" );
    
    MatrixStolcke   *stolcke;
    GenericStolcke  *gen;
    Grammar         *gram;
    
    Chart   *chart;
    VALUE    prob, *chartary;
    Rule    *r;
    int      i, loopcount;
    
    if( rb_block_given_p() ) /* ubN^ꂽƂ͒ʏ̃p[WO */
        return stolcke_matrix_update( self );
    
    /* {Ƃ update() ꕔ폜̂ */
    Data_Get_Struct( self, MatrixStolcke, stolcke );
    gen  = &stolcke->gen;
    gram = gen->grammar;
    
    gen->probtbl[ 0 ] = 0.0;
    for( i = 1; i < gram->size_rule; ++i ) {
        prob = rb_funcall( gram->rules[i].orig, id_prob, 0 );
        if( NIL_P( prob ) )
            rb_raise( rb_eNLPException, "Probability of Rule-ID [%d] is not assigned.", i );
        
        gen->probtbl[ i ] = NUM2DBL( rb_Float( prob ) );
    }
    
    compute_unit_matrix( stolcke );
    
    loopcount = (gen->loopcount > 0)? gen->loopcount: 1;
    for( ; loopcount > 0; --loopcount ) {
        rb_zero_memory( gen->xnumtbl, double, gram->size_rule );
        
        chartary = RARRAY(gen->chart_ary)->ptr;
        for( i = 0; i < RARRAY(gen->chart_ary)->len; ++i ) {
            stolcke_matrix_visit( self, chartary[i] );
            Data_Get_Struct( chartary[i], Chart, chart );
            chart_likelihood_internal( chart );
        }
    }
    
    return INT2FIX(0);
}


