/*                                                     project Gemini 
----------------------------------------------------------------------
    ͂̕⏕
      $Id: strparse.c,v 1.6 2002/02/08 04:33:58 Black Exp $

        Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp> 
----------------------------------------------------------------------
*/
#define DEBUGSW 0
#include "nlputil.h"


/*   xOt: pXH
----------------------------------------------------------------------
 - ͑Ώۂ
    +- (:...:) @... @...
   ̂悤ȍs
*/
static VALUE strparse_sgraph_path_p ( VALUE module, VALUE line )
{
    VALUE volatile localpath;
    VALUE volatile ret;
    char  *p, *q, *r;
    int    len, nodecount;
    
    Check_Type( line, T_STRING );
    p = RSTRING(line)->ptr;
    
    /* /\s+\+-+/ */
    if( *p != ' ' && *p != '\t' )
        return Qnil;
    for( ++p; *p == ' ' || *p == '\t'; ++p ) ;
    
    if( p[0] != '+' && p[1] != '-' )
        return Qnil;
    for( p += 2; *p == '-'; ++p ) ;
    
    /* pXm */
    localpath = rb_ary_new();
    
    /* m[hEĂ */
    while( *p != '\n' && *p != '\0' ) {
        while( *p == ' ' || *p == '\t' )
            ++p;
        if( *p == '\n' || *p == '\0' )
            break;
        
        if( p[0] == '(' && p[1] == ':' ) {
            /* (: ... :) */
            p += 2;
            while( *p == ' ' || *p == '\t' )
                ++p;
            q = p;
            while( p[0] != ':' || p[1] != ')' ) {
                if( *p == '\0' )
                    rb_raise( rb_eNLPException, "Syntax error. (':)' is not found.)" );
                ++p;
            }
            r = p;
            while( r > q && (r[-1] == ' ' || r[-1] == '\t') )
                --r;
            rb_ary_push( localpath, rb_str_new( q, r-q ) );
            p += 2;
        }
        else if( p[0] == '@' ) {
            /* @... */
            ++p;
            len = 0;
            
            /* ꌅ */
            if( !('0' <= p[0] && p[0] <= '9') )
                return Qnil;
            len += (*(p++) - '0');
            
            /* 񌅈ȍ~ */
            while( '0' <= p[0] && p[0] <= '9' ) /* ASCIIˑǁcˁH */
                len = len*10 + (*(p++) - '0');
            
            rb_ary_push( localpath, INT2FIX(len) );
        }
        else {
            rb_raise( rb_eNLPException, "Syntax error. (invalid node) (rest string=%s)", p );
        }
    }
    
    nodecount = RARRAY(localpath)->len;
    ret = rb_ary_new3( 2, INT2FIX(nodecount), localpath );
    
    return ret;
}


/*   @K
----------------------------------------------------------------------
 - ͑Ώۂ
    <...> --> <...> ... <...>
   ̂悤ȍs
*/
static const char * parse_cat ( const char * ptr, const char * * head, int * len )
{
    int  l, r;
    
    while( *ptr == ' ' || *ptr == '\t' )
        ++ptr;
    
    if( *ptr == '\0' || *ptr == '\n' )
        return NULL;
    
    *head = ptr;
    if( *ptr == '<' ) {
        ++ptr;
        l = 1;
        r = 0;
        while( l != r ) {
            switch( *(ptr++) ) {
            case '\0':
                rb_raise( rb_eNLPException, "Rule: Closing '>' is not found." );
            case '<':
                ++l;
                break;
            case '>':
                ++r;
                break;
            case '-':
                if( ptr[0] == '-' && ptr[1] == '>' )
                    ptr += 2;
                break;
            }
        }
    }
    else {
        while( *ptr != '\0'&& *ptr != '\t' && *ptr != '\n' && *ptr != ' ' && *ptr != '<' )
            ++ptr;
    }
    *len = ptr - *head;
    
    return ptr;
}

static VALUE strparse_parse_rule ( VALUE module, VALUE line )
{
    VALUE volatile ret = Qnil;
    const char  *p, *head;
    int  len;
    
    Check_Type( line, T_STRING );
    p = RSTRING(line)->ptr;
    ret = rb_ary_new();
    
    p = parse_cat( p, &head, &len );
    if( !p )
        rb_raise( rb_eNLPException, "Rule: No leftside." );
    rb_ary_push( ret, rb_str_new( head, len ) );
    
    while( *p == ' ' || *p == '\t' )
        ++p;
    if( *p != '-' )
        rb_raise( rb_eNLPException, "Rule: '-->' is not found." );
    for( ++p; *p == '-'; ++p ) ;
    if( *p != '>' )
        rb_raise( rb_eNLPException, "Rule: '-->' is not found." );
    ++p;
    
    while( p = parse_cat( p, &head, &len ) )
        rb_ary_push( ret, rb_str_new( head, len ) );
    
    return ret;
}


/*   \͌ʂ̈ʒuXg
----------------------------------------------------------------------
 - [1,2,3] Ȃ
*/
static VALUE strparse_parse_item_poslist ( VALUE module, VALUE line, VALUE ary )
{
    VALUE volatile ret = Qnil;
    const char  *p;
    int  n;
    
    Check_Type( line, T_STRING );
    Check_Type( ary, T_ARRAY );
    p = RSTRING(line)->ptr;
    rb_ary_clear( ary );
    
    while( *p == ' ' || *p == '\t' )
        ++p;
    if( *p != '[' )
        return Qnil;
    ++p;
    
    while( *p != ']' ) {
        while( *p == ' ' || *p == '\t' )
            ++p;
        
        /* ꌅ */
        n = 0;
        if( !('0' <= *p && *p <= '9') )
            rb_raise( rb_eNLPException, "Invalid item format. (invalid index)" );
        n += (*(p++) - '0');
        
        /* 񌅖ڈȍ~ */
        while( '0' <= *p && *p <= '9' ) {
            n = n*10 + (*(p++) - '0');
        }
        
        while( *p == ' ' || *p == '\t' )
            ++p;
        switch( *p ) {
        case ',':
            ++p;
        case ']':
            rb_ary_push( ary, INT2FIX(n) );
            break;
        default:
            rb_raise( rb_eNLPException, "Invalid item format. (index separator is not found)" );
        }
    }
    ++p;
    while( *p == ' ' || *p == '\t' )
        ++p;
    
    /* @K̎擾 */
    if( '0' <= *p && *p <= '9' ) {
        n = (*(p++) - '0');
        while( '0' <= *p && *p <= '9' ) {
            n = n*10 + (*(p++) - '0');
        }
        ret = INT2FIX(n);
    }
    else {
        ret = rb_str_new2( p );
    }
    
    return ret;
}


/*   xOt̃Otԍ
----------------------------------------------------------------------
*/

static VALUE strparse_sgraph_index_p ( VALUE module, VALUE line )
{
    char  *p, *q;
    int    n;
    
    Check_Type( line, T_STRING );
    p = RSTRING(line)->ptr;
    
    /* ^[\s*(\d+?)\s*] */
    if( p[0] != '[' )
        return Qnil;
    ++p;
    
    while( *p == ' ' || *p == '\t' )
        ++p;
    
    n = 0;
    if( !('0' <= *p && *p <= '9') )
        rb_raise( rb_eNLPException, "Invalid graph index format. (Invalid index)" );
    n += (*(p++) - '0');
    while( '0' <= *p && *p <= '9' )
        n = n*10 + (*(p++) - '0');
    
    while( *p == ' ' || *p == '\t' )
        ++p;
    
    if( p[0] != ']' )
        rb_raise( rb_eNLPException, "Invalid graph index format. (No ']')" );
    ++p;
    while( *p == ' ' || *p == '\t' )
        ++p;
    
    /* #... */
    if( *p == '#' ) {
        ++p;
        while( *p == ' ' || *p == '\t' )
            ++p;
        q = p;
        while( *p != '\0' && *p != '\n' )
            ++p;
        
        return rb_ary_new3( 2, INT2FIX(n), rb_str_new( q, p-q ) );
    }
    
    return rb_ary_new3( 2, INT2FIX(n), Qnil );
}


/*   StringParser W[̍\z
----------------------------------------------------------------------
*/
void init_strparse ( void )
{
    VALUE  m_strparse;
    
    /* lCeBuŕ̓W[̍쐬 */
    m_strparse = rb_define_module_under( rb_mExt, "StringParser" );
    
    /* W[֐̒` */
    rb_define_module_functions(
        m_strparse,
        
        "sgraph_path?"		, strparse_sgraph_path_p, 1,
        
        "parse_rule"		, strparse_parse_rule, 1,
        
        "parse_item_poslist"	, strparse_parse_item_poslist, 2,
        
        "sgraph_index?"		, strparse_sgraph_index_p, 1,
        
        NULL);
}


