#!/usr/local/bin/ruby
# $Id: gem.rb,v 1.12 2002/02/08 04:42:17 Black Exp $
#                                                      Project Gemini
#---------------------------------------------------------------------
#
#   graphical EM
#
#       Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp>
######################################################################
require File.dirname(File.expand_path(__FILE__)) + "/nlputil.rb"


#
#   Global Method
######################################################################

# ftHgl
$Epsilon = 0.000001

$VERSION = <<"--------"
graphical EM algorithm   ver. 1.5                         [gemini 1.2]
         Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp>

--------

$USAGE = <<"--------"
usage:
  ruby gem.rb (options) (corpus file)

options:
  -c [count]
    iteration count (> 0)
  -e [precision]
    threshold to stop EM iteration  (default = #{$Epsilon})
  -g [grammar file]
    using grammar
  --random
    force to set random probability
  --loop [count]
    set learning loop\'s count
--------


#
#   Main Program
######################################################################

getopts_ex( nil, 'c:', 'e:', 'random', 'log:', 'bench:', 'loop:' )
v_print $VERSION

$files = ARGV.dup
$files = [ STDIN ] if $files.empty?

# EM wK臒l
if $OPT_e then
  $_ = $OPT_e.to_f
  $Epsilon = $_  if $_ > 0
end

# JԂ
$OPT_c = $OPT_c.to_i if $OPT_c

# o͂Ɏgp郍Of[^p̃nbVe[u
$ResultLog = { }


#
#   Prepare grammar
######################################################################

if $grammar then
  v_print { " + grammar = #{$OPT_g}\n" }
else
  $grammar = Grammar.new()
  v_print { " + generated empty grammar.\n" }
end


#
#   Read corpus
######################################################################

def print_progress ()
  v_print { sprintf( "    - \#%d (%.1f KB)\r", $total, $memtotal/1024 ) }
end

$total     = 0
$memtotal  = 0
$sglist    = [ ] if $OPT_log
$extsglist = [ ]

time_block do
  $files.each do |filename|
    v_print { " + #{io_to_name(filename)}\n" }
    if String === filename then
      unless File::exist?( filename ) then
        if $OPT_v then
          v_print { "   ! No such file\n" }
        else
          $stderr << "No such file -- #{filename}\n"
        end
        next
      end
      stream = open( filename )
    else
      stream = filename
    end

    CorpusStream.new( stream ) do |root|
      root.each do |corpus|
        t = Time.now.to_i ; print_progress()
        corpus.each do |sentence|
          $total += 1
          begin
            sg = nil
            sentence.each( SupportGraphBlock ) do |c, block|
              next unless c
              if sg then
                raise NLPException, "Found too many <dSupportGraph>."
              else
                sg = c.new(block).read
              end
            end
            next unless sg
            
            #sg.compile( $grammar )
            extsg = Ext::SupportGraph.new( sg )
            $memtotal += extsg.memsize
            
            $sglist << sg if $OPT_log
            $extsglist << extsg
            if t+1 <= Time.now.to_i then
              t = Time.now.to_i ; print_progress()
            end
          rescue
            v_print "\n"
            $stderr << "\##{$total} (#{root.stream.path}:#{root.stream.lineno}): #{$!}\n"
            t = 0
          end
        end # corpus.each
        print_progress() ; v_print "\n"
      end # root.each
    end # CorpusStream.new
  end
end
readtime = $time.ptime
if $extsglist.empty? then
  $stderr << "Error: No supportgraphs.\n"
  exit
end


#
#   Create learner
######################################################################

$extgrammar = toPCFG( Ext::Grammar.new( $grammar ), $OPT_random )
if $OPT_bench then
  $learner = Ext::GraphicalEM::BenchMark.new( $extgrammar )
  $OPT_loop = $OPT_bench
else
  $learner = Ext::GraphicalEM.new( $extgrammar )
end
if $OPT_loop then
  $learner.loopcount = $ResultLog['loopcount'] = $OPT_loop.to_i
  v_print { " + loopcount = #{$learner.loopcount}\n" }
end
$extsglist.each do |extsg|
  $learner << extsg
end
$extsglist = [ ]


#
#   Helper
######################################################################

def print_likelihood ()
  v_print do
    tstr = Time.now.strftime('%Y/%m/%d %H:%M:%S')
    sprintf( "\#%4d (#{tstr}): likelihood = %.6e (%f sec)\n", $itrcount, $likelihood, $itrtime )
  end
end


#
#   Learning
######################################################################

HistoryData = Struct.new( :num, :likelihood, :tms )
$history  = [ ]
$itrcount = 0
time_learn = 0.0

if $OPT_log then
  logfile = open( $OPT_log, "w" )
  extsglist = $learner.sglist
end

loopcount = $OPT_c || -1
begin
  until loopcount >= 0 and $itrcount >= loopcount do
    $likelihood = time_block { $learner.update() }
    $itrtime    = $time.ptime
    $itrcount  += $learner.loopcount
    time_learn += $itrtime
    
    print_likelihood()
    $history << HistoryData.new( $itrcount, $likelihood, $time )
    
    if $OPT_log then
      logfile << "---- iteration \##{$itrcount}\n"
      extsglist.each_with_index do |extsg, n|
        sg = $sglist[n]
        for i in 0 ... extsg.size_graph do
          logfile << sprintf( "%s :: i=%g, o=%g\n", sg.comment[i], extsg.inside(i), extsg.outside(i) )
        end
        logfile << "\n"
      end
    end

    break if $likelihood == 0
    if $history.size > 1 then
      delta = $history[-1].likelihood - $history[-2].likelihood
      if delta < 0 then
        $stderr << "Error: Likelihood is down. (fatal error)\n"
        exit
      end
      break if delta < $Epsilon
    end
  end
rescue NLPException
  $stderr << "Error: itr. \##{$itrcount}: #{$!}\n"
end


#
#   Output parameter
######################################################################

t_per_itr = ($itrcount <= 0)? 0: time_learn / $itrcount

$ResultLog.update(
  'corpus-size' => $learner.size,
  'itr-count'   => $itrcount,
  'read-time'   => "#{readtime} \# sec"
)
$ResultLog['learn-time'] = "#{t_per_itr} \# sec/itr" if $itrcount > 0

$grammar.display( $defout )
Ext::Block.display( "dLog" ) do
  $defout << hash_to_pstr( $ResultLog )
end
v_print { " + finished.\n" }

