#!/usr/local/bin/ruby
# $Id: gutil.rb,v 1.4 2002/02/08 04:43:07 Black Exp $
#                                                      Project Gemini
#---------------------------------------------------------------------
#
#   Grammar Utility
#
#       Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp>
######################################################################
require File.dirname(File.expand_path(__FILE__)) + "/nlputil.rb"


#
#   Global Method
######################################################################

$VERSION = <<_VERSIONEND_
Grammar Utility   ver. 1.1                                [gemini 1.2]
         Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp>

_VERSIONEND_

$USAGE = <<_USAGEEND_
usage:
  gutil.rb [mode] (option)

mode:
  first [grammar file]
    Output the grammar added FIRST set.

  check [grammar files]
    Check the grammars.
_USAGEEND_


#
#   FIRST
######################################################################

class ArrayIO
  def initialize ( src )
    @lines  = src
    @lineno = 0
  end
  def gets ( rs = nil )
    @lineno += 1
    @lines.shift
  end
  def close ()
    @lines.clear
  end
  attr_reader :lineno
  def concat ( line )
    @lines.push( line )
  end
  alias write concat
  alias << concat
end

def mode_first ()
  v_print { " + MODE = first\n" }
  
  grammarfile = $file.first
  if grammarfile then
    unless File::exist?( grammarfile ) then
      $stderr << "ERROR: No such file -- #{grammarfile}\n"
      exit
    end
    stream = open( grammarfile )
    grampath = io_to_name( stream )
  else
    $gramlines = $stdin.readlines
    grampath = io_to_name( $stdin )
    stream = ArrayIO.new( $gramlines.dup )
  end
  
  # @̓ǂݍ
  v_print { "   - #{grampath}\n" }
  begin
    $grammar = stream_read_grammar( stream )
  rescue NLPException
    $stderr << $! << "\n"
    exit
  end
  unless $grammar then
    $stderr << "No <pGrammar> in #{grampath}.\n"
    exit
  end
  
  # FIRST W̌vZ
  rules    = $grammar.rules
  firstset = Array.new( $grammar.size_nonterm, nil )
  v_print { "   - #{rules.size} rules\n" \
            "   - #{$grammar.size_nonterm} non-terminals\n" }
  rules.each_with_index do |rule, n|
    next unless rule
    top = rule.lhs.to_i
    lc  = rule.rhs.first.to_i
    firstset[top] = { } unless firstset[top]
    firstset[top][lc] = true
  end
  
  buffer  = { }
  changed = true
  itr     = 1
  t       = Time.now.to_i
  while changed do
    changed = false
    firstset.each_with_index do |localset, cat|
      # localset == FIRST(cat)
      if t+1 <= Time.now.to_i then
        t = Time.now.to_i
        v_print { sprintf( " + #%d : %5d\r", itr, cat ) }
      end
      next unless localset
      buffer.clear
      prevsize = localset.size
      localset.each_key do |keycat|
        next unless firstset[keycat]
        firstset[keycat].each_key do |i|
          buffer[i] = true
        end
      end
      localset.update( buffer )
      changed = true if prevsize != localset.size
    end
    itr += 1 if changed
  end
  firstset.collect!{|localset| (localset)? localset.keys: nil }
  v_print " + finished.          \n"
  
  # o
  if grammarfile then
    stream = open( grammarfile )
  else
    stream = ArrayIO.new( $gramlines )
  end
  root = Ext::Block.open( stream )
  
  root.out = $defout
  begin
    root.each do |block|
      if Grammar === block then
        block.display do
          block.each do |elem|
            if /^[dD]([fF]irst|FIRST)$/ === elem.name then
              elem.skip()
            else
              elem.dump()
            end
          end
          block.close()
          
          Ext::Block.display( "dFIRST" ) do
            lines = [ ]
            firstset.each_with_index do |list, n|
              next unless list
              list.sort!
              list.collect!{|i| $grammar.getcat(i).to_s }
              $defout << "FIRST( #{$grammar.getcat(n)} ) = #{list.join(' ')}\n"
            end
          end # <dFIRST>.display
        end # block.display
        root.close()
        break
      end
    end
  end
end


#
#   check
######################################################################

def mode_check ()
  v_print { " + MODE = check\n" }
  
  while grammarfile = $file.shift do
    unless File::exist?( grammarfile ) then
      $stderr << "Warning: Grammar `#{grammarfile}' is not found.\n"
      next
    end
    
    # @̓ǂݍ
    v_print { " + #{grammarfile}\n" }
    begin
      $grammar = stream_read_grammar( open( grammarfile ) )
    rescue NLPException
      $stderr << $! << "\n"
      next
    end
    nofailure = true
    
    # L̃`FbN
    v_print { "   - check symbols ...\n" }
    catlist = $grammar.categories
    catlist.each do |cat|
      n = cat.to_s
      if cat.preterm? then
        if /<|>/ === n then
          $defout << "Warning: #{grammarfile}: `#{n}'(\##{cat.to_i}) has isolated < or >.\n"
          nofailure = false
        end
      else
        if /\s/ === n then
          $defout << "Warning: #{grammarfile}: `#{n}'(\##{cat.to_i}) has a space character.\n"
          nofailure = false
        end
      end
    end
    
    # K̃`FbN
    v_print { "   - check rules ...\n" }
    rulelist = $grammar.rules
    rulelist.each do |r|
      next unless r
      pr = r.pr.to_f
      if pr and (pr < 0.0 or pr > 1.0) then
        $defout << "Error: #{grammarfile}: Probability of rule \##{r.to_i} is out of range. (p=#{pr})\n"
        nofailure = false
      end
    end
    
    if nofailure then
      $defout << " * #{grammarfile} is ready.\n"
    else
      $defout << " * #{grammarfile} is not ready.\n"
    end
  end
end


#
#   Main Program
######################################################################

getopts_ex( nil )
v_print $VERSION

$mode = ARGV.shift
$file = ARGV.dup
case $mode
when "first", "FIRST" then
  mode_first()
when "check" then
  mode_check()
else
  if $mode then
    $stderr << "ERROR: Mode `#{$mode}' is unknown."
  end
  printUsage()
end

