#!/usr/local/bin/ruby
# $Id: logcmp.rb,v 1.5 2002/02/08 04:43:07 Black Exp $
#                                                      Project Gemini
#---------------------------------------------------------------------
#
#   Log Comparison Utility
#
#       Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp>
######################################################################
require "getopts.rb"

getopts( '', 'noonly' )

$files = ARGV.dup
if $files.size != 2 then
  $stderr << <<'----'
Log Comparison Utility   ver. 1.0                         [gemini 1.2]
         Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp>

usage:
  ruby logcmp.rb (option) [log1] [log2]

option:
  --noonly
    delete lines appeared in one side only
----
end

# f[^̎荞
RE_NULLLINE = /\A\s*\z/
$data = [ ]
$files.each do |filename|
  open( filename ) do |input|
    image = input.read()
    itrary = image.split( /^-+\s+iteration\s+\#\d+$/ )
    itrary.delete_if do |elem|
      RE_NULLLINE === elem # svȕ؂̂Ă
    end
    itrary.map! do |itr|
      sentary = itr.split( /^\s*$/ )
      sentary.delete_if do |elem|
        RE_NULLLINE === elem # svȕ؂̂Ă
      end
      sentary.map! do |sent|
        sent.split( /\n/ )
      end
      sentary
    end
    $data << itrary
  end
end

# f[^̔r
RE_LINESPLITTER = / :: /
RE_IGNORE = / o=0$/

endindex = [ $data[0].size, $data[1].size ].min
$data[0] = $data[0][ 0 ... endindex ]
$data[1] = $data[1][ 0 ... endindex ]
$defout << "log1: #{$files[0]}\nlog2: #{$files[1]}\n=====\n\n"

for i in 0...endindex do
  sentary = [ $data[0][i], $data[1][i] ]
  sentcount = 0
  loop do
    lineary = [ sentary[0][sentcount], sentary[1][sentcount] ]
    if !lineary[0] and !lineary[1] then
      break
    end
    if !lineary[0] then
      $defout << "sentence #{sentcount}...#{sentary[1].size} of #{$files[1]} are ignored.\n"
      break
    end
    if !lineary[1] then
      $defout << "sentence #{sentcount}...#{sentary[0].size} of #{$files[0]} are ignored.\n"
      break
    end
    $defout << "itr. \##{i+1}, sent. \##{sentcount+1}\n****\n"
    
    table = { }
    only = [ [ ], [ ] ]
    lineary[0].each do |line|
      elem = line.split( RE_LINESPLITTER )
      table[ elem[0] ] = line
    end
    lineary[1].each do |line|
      elem = line.split( RE_LINESPLITTER )
      opp = table.delete( elem[0] )
      if opp then
        # ̃Oŏof[^
        if line != opp then
          unless RE_IGNORE === line or RE_IGNORE === opp then
            $defout << "1: #{opp}\n2: #{line}\n----\n"
          end
        end
      else
        # O 2 ɂȂf[^
        only[1] << line
      end
    end
    only[0] = table.values
    unless $OPT_noonly then
      $defout << "\n(only #{$files[0]})\n" << only[0].join( "\n" ) << "\n" unless only[0].empty?
      $defout << "\n(only #{$files[1]})\n" << only[1].join( "\n" ) << "\n" unless only[1].empty?
    end
    $defout << "\n"
    
    sentcount += 1
  end
end

