#!/usr/local/bin/ruby
# $Id: mksg.rb,v 1.9 2002/02/08 04:43:07 Black Exp $
#                                                      Project Gemini
#---------------------------------------------------------------------
#
#   ItemSet --> SupportGraph converter
#
#       Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp>
######################################################################
require File.dirname(File.expand_path(__FILE__)) + "/nlputil.rb"


#
#   Global Method
######################################################################

# ftHgl
$VERSION = <<"--------"
ItemSet to SupportGraph converter   ver. 1.3              [gemini 1.2]
         Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp>

--------

$USAGE = <<"--------"
usage:
  ruby mksg.rb (options) (corpus file)

options:
  --cyk
    output CYK-supportgraph  (default)
  --earley
    output Earley-supportgraph
--------


#
#   Main Program
######################################################################

$grammar = Grammar.new()

getopts_ex( nil, 'cyk', 'earley' )
v_print $VERSION

$files = ARGV.dup
$files = [ STDIN ] if $files.empty?


#
#   Conversion
######################################################################

def print_progress ()
  v_print { "    - \##{$total}\r" }
end

$total = 0

Ext::Block.display( "pCorpus" ) do
  $files.each do |filename|
    v_print { " + #{io_to_name(filename)}\n" }
    if String === filename then
      unless File::exist?( filename ) then
        if $OPT_v then
          v_print { "   ! No such file\n" }
        else
          $stderr << "No such file -- #{filename}\n"
        end
        next
      end
      stream = open( filename )
    else
      stream = filename
    end
    
    CorpusStream.new( stream ) do |root|
      root.each do |corpus|
        t = Time.now.to_i ; print_progress()
        corpus.each do |sentence|
          $total += 1
          begin
            sentence.out = $defout
            sentence.display do
              
              # ̓ǂݍ
              itemset = nil
              sentence.each( ItemSetBlock ) do |c, block|
                if c then
                  if itemset then
                    raise NLPException, "Found too many <dItemSet>."
                  else
                    block.display do
                      itemset = c.new( block ).read
                    end
                  end
                else
                  if SupportGraphBlock === block then
                    block.skip
                  else
                    block.dump
                  end
                end
              end
              next unless itemset
              
              # ϊ
              itemset.compile( $grammar )
              extitemset = Ext::ItemSet.new( itemset )
              Ext::Block.display( "dSupportGraph" ) do
                if $OPT_earley then
                  extitemset.display_earley( $defout )
                else
                  extitemset.display_cyk( $defout )
                end
              end
              if t+1 <= Time.now.to_i then
                t = Time.now.to_i ; print_progress()
              end
            end # sentence.display
          rescue
            v_print "\n"
            $stderr << "\##{$total} (#{root.stream.path}:#{root.stream.lineno}): #{$!}\n"
            t = 0
          end
        end # corpus.each
        print_progress() ; v_print { "\n" }
      end # root.each
    end # CorpusStream.new
  end # $files.each
end # <pCorpus>.display
v_print { " + finished.\n" }

